#!/usr/bin/env python3
import json
import os
import sqlite3
import uuid
from datetime import datetime, timezone
from pathlib import Path

from django.core.management.base import BaseCommand, CommandError
from django.db.models import CharField, Value
from django.db.models.functions import Cast, Concat
from teklia_toolbox.time import Timer

from arkindex.documents.export import BATCH_SIZE
from arkindex.documents.models import (
    Classification,
    Corpus,
    Element,
    ElementPath,
    ElementType,
    Entity,
    EntityLink,
    EntityRole,
    EntityType,
    MetaData,
    MLClass,
    Transcription,
    TranscriptionEntity,
)
from arkindex.images.models import Image, ImageServer
from arkindex.process.models import Repository, Revision, Worker, WorkerType, WorkerVersion
from arkindex.users.models import Role, User

EXPORT_VERSION = 5

TABLE_NAMES = {
    'export_version',
    'image_server',
    'image',
    'worker_version',
    'element',
    'element_path',
    'entity',
    'entity_type',
    'entity_role',
    'entity_link',
    'transcription',
    'transcription_entity',
    'metadata',
    'classification',
}

SQL_TABLES_QUERY = "SELECT name FROM sqlite_master WHERE type ='table' AND name NOT LIKE 'sqlite_%'"
SQL_VERSION_QUERY = "SELECT version FROM export_version"

SQL_ELEMENT_TYPE_QUERY = """
    SELECT DISTINCT element.type as slug, max(element.polygon) IS NULL as folder
    FROM element
    GROUP BY element.type
"""

SQL_ML_CLASS_QUERY = "SELECT DISTINCT class_name FROM classification"

SQL_REPOSITORY_QUERY = "SELECT DISTINCT repository_url FROM worker_version"
SQL_REVISION_QUERY = "SELECT DISTINCT revision, repository_url FROM worker_version"
SQL_WORKER_TYPE_QUERY = "SELECT DISTINCT type FROM worker_version"
SQL_WORKER_VERSION_QUERY = "SELECT * FROM worker_version"

SQL_IMAGE_SERVER_QUERY = "SELECT * FROM image_server"
SQL_IMAGE_QUERY = """
    SELECT image.*, image_server.url as server_url
    FROM image
    INNER JOIN image_server ON (image_server.id = image.server_id)
"""

# A temporary table used to map existing image IDs found in the PostgreSQL database
# to SQLite image IDs using the image URL.
SQL_IMAGE_MAP_TABLE = """
CREATE TEMPORARY TABLE image_map (
    postgres_id VARCHAR(37) NOT NULL,
    url TEXT NOT NULL,
    PRIMARY KEY (postgres_id),
    UNIQUE (url)
)
"""

SQL_ELEMENT_QUERY = """
    SELECT
        element.id,
        name,
        type,
        COALESCE(image_map.postgres_id, image_id) AS image_id,
        polygon,
        rotation_angle,
        mirrored,
        worker_version_id,
        confidence
    FROM element
    LEFT JOIN image ON (image.id = element.image_id)
    LEFT JOIN image_map ON (image_map.url = image.url)
"""
SQL_ELEMENT_PATH_QUERY = "SELECT * FROM element_path"
SQL_PARENT_QUERY = "SELECT parent_id FROM element_path WHERE child_id = '{}'"
SQL_TOP_LEVEL_PATH_QUERY = """
    SELECT element.id
    FROM element
    WHERE NOT EXISTS (
        SELECT 1
        FROM element_path
        WHERE child_id = element.id
        LIMIT 1
    )
"""

SQL_ENTITY_QUERY = "SELECT * FROM entity"
SQL_ENTITY_TYPE_QUERY = "SELECT * FROM entity_type"
SQL_ENTITY_ROLE_QUERY = "SELECT * FROM entity_role"
SQL_ENTITY_LINK_QUERY = "SELECT * FROM entity_link"

SQL_TRANSCRIPTION_QUERY = "SELECT * FROM transcription"
SQL_TRANSCRIPTION_ENTITY_QUERY = "SELECT * FROM transcription_entity"

SQL_METADATA_QUERY = "SELECT * FROM metadata"

SQL_CLASSIFICATION_QUERY = "SELECT * FROM classification"


class Command(BaseCommand):
    help = "Import an SQLite database generated by an Arkindex export"

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            "db_path", help="The path to the SQL database file", type=Path,
        )
        parser.add_argument(
            "--email", help="The user email to give corpus rights", type=str, required=True
        )
        parser.add_argument("--corpus-name", help="The name of the corpus to create")

    def sql_chunk(self, sql_query):
        execute = self.db.execute(sql_query)
        while True:
            chunk = execute.fetchmany(BATCH_SIZE)
            if not chunk:
                return
            yield chunk

    def build_element_paths(self, parent_id):
        """
        Reconstructs the paths of a child element on a single parent element at a time.
        The SQLite database only stores direct links, not arrays like in Arkindex,
        so this function calls itself recursively to reconstruct the arrays.
        """
        sql_query = SQL_PARENT_QUERY.format(parent_id)
        found = False

        for db_chunk in self.sql_chunk(sql_query):
            for row in db_chunk:
                parent_paths = self.build_element_paths(row["parent_id"])
                for parent_path in parent_paths:
                    found = True
                    yield parent_path + [parent_id]

        if not found:
            # When this element has no parents, return a path with the element alone
            yield [parent_id]

    def convert_images(self, row):
        assert row["url"].startswith(row["server_url"]), "The url of the image does not start with the url of its server"
        path = row["url"][len(row["server_url"].strip('/')) + 1:]
        return [Image(
            id=row["id"],
            path=path,
            width=row["width"],
            height=row["height"],
            server_id=self.image_server_map[row["server_id"]],
        )]

    def convert_elements(self, row):
        return [Element(
            id=row["id"],
            name=row["name"],
            type_id=self.element_type_map[row["type"]],
            polygon=json.loads(row["polygon"]) if row["polygon"] else None,
            image_id=row["image_id"],
            rotation_angle=row["rotation_angle"],
            mirrored=row["mirrored"],
            worker_version_id=self.worker_version_map[row["worker_version_id"]] if row["worker_version_id"] else None,
            confidence=row["confidence"],
            corpus=self.corpus
        )]

    def convert_element_paths(self, row):
        paths = self.build_element_paths(row["parent_id"])
        return [ElementPath(
            element_id=row["child_id"],
            path=path,
            ordering=row["ordering"]
        ) for path in paths]

    def convert_top_level_paths(self, row):
        return [ElementPath(
            element_id=row["id"],
            path=[],
            ordering=0
        )]

    def convert_entities(self, row):
        return [Entity(
            id=row["id"],
            name=row["name"],
            type_id=row["type_id"],
            validated=row["validated"],
            moderator=User.objects.get(email=row["moderator"]) if row["moderator"] else None,
            metas=json.loads(row["metas"]) if row["metas"] else None,
            worker_version_id=self.worker_version_map[row["worker_version_id"]] if row["worker_version_id"] else None,
            corpus=self.corpus
        )]

    def convert_entity_types(self, row):
        return [EntityType(
            id=row["id"],
            name=row["name"],
            color=row["color"],
            corpus=self.corpus
        )]

    def convert_entity_roles(self, row):
        return [EntityRole(
            id=row["id"],
            parent_name=row["parent_name"],
            child_name=row["child_name"],
            parent_type_id=row["parent_type_id"],
            child_type_id=row["child_type_id"],
            corpus=self.corpus
        )]

    def convert_entity_links(self, row):
        return [EntityLink(
            id=row["id"],
            parent_id=row["parent_id"],
            child_id=row["child_id"],
            role_id=row["role_id"],
        )]

    def convert_transcriptions(self, row):
        return [Transcription(
            id=row["id"],
            element_id=row["element_id"],
            text=row["text"],
            confidence=row["confidence"],
            orientation=row["orientation"],
            worker_version_id=self.worker_version_map[row["worker_version_id"]] if row["worker_version_id"] else None,
        )]

    def convert_transcription_entities(self, row):
        return [TranscriptionEntity(
            id=row["id"],
            transcription_id=row["transcription_id"],
            entity_id=row["entity_id"],
            offset=row["offset"],
            length=row["length"],
            worker_version_id=self.worker_version_map[row["worker_version_id"]] if row["worker_version_id"] else None,
            confidence=row["confidence"],
        )]

    def convert_metadatas(self, row):
        return [MetaData(
            id=row["id"],
            element_id=row["element_id"],
            name=row["name"],
            type=row["type"],
            value=row["value"],
            entity_id=row["entity_id"],
            worker_version_id=self.worker_version_map[row["worker_version_id"]] if row["worker_version_id"] else None,
        )]

    def convert_classifications(self, row):
        return [Classification(
            id=row["id"],
            element_id=row["element_id"],
            ml_class_id=self.ml_class_map[row["class_name"]],
            state=row["state"],
            moderator=User.objects.get(email=row["moderator"]) if row["moderator"] else None,
            confidence=row["confidence"],
            high_confidence=row["high_confidence"],
            worker_version_id=self.worker_version_map[row["worker_version_id"]] if row["worker_version_id"] else None,
        )]

    def bulk_create_objects(self, ModelClass, convert_method, sql_query, ignore_conflicts=True):
        # Model name for logs
        verbose_name_plural = ModelClass._meta.verbose_name_plural.lower()
        verbose_name = ModelClass._meta.verbose_name.capitalize()

        self.stdout.write(f"Creating {verbose_name_plural}")

        with Timer() as t:
            count, failed = 0, 0
            for db_chunk in self.sql_chunk(sql_query):
                # Create instances
                objects = []
                for row in db_chunk:
                    try:
                        objects.extend(convert_method(row))
                    except Exception as e:
                        self.stderr.write(self.style.WARNING(f"{verbose_name} conversion failed: {e}"))
                        failed += 1

                # Create objects in bulk
                try:
                    ModelClass.objects.bulk_create(objects, ignore_conflicts=ignore_conflicts)
                    count += len(objects)
                except Exception as e:
                    self.stderr.write(self.style.WARNING(f"{verbose_name_plural.title()} insertion failed: {e}"))
                    failed += len(objects)

        self.stdout.write(f"Ran on {count+failed} rows: {count} completed, {failed} failed in {t.delta}")

    def create_objects(self, ModelClass, convert_method, sql_query):
        """
        Create objects one by one, with custom conflict resolution managed by a conversion method.

        The `convert_method` takes an SQLite row and should return the same arguments as a `ModelClass.objects.get_or_create`:
        an instance of the model and a boolean indicating if the row was created or already existed.
        If the import of a single row cannot be guaranteed for any reason, the conversion method should cause an exception.

        Returns a dict mapping the first column of the SQLite SQL query to the PostgreSQL ID of the object returned by the conversion method.
        """
        # Model name for logs
        verbose_name_plural = ModelClass._meta.verbose_name_plural.lower()
        verbose_name = ModelClass._meta.verbose_name.capitalize()
        id_map = {}

        self.stdout.write(f"Importing {verbose_name_plural}")

        with Timer() as t:
            total, created_count, error_count = 0, 0, 0
            for db_chunk in self.sql_chunk(sql_query):
                for row in db_chunk:
                    try:
                        instance, created = convert_method(row)
                        assert row[0] not in id_map, f'Duplicate identifier {row[0]!r}'
                        id_map[row[0]] = instance.id
                        created_count += created
                    except Exception as e:
                        self.stderr.write(self.style.WARNING(f"{verbose_name} import failed: {e}"))
                        error_count += 1
                    total += 1

        self.stdout.write(f"Ran on {total} rows: {created_count} created, {error_count} errors in {t.delta}")
        return id_map

    def create_element_type(self, row):
        return self.corpus.types.get_or_create(
            display_name=row["slug"].title().replace("_", " "),
            slug=row["slug"],
            folder=row["folder"],
        )

    def create_ml_class(self, row):
        return self.corpus.ml_classes.get_or_create(name=row["class_name"])

    def create_repository(self, row):
        repo, created = Repository.objects.get_or_create(
            url=row['repository_url'],
            defaults={
                'hook_token': str(uuid.uuid4()),
            },
        )
        return repo, created

    def create_revision(self, row):
        return Revision.objects.get_or_create(
            repo_id=self.repository_map[row['repository_url']],
            hash=row['revision'],
            defaults={
                "message": "Fake revision",
                "author": self.user.display_name,
            }
        )

    def create_worker_type(self, row):
        return WorkerType.objects.get_or_create(
            slug=row["type"],
            defaults={
                "display_name": row["type"]
            }
        )

    def create_worker(self, row):
        return Worker.objects.get_or_create(
            slug=row["slug"],
            repository_id=self.repository_map[row['repository_url']],
            defaults={
                "name": row["name"],
                "type_id": self.worker_type_map[row["type"]],
            }
        )

    def create_worker_version(self, row):
        return WorkerVersion.objects.get_or_create(
            worker_id=self.worker_map[row["id"]],
            revision_id=self.revision_map[row["revision"]],
            defaults={
                'configuration': {'configuration': {}}
            }
        )

    def create_image_server(self, row):
        return ImageServer.objects.get_or_create(
            url=row['url'],
            defaults={
                'display_name': row['display_name'],
                'max_width': row['max_width'],
                'max_height': row['max_height'],
            }
        )

    def create_elements(self):
        # Build a temporary table to map existing PostgreSQL image IDs to SQLite image IDs and avoid conflicts
        self.db.execute(SQL_IMAGE_MAP_TABLE)
        self.db.executemany(
            "INSERT INTO image_map (postgres_id, url) VALUES (?, ?)",
            Image.objects.filter(server_id__in=self.image_server_map.values())
            .values_list(
                # sqlite3 is not happy about getting uuid.UUID parameters.
                Cast('id', output_field=CharField()),
                Concat(
                    'server__url',
                    Value('/'),
                    # Django is not happy about concatenating a CharField and a TextField.
                    Cast('path', output_field=CharField())
                )
            )
        )

        # Create elements
        self.bulk_create_objects(Element, self.convert_elements, SQL_ELEMENT_QUERY)

        # Drop the table early
        self.db.execute("DROP TABLE image_map")

    def handle(self, db_path, email, **options):
        # Check the database file
        if os.path.splitext(db_path.name)[1].lower() not in [".db", ".sqlite"]:
            raise CommandError(f"File {db_path} is not an SQLite database")

        if not os.path.isfile(db_path):
            raise CommandError(f"File {db_path} does not exist")

        # Check if user exists
        try:
            self.user = User.objects.get(email=email)
        except User.DoesNotExist:
            raise CommandError(f"User with the email {email} does not exist")

        self.db = sqlite3.connect(db_path)
        self.db.row_factory = sqlite3.Row

        # Check database tables
        db_results = self.db.execute(SQL_TABLES_QUERY).fetchall()
        if not set([table["name"] for table in db_results]) == TABLE_NAMES:
            raise CommandError(f"The SQLite database {db_path} is not a correct Arkindex export")

        # Check export version
        db_results = self.db.execute(SQL_VERSION_QUERY).fetchall()
        if len(db_results) != 1 or db_results[0]["version"] != EXPORT_VERSION:
            raise CommandError(f"The SQLite database {db_path} does not have the correct export version")

        # Retrieve corpus name
        date = datetime.now(timezone.utc).strftime("%Y-%m-%d %H:%M")
        corpus_name = options.get("corpus_name")
        if not corpus_name:
            # An Arkindex export has a name like "corpus_name-%Y%m%d-%H%M%S.sqlite"
            split_name = db_path.name.split("-")
            corpus_name = " ".join(split_name[:len(split_name) - 2]).title()

        if not corpus_name:
            corpus_name = f"Corpus import {date}"

        self.stdout.write(f"Creating corpus {corpus_name}")
        with Timer() as t:
            # Create corpus
            self.corpus = Corpus.objects.create(name=corpus_name, description=f"Corpus import {date}")
            self.corpus.memberships.create(
                user=self.user,
                level=Role.Admin.value
            )

            # Create element types
            self.element_type_map = self.create_objects(ElementType, self.create_element_type, SQL_ELEMENT_TYPE_QUERY)

            # Create ML classes
            self.ml_class_map = self.create_objects(MLClass, self.create_ml_class, SQL_ML_CLASS_QUERY)

            # Create worker versions
            self.repository_map = self.create_objects(Repository, self.create_repository, SQL_REPOSITORY_QUERY)
            self.revision_map = self.create_objects(Revision, self.create_revision, SQL_REVISION_QUERY)
            self.worker_type_map = self.create_objects(WorkerType, self.create_worker_type, SQL_WORKER_TYPE_QUERY)
            self.worker_map = self.create_objects(Worker, self.create_worker, SQL_WORKER_VERSION_QUERY)
            self.worker_version_map = self.create_objects(WorkerVersion, self.create_worker_version, SQL_WORKER_VERSION_QUERY)

            # Create images and servers
            self.image_server_map = self.create_objects(ImageServer, self.create_image_server, SQL_IMAGE_SERVER_QUERY)
            self.bulk_create_objects(Image, self.convert_images, SQL_IMAGE_QUERY)

            self.create_elements()

            # Create element paths
            self.bulk_create_objects(ElementPath, self.convert_element_paths, SQL_ELEMENT_PATH_QUERY, ignore_conflicts=False)
            self.bulk_create_objects(ElementPath, self.convert_top_level_paths, SQL_TOP_LEVEL_PATH_QUERY, ignore_conflicts=False)

            # Create entities, entity types, roles and links
            self.bulk_create_objects(EntityType, self.convert_entity_types, SQL_ENTITY_TYPE_QUERY)
            self.bulk_create_objects(Entity, self.convert_entities, SQL_ENTITY_QUERY)
            self.bulk_create_objects(EntityRole, self.convert_entity_roles, SQL_ENTITY_ROLE_QUERY)
            self.bulk_create_objects(EntityLink, self.convert_entity_links, SQL_ENTITY_LINK_QUERY)

            # Create transcriptions and transcription entities
            self.bulk_create_objects(Transcription, self.convert_transcriptions, SQL_TRANSCRIPTION_QUERY)
            self.bulk_create_objects(TranscriptionEntity, self.convert_transcription_entities, SQL_TRANSCRIPTION_ENTITY_QUERY)

            # Create metadatas
            self.bulk_create_objects(MetaData, self.convert_metadatas, SQL_METADATA_QUERY)

            # Create classifications
            self.bulk_create_objects(Classification, self.convert_classifications, SQL_CLASSIFICATION_QUERY)

        self.stdout.write(self.style.SUCCESS(f"Created corpus {corpus_name} in {t.delta}"))

        self.db.close()
