import math
import uuid
from collections import defaultdict
from textwrap import dedent

from django.conf import settings
from django.contrib.gis.geos import LinearRing
from django.core.exceptions import ValidationError as DjangoValidationError
from django.core.validators import URLValidator
from django.db import transaction
from drf_spectacular.utils import extend_schema_field, extend_schema_serializer
from rest_framework import serializers
from rest_framework.exceptions import ValidationError

from arkindex.documents.models import Corpus, Element, ElementPath, ElementType, Entity, MetaData, MetaType
from arkindex.documents.serializers.entities import BaseEntitySerializer
from arkindex.documents.serializers.light import (
    CorpusLightSerializer,
    ElementLightSerializer,
    ElementTypeLightSerializer,
    MetaDataLightSerializer,
)
from arkindex.documents.serializers.ml import ClassificationSerializer, WorkerRunSummarySerializer
from arkindex.images.models import Image
from arkindex.images.serializers import ZoneSerializer
from arkindex.ponos.utils import is_admin_or_ponos_task
from arkindex.process.models import WorkerRun, WorkerVersion
from arkindex.project.fields import Array
from arkindex.project.mixins import SelectionMixin
from arkindex.project.serializer_fields import ForbiddenField, LinearRingField
from arkindex.project.tools import polygon_outside_image
from arkindex.project.triggers import add_parent_selection, move_selection
from arkindex.project.validators import ForbiddenValidator
from arkindex.users.models import Group, Right, Role
from arkindex.users.utils import get_max_level


def metadata_number_validator(data, serializer):
    # Those can be None if someone messed up their input; DRF will catch it itself afterwards, so we ignore None.
    meta_type = data.get('type', serializer.instance.type if serializer.instance else None)
    value = data.get('value', serializer.instance.value if serializer.instance else None)

    if meta_type != MetaType.Numeric or value is None:
        return

    try:
        # Python allows _ to separate numbers (123_456.789_101), but Postgres does not, so we forbid them.
        assert '_' not in value
        value = float(value)
        assert not math.isinf(value)
        assert not math.isnan(value)
    except (AssertionError, TypeError, ValueError):
        raise serializers.ValidationError({
            'value': ['The value of a numeric metadata should be a valid floating-point number.']
        })


def metadata_url_validator(data, serializer):
    # Those can be None if someone messed up their input; DRF will catch it itself afterwards, so we ignore None.
    meta_type = data.get('type', serializer.instance.type if serializer.instance else None)
    value = data.get('value', serializer.instance.value if serializer.instance else None)

    if meta_type != MetaType.URL or value is None:
        return

    try:
        URLValidator(schemes=['http', 'https'])(value)
    except DjangoValidationError:
        raise serializers.ValidationError({
            'value': ['The value of a URL metadata must be a valid HTTP or HTTPS URL.']
        })


metadata_number_validator.requires_context = True
metadata_url_validator.requires_context = True


@extend_schema_serializer(deprecate_fields=('worker_version', ))
class MetaDataSerializer(MetaDataLightSerializer):
    entity = BaseEntitySerializer(read_only=True, allow_null=True)
    entity_id = serializers.PrimaryKeyRelatedField(
        source='entity',
        queryset=Entity.objects.none(),
        write_only=True,
        required=False,
        allow_null=True,
        style={'base_template': 'input.html'},
    )
    worker_run = WorkerRunSummarySerializer(read_only=True, allow_null=True)

    class Meta(MetaDataLightSerializer.Meta):
        fields = MetaDataLightSerializer.Meta.fields + (
            'entity_id',
            'entity',
            'worker_version',
            'worker_run',
        )
        read_only_fields = MetaDataLightSerializer.Meta.read_only_fields + (
            'entity',
            'worker_version',
            'worker_run',
        )
        validators = [
            metadata_number_validator,
            metadata_url_validator,
        ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        corpus_id = None
        # When used for CreateMetaData, the API adds the current element to the context
        if self.context.get('element') is not None:
            corpus_id = self.context['element'].corpus_id
        # self.instance can be a queryset if this serializer is used for a list; just ignore anything that isn't a metadata
        elif isinstance(self.instance, MetaData):
            corpus_id = self.instance.element.corpus_id
        if corpus_id:
            self.fields['entity_id'].queryset = Entity.objects.filter(corpus_id=corpus_id)

    def validate(self, data):
        worker_run = data.get('worker_run')
        if worker_run is not None:
            data['worker_version'] = WorkerVersion(id=worker_run.version_id)
        return data


@extend_schema_serializer(deprecate_fields=('worker_version', ))
class MetaDataCreateSerializer(MetaDataSerializer):
    worker_version = serializers.UUIDField(
        required=False,
        allow_null=True,
        source='worker_version_id',
        validators=[
            ForbiddenValidator(),
        ],
        help_text=dedent("""
            ID of a WorkerVersion that created this MetaData.

            Creating new MetaData with a WorkerVersion is forbidden. Use `worker_run_id` instead.
        """),
    )
    worker_run_id = serializers.PrimaryKeyRelatedField(
        queryset=WorkerRun.objects.all(),
        required=False,
        write_only=True,
        allow_null=True,
        style={'base_template': 'input.html'},
        source='worker_run'
    )

    class Meta(MetaDataSerializer.Meta):
        fields = MetaDataSerializer.Meta.fields + (
            'worker_run_id',
        )
        read_only_fields = ()


class MetaDataBulkItemSerializer(MetaDataLightSerializer):
    """
    Serialises some Metadata for any Element
    """
    entity_id = serializers.PrimaryKeyRelatedField(
        source='entity',
        queryset=Entity.objects.none(),
        required=False,
        allow_null=True,
        style={'base_template': 'input.html'},
    )

    class Meta:
        model = MetaData
        fields = MetaDataLightSerializer.Meta.fields + ('entity_id', )
        validators = [
            metadata_number_validator,
            metadata_url_validator,
        ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.context.get('element') is not None:
            self.fields['entity_id'].queryset = Entity.objects.filter(corpus_id=self.context['element'].corpus_id)


class MetaDataBulkSerializer(serializers.Serializer):
    """
    Allow editing several MetaData
    """
    worker_version = ForbiddenField()
    worker_run_id = serializers.PrimaryKeyRelatedField(
        queryset=WorkerRun.objects.all(),
        style={'base_template': 'input.html'},
        source='worker_run',
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['metadata_list'] = MetaDataBulkItemSerializer(many=True, allow_empty=False, context=self.context)

    def make_metadata_tuples(self, meta_list):
        tuples_list = []
        for item in meta_list:
            item_tuple = (
                item.get('name'),
                item.get('type').value,
                item.get('value')
            )
            tuples_list.append(item_tuple)
        return tuples_list

    def validate(self, data):
        data = super().validate(data)
        request_metadata = self.make_metadata_tuples(data['metadata_list'])
        unique_metadata = set(request_metadata)
        if len(unique_metadata) != len(request_metadata):
            raise ValidationError('Metadata in metadata list must make a unique set.')
        names = [item['name'] for item in data['metadata_list']]
        types = [item['type'] for item in data['metadata_list']]
        filtered_metadata = MetaData.objects.filter(element=self.context['element'].id, name__in=names, type__in=types).values()
        existing_metadata = self.make_metadata_tuples(filtered_metadata)
        meta_matches = set(request_metadata) & set(existing_metadata)
        if meta_matches:
            raise ValidationError(f'Metadata {meta_matches} already exist(s) on this element.')
        return data

    def create(self, validated_data):
        base_attrs = {}
        worker_run = validated_data['worker_run']
        # Retrieve the version from the specified worker run
        # No need for an extra query here
        base_attrs.update({
            "worker_run": worker_run,
            "worker_version_id": worker_run.version_id,
        })

        validated_data['metadata_list'] = MetaData.objects.bulk_create([
            MetaData(
                **m,
                **base_attrs,
                id=uuid.uuid4(),
                element=self.context['element'],
            )
            for m in validated_data['metadata_list']
        ])

        return validated_data


class CorpusSerializer(serializers.ModelSerializer):
    """
    Serialize a corpus
    """
    rights = serializers.SerializerMethodField(read_only=True)
    types = ElementTypeLightSerializer(many=True, read_only=True)
    top_level_type = serializers.SlugRelatedField(
        queryset=ElementType.objects.none(),
        slug_field='slug',
        allow_null=True,
        default=None,
        help_text=dedent("""
            Slug of a type that will be used by default by the frontend when browsing this corpus' elements.

            When a user opens this corpus in the Arkindex frontend, the frontend will use the `top_level=true`
            and `folder=true` filters on the `ListElements` endpoint. When this attribute is set, the frontend
            will use the `type=<top_level_type>` and the `top_level=true` filters instead.
        """).strip(),
    )
    authorized_users = serializers.SerializerMethodField(
        read_only=True,
        help_text='Number of users or groups that have been granted access rights on this corpus.',
    )

    class Meta:
        model = Corpus
        fields = (
            'id',
            'name',
            'description',
            'public',
            'rights',
            'types',
            'top_level_type',
            'created',
            'authorized_users',
            'indexable',
        )
        extra_kwargs = {
            'public': {
                'help_text': 'Make the corpus visible to users without authentication. '
                             'This attribute can only be changed by instance administrators.'
            },
            'indexable': {
                'help_text': 'Whether this corpus should be indexed in Solr and searchable using the `SearchCorpus` endpoint.'
            }
        }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if isinstance(self.instance, Corpus):
            # There is an existing corpus; allow its types as top level types
            self.fields['top_level_type'].queryset = self.instance.types.all()

    @extend_schema_field(serializers.ListField(child=serializers.ChoiceField(['read', 'write', 'admin'])))
    def get_rights(self, corpus):

        if getattr(corpus, 'access_level', None):
            # Use annotated rights if available
            level = corpus.access_level
        else:
            level = get_max_level(self.context['request'].user, corpus)

        rights = ['read']
        if level >= Role.Contributor.value:
            rights.append('write')
        if level >= Role.Admin.value:
            rights.append('admin')
        return rights

    def get_authorized_users(self, corpus) -> int:
        count = getattr(corpus, 'authorized_users', None)
        if count is None:
            count = corpus.memberships.count()
        return count

    def validate_public(self, public):
        """
        Only an admin can toggle a corpus public property
        Normal users may create private corpus only
        """
        toggled = self.instance and self.instance.public != public
        if (toggled or public) and not self.context['request'].user.is_admin:
            raise ValidationError(['Only admin users are allowed to edit the public attribute on a corpus.'])
        return public

    @transaction.atomic
    def create(self, validated_data):
        corpus = Corpus.objects.create(**validated_data)
        default_groups = Group.objects.filter(use_in_new_project=True).all()
        memberships = [
            Right(group=user_group, level=Role.Admin.value, content_object=corpus)
            for user_group in default_groups
        ]
        memberships.append(Right(user=self.context['request'].user, level=Role.Admin.value, content_object=corpus))
        Right.objects.bulk_create(
            memberships
        )
        corpus.create_default_types()
        # Cheat annotating corpus to avoid a stale read when serializing it
        corpus.access_level = Role.Admin.value
        corpus.authorized_users = 1
        return corpus


class ElementTinySerializer(serializers.ModelSerializer):
    """
    Serialises a document
    """
    type = serializers.SlugRelatedField(queryset=ElementType.objects.none(), slug_field='slug')
    corpus = serializers.SerializerMethodField(read_only=True)
    thumbnail_url = serializers.SerializerMethodField(read_only=True)
    zone = serializers.SerializerMethodField(read_only=True)
    rotation_angle = serializers.IntegerField(min_value=0, max_value=359, default=0)
    mirrored = serializers.BooleanField(default=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('element'):
            # Do not raise Error in order to create OpenAPI schema
            return
        self.fields['type'].queryset = ElementType.objects.filter(corpus=self.context['element'].corpus)

    @extend_schema_field(serializers.CharField(allow_null=True))
    def get_thumbnail_url(self, element):
        if element.type.folder:
            return element.thumbnail.s3_url

    @extend_schema_field(CorpusLightSerializer(read_only=True))
    def get_corpus(self, element):
        """
        This serializer accepts a Corpus instance in its context,
        which will be used instead of the element's corpus when available.

        This is particularly helpful for element list endpoints as they can avoid
        fetching an already-known corpus repeatedly using SQL joins.
        """
        corpus = self.context.get('corpus')
        if not corpus:
            corpus = element.corpus
        return CorpusLightSerializer(corpus).data

    @extend_schema_field(ZoneSerializer(read_only=True))
    def get_zone(self, element):
        """
        Elements used to have zones, which held images and polygons.
        Those attributes are now directly on elements. While we could just use
        ZoneSerializer(source='*') in the attributes directly, this would cause
        the ZoneSerializer to always be present. Instead of having zone=None,
        we would see zone={'image': None, 'polygon': None, 'url': None}.
        This method ensures the zone attribute is None.
        """
        if not element.image_id or not element.polygon:
            return
        return ZoneSerializer(element).data

    class Meta:
        model = Element
        fields = (
            'id',
            'type',
            'name',
            'corpus',
            'thumbnail_url',
            'zone',
            'rotation_angle',
            'mirrored',
        )
        read_only_fields = (
            'id',
            'corpus',
            'thumbnail_url',
            'zone',
        )


class ElementSlimSerializer(ElementTinySerializer):
    """
    Fully serialises a document
    """
    thumbnail_put_url = serializers.SerializerMethodField(read_only=True)

    @extend_schema_field(serializers.CharField(allow_null=True))
    def get_thumbnail_put_url(self, element):
        if is_admin_or_ponos_task(self.context['request']) and element.type.folder:
            return element.thumbnail.s3_put_url

    class Meta(ElementTinySerializer.Meta):
        model = Element
        fields = ElementTinySerializer.Meta.fields + ('thumbnail_put_url',)
        read_only_fields = ElementTinySerializer.Meta.read_only_fields + ('thumbnail_put_url',)


@extend_schema_serializer(deprecate_fields=('worker_version_id', ))
class ElementListSerializer(ElementTinySerializer):
    created = serializers.DateTimeField(read_only=True)
    classes = ClassificationSerializer(
        many=True,
        default=None,
        read_only=True,
        help_text='Classifications on this element, if the `with_classes` option has been enabled.',
    )
    metadata = MetaDataLightSerializer(
        many=True,
        default=None,
        read_only=True,
        help_text='Metadata without entities on this element, if the `with_metadata` option has been enabled.',
        # Use a custom attribute here, so that the serializer does not try to load
        # all of the element's metadata if they have not been prefetched.
        source='prefetched_metadata',
    )
    has_children = serializers.BooleanField(default=None, read_only=True)
    worker_run = WorkerRunSummarySerializer(read_only=True, allow_null=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if 'with_zone' in self.context and not self.context['with_zone']:
            self.fields.pop('zone')
        if 'with_corpus' in self.context and not self.context['with_corpus']:
            self.fields.pop('corpus')

    @extend_schema_field(serializers.CharField(allow_null=True))
    def get_thumbnail_url(self, element):
        # Do not return full element zone when with_zone query parameter is false, instead just return a thumbnail
        if 'with_zone' in self.context and not self.context['with_zone'] and not element.type.folder and element.image_id:
            return element.iiif_thumbnail_url
        return super().get_thumbnail_url(element)

    class Meta(ElementTinySerializer.Meta):
        model = Element
        fields = ElementTinySerializer.Meta.fields + (
            'created',
            'classes',
            'metadata',
            'has_children',
            'worker_version_id',
            'worker_run',
            'confidence',
        )


class ElementParentSerializer(serializers.Serializer):
    """
    Serialize a relation between an element and one of its parents
    """
    child = serializers.PrimaryKeyRelatedField(queryset=Element.objects.none())
    parent = serializers.PrimaryKeyRelatedField(queryset=Element.objects.none())
    type_ordering = serializers.BooleanField(
        default=True,
        help_text='Add the child element at the last position in the parent relative to '
                  'only the elements of the same type, or to all elements.',
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):
            # Do not raise Error in order to create OpenAPI schema
            return
        corpora = Corpus.objects.writable(self.context['request'].user)
        self.fields['child'].queryset = Element.objects.using('default').filter(corpus__in=corpora)
        self.fields['parent'].queryset = Element.objects.using('default').filter(corpus__in=corpora)

    def validate(self, data):
        errors = defaultdict(list)
        data = super().validate(data)
        child = data.get('child')
        parent = data.get('parent')

        if parent.corpus_id != child.corpus_id:
            errors['parent'].append("Parent is not from corpus '{}'".format(child.corpus.name))
        if parent.id == child.id:
            errors['parent'].append('A child cannot be its own parent')
        if errors:
            raise ValidationError(errors)
        return data

    def perform_create_checks(self, child, parent):
        # Assert parent is not an child's ancestor already
        if ElementPath.objects.filter(element_id=child.id, path__contains=[parent.id]).exists():
            raise ValidationError({'parent': [
                "'{}' is already a parent of '{}'".format(parent.id, child.id)
            ]})
        # Assert parent is not an alement's descendant
        if ElementPath.objects.filter(element_id=parent.id, path__contains=[child.id]).exists():
            raise ValidationError({'parent': [
                "'{}' is a child of element '{}'".format(parent.id, child.id)
            ]})

    def create(self, validated_data):
        child = validated_data['child']
        parent = validated_data['parent']
        type_ordering = validated_data['type_ordering']
        self.perform_create_checks(child, parent)
        child.add_parent(parent, type_ordering=type_ordering)

    def delete(self, validated_data):
        child = validated_data['child']
        parent = validated_data['parent']
        parent.remove_child(child)


@extend_schema_serializer(deprecate_fields=('worker_version', ))
class ElementSerializer(ElementSlimSerializer):
    """
    Serialize an element with its metadata and classifications
    """

    created = serializers.DateTimeField(read_only=True)
    creator = serializers.CharField(
        source='creator.display_name',
        read_only=True,
        allow_null=True,
        help_text='Display name of the user who created the element, if there is one.',
    )
    rights = serializers.SerializerMethodField()
    metadata_count = serializers.IntegerField(
        help_text='Number of metadata. You can list them using `ListElementMetadata`.',
        read_only=True,
    )
    classifications = ClassificationSerializer(many=True, read_only=True)
    image = serializers.PrimaryKeyRelatedField(
        queryset=Image.objects.all().select_related('server'),
        required=False,
        write_only=True,
        help_text='Link this element to an image by UUID via a polygon. '
                  'When the image is updated, if there was an image before and the polygon is not updated, '
                  'the previous polygon is reused. Otherwise, a polygon filling the new image is used.',
        style={'base_template': 'input.html'},
    )
    polygon = LinearRingField(
        required=False,
        write_only=True,
        help_text='Set the polygon linking this element to the image. '
                  '`image` must be set when this field is set and there was no image or polygon defined before.',
    )

    worker_run = WorkerRunSummarySerializer(read_only=True, allow_null=True)

    class Meta:
        model = Element
        fields = ElementSlimSerializer.Meta.fields + (
            'created',
            'creator',
            'rights',
            'metadata_count',
            'classifications',
            'image',
            'polygon',
            'worker_version',
            'confidence',
            'worker_run'
        )
        read_only_fields = ElementSlimSerializer.Meta.read_only_fields + (
            'created',
            'creator',
            'rights',
            'metadata_count',
            'classifications',
            'worker_version',
            'worker_run'
        )

    @extend_schema_field(serializers.ListField(child=serializers.ChoiceField(['read', 'write', 'admin'])))
    def get_rights(self, element):
        user = self.context['request'].user
        level = get_max_level(user, element.corpus)

        rights = ['read']
        if level >= Role.Contributor.value:
            rights.append('write')
        # Admin access is granted to both corpus admins and element creators
        if level >= Role.Admin.value or (level >= Role.Contributor.value and element.creator_id == user.id):
            rights.append('admin')

        return rights

    def update(self, instance, validated_data):
        image = validated_data.pop('image', None)
        polygon = validated_data.pop('polygon', None)
        if polygon or image:
            if not image:
                if not instance.image_id:
                    # A polygon was set but there is no image
                    raise ValidationError({
                        'image': ['Image is required when defining a polygon on an element without an existing zone'],
                    })
                image = instance.image

            if image.width == 0 or image.height == 0:
                raise ValidationError({'image': ['This image does not have valid dimensions.']})

            if not polygon:
                if instance.polygon:
                    polygon = instance.polygon
                else:
                    polygon = LinearRing(
                        (0, 0),
                        (0, image.height),
                        (image.width, image.height),
                        (image.width, 0),
                        (0, 0)
                    )
        if polygon and image:
            if polygon_outside_image(image, polygon):
                raise ValidationError({
                    'polygon': ["An element's polygon must not exceed its image's bounds."]
                })

            validated_data.update(image=image, polygon=polygon)

        instance = super().update(instance, validated_data)
        return instance


class ElementNeighborsSerializer(serializers.Serializer):
    position = serializers.IntegerField(source='ordering')
    element = ElementLightSerializer()
    parents = serializers.ListField(
        child=ElementLightSerializer()
    )


@extend_schema_serializer(deprecate_fields=('worker_version', ))
class ElementCreateSerializer(ElementLightSerializer):
    """
    Serialize an Element with a possible parent and image
    """
    type = serializers.SlugField(source='type.slug')
    corpus = serializers.PrimaryKeyRelatedField(
        queryset=Corpus.objects.none(),
        style={'base_template': 'input.html'},
    )
    image = serializers.PrimaryKeyRelatedField(
        queryset=Image.objects.using('default').all().select_related('server'),
        required=False,
        help_text='Link this element to an image by UUID via a polygon.',
        style={'base_template': 'input.html'},
    )
    parent = serializers.PrimaryKeyRelatedField(
        queryset=Element.objects.none(),
        required=False,
        allow_null=True,
        style={'base_template': 'input.html'},
    )
    polygon = LinearRingField(
        required=False,
        help_text='Set the polygon linking this element to the image. '
                  'Either `image` must be set, or a `parent` element with an image must be set, to use a polygon. '
                  'Defaults to a rectangle taking up the whole image.',
    )
    rotation_angle = serializers.IntegerField(min_value=0, max_value=359, required=False)
    mirrored = serializers.BooleanField(required=False)
    worker_version = serializers.UUIDField(
        required=False,
        allow_null=True,
        source='worker_version_id',
        validators=[
            ForbiddenValidator(),
        ],
        help_text=dedent("""
            ID of a WorkerVersion that created this element.

            Creating new elements with a WorkerVersion is forbidden. Use `worker_run_id` instead.
        """),
    )
    worker_run_id = serializers.PrimaryKeyRelatedField(
        queryset=WorkerRun.objects.all(),
        required=False,
        allow_null=True,
        help_text='A WorkerRun ID that the new elements will refer to',
        source='worker_run',
        # Avoids thousands of SQL queries when opening the endpoint in a browser
        style={'base_template': 'input.html'}
    )

    class Meta(ElementLightSerializer.Meta):
        model = Element
        fields = ElementLightSerializer.Meta.fields + (
            'image',
            'corpus',
            'parent',
            'polygon',
            'rotation_angle',
            'mirrored',
            'worker_version',
            'worker_run_id',
            'confidence',
        )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):
            # Do not raise Error in order to create OpenAPI schema
            return
        corpora = Corpus.objects.writable(self.context['request'].user)
        self.fields['corpus'].queryset = corpora.prefetch_related('types')

        # Avoid a stale read issue when creating a child on a freshly created parent
        self.fields['parent'].queryset = Element.objects.filter(corpus__in=corpora).select_related('image').using('default')

    def validate(self, data):
        errors = defaultdict(list)
        data = super().validate(data)

        try:
            # Use types already loaded with the corpus queryset
            # instead of making another SQL query
            data['type'] = next(t for t in data['corpus'].types.all() if t.slug == data['type']['slug'])
        except StopIteration:
            # Fail immediately here because the type is required for other checks
            raise ValidationError({
                'type': ['Type {!r} not found in corpus {}'.format(data['type']['slug'], data['corpus'].name)]
            })

        parent = data.get('parent')
        image = data.get('image')
        polygon = data.get('polygon')

        if parent:
            if parent.corpus_id != data['corpus'].id:
                errors['corpus'].append(
                    'Parent and child must be in the same corpus'
                )
            # Make the element inherit rotation and mirroring from its parent
            data.setdefault('rotation_angle', parent.rotation_angle)
            data.setdefault('mirrored', parent.mirrored)
        else:
            # Turn any None values into valid default values
            data.setdefault('rotation_angle', 0)
            data.setdefault('mirrored', False)

        if polygon and not image:
            # Deduce the image from the parent if there is one
            if parent is not None and parent.image_id:
                image = parent.image
                data['image'] = image
            else:
                errors['polygon'].append('An image or a parent with an image is required to create an element with a polygon.')

        if image and (image.width == 0 or image.height == 0):
            # Element creation with images with a width or height equal to zero
            # will lead to many errors everywhere as this would create impossible polygons
            errors['image'].append('This image does not have valid dimensions.')

        if polygon and image:
            if polygon_outside_image(image, polygon):
                errors['polygon'].append("An element's polygon must not exceed its image's bounds.")

        worker_run = data.get('worker_run', None)
        if worker_run:
            if is_admin_or_ponos_task(self.context['request']):
                data['worker_version_id'] = worker_run.version_id
            else:
                errors['worker_run_id'].append('Only an instance admin or a Ponos task can create an element with a worker run.')
        else:
            # Set the element creator for a manual element
            data['creator'] = self.context['request'].user

        if errors:
            raise ValidationError(errors)

        return data

    def create(self, validated_data):
        image = validated_data.get('image')
        polygon = validated_data.get('polygon')

        if image and not polygon:
            # Create a new zone containing the image
            polygon = LinearRing(
                (0, 0),
                (0, image.height),
                (image.width, image.height),
                (image.width, 0),
                (0, 0)
            )

        element = Element.objects.create(
            corpus=validated_data['corpus'],
            type=validated_data['type'],
            name=validated_data['name'],
            worker_version_id=validated_data.get('worker_version_id'),
            worker_run=validated_data.get('worker_run'),
            creator=validated_data.get('creator'),
            image=image,
            polygon=polygon,
            rotation_angle=validated_data['rotation_angle'],
            mirrored=validated_data['mirrored'],
            confidence=validated_data.get('confidence'),
        )

        # Make relation to parent, without dealing with children as there cannot be
        # any children on a newly created element
        parent = validated_data.get('parent')
        if parent:
            element.add_parent(parent, skip_children=True)
        else:
            element.paths.create(path=[])

        return element

    def to_representation(self, instance):

        # Add a mode to only output the newly created element's id
        # This speeds up the endpoint execution time as it makes a few less SQL queries
        # and does not spend time serializing. This should be used by all workers
        if self.context['request'].query_params.get('slim_output') is not None:
            return {"id": instance.id}

        return ElementSerializer(instance, context=self.context).data


class ElementTypeSerializer(ElementTypeLightSerializer):
    """
    Serialize an element type with its corpus
    """
    corpus = serializers.PrimaryKeyRelatedField(queryset=Corpus.objects.none())

    class Meta:
        model = ElementType
        fields = ElementTypeLightSerializer.Meta.fields + ('corpus', )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):
            # Keep original fields so OpenAPI can build its schemas
            return
        self.fields['corpus'].queryset = Corpus.objects.admin(self.context['request'].user)


class ElementBulkItemSerializer(serializers.Serializer):
    """
    A single element in a bulk element creation.
    All validation is done at once in ElementBulkSerializer. Do not use this serializer alone.
    """
    name = serializers.CharField()
    type = serializers.SlugField()
    polygon = LinearRingField()
    confidence = serializers.FloatField(
        min_value=0,
        max_value=1,
        allow_null=True,
        default=None,
    )


class ElementBulkSerializer(serializers.Serializer):
    """
    Serialize child elements on a common parent
    """
    worker_version = ForbiddenField()
    worker_run_id = serializers.PrimaryKeyRelatedField(
        queryset=WorkerRun.objects.all(),
        help_text='A WorkerRun ID that the new elements will refer to',
        source='worker_run',
        # Avoids thousands of SQL queries when opening the endpoint in a browser
        style={'base_template': 'input.html'},
    )

    elements = ElementBulkItemSerializer(
        many=True,
        allow_empty=False,
        help_text='A list of elements with their name, type and polygon'
    )

    def validate(self, data):
        errors = defaultdict()
        type_slugs = set(element['type'] for element in data['elements'])
        type_ids = dict(
            ElementType
            .objects
            .filter(
                corpus_id=self.context['element'].corpus_id,
                slug__in=type_slugs)
            .values_list('slug', 'id')
        )
        missing_types = list(type_slugs - set(type_ids.keys()))
        if missing_types:
            raise ValidationError([
                f"Element types with slugs {', '.join(missing_types)} do not exist in the parent element's corpus"
            ])

        element_errors = {}
        for i, element in enumerate(data['elements']):
            # Replace slugs with type IDs
            element['type'] = type_ids[element['type']]
            # Check that polygons are not outside the image
            if polygon_outside_image(self.context['element'].image, element['polygon']):
                element_errors[i] = {'polygon': ["An element's polygon must not exceed its image's bounds."]}

        if element_errors:
            errors['elements'] = element_errors

        if errors:
            raise ValidationError(errors)

        return data


class WorkerStatisticsSerializer(serializers.Serializer):
    """
    Serialize activity statistics of a worker version
    """
    worker_version_id = serializers.UUIDField(read_only=True)
    queued = serializers.IntegerField(read_only=True)
    started = serializers.IntegerField(read_only=True)
    processed = serializers.IntegerField(read_only=True)
    error = serializers.IntegerField(read_only=True)


class ElementDestinationSerializer(serializers.Serializer):
    source = serializers.PrimaryKeyRelatedField(queryset=Element.objects.none())
    destination = serializers.PrimaryKeyRelatedField(queryset=Element.objects.none())

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):
            # Do not raise Error in order to create OpenAPI schema
            return
        corpora = Corpus.objects.writable(self.context['request'].user)
        self.fields['source'].queryset = Element.objects.filter(corpus__in=corpora).select_related('corpus')
        self.fields['destination'].queryset = Element.objects.filter(corpus__in=corpora).select_related('corpus')

    def validate(self, data):
        data = super().validate(data)
        source = data.get('source')
        destination = data.get('destination')

        if destination.id == source.id:
            raise ValidationError({'destination': ['A source element cannot be moved into itself']})
        if destination.corpus != source.corpus:
            raise ValidationError({'destination': ['A source element cannot be moved to a destination from another corpus']})
        return data

    def perform_create_checks(self, source, destination):
        # Assert destination is not a source's direct ancestor already
        if ElementPath.objects.filter(element_id=source.id, path__last=destination.id).exists():
            raise ValidationError({'destination': [
                "'{}' is already a direct parent of '{}'".format(destination.id, source.id)
            ]})
        # Assert destination is not a source's descendant
        if ElementPath.objects.filter(element_id=destination.id, path__contains=[source.id]).exists():
            raise ValidationError({'destination': [
                "'{}' is a child of element '{}'".format(destination.id, source.id)
            ]})


class SelectionMoveSerializer(serializers.Serializer, SelectionMixin):
    corpus_id = serializers.PrimaryKeyRelatedField(queryset=Corpus.objects.none())
    destination = serializers.PrimaryKeyRelatedField(queryset=Element.objects.none())

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):
            # Do not raise Error in order to create OpenAPI schema
            return
        corpora = Corpus.objects.writable(self.context['request'].user)
        self.fields['corpus_id'].queryset = corpora
        self.fields['destination'].queryset = Element.objects.filter(corpus__in=corpora).select_related('corpus')

    def validate(self, data):
        data = super().validate(data)
        corpus = data.get('corpus_id')
        destination = data.get('destination')

        if not settings.ARKINDEX_FEATURES['selection']:
            raise ValidationError('Selection is not available on this instance.')

        if not self.context['request'].user.selected_elements.filter(corpus=corpus).exists():
            raise ValidationError({'corpus_id': ['No elements from the specified corpus have been selected.']})

        if destination.corpus.id != corpus.id:
            raise ValidationError({'destination': ['An element cannot be moved to a destination from another corpus.']})
        selected_elements = self.context['request'].user.selected_elements.filter(corpus=corpus.id)
        if selected_elements.filter(id=destination.id).exists():
            raise ValidationError({'destination': ['An element cannot be moved into itself.']})

        # Assert destination is not a source's direct ancestor already
        if ElementPath.objects.filter(element__corpus_id=corpus.id, element__selections__user_id=self.context['request'].user.id, path__last=destination.id).exists():
            raise ValidationError({'destination': [
                "'{}' is already a direct parent of one or more selected elements.".format(destination.id)
            ]})
        # Assert destination is not a source's descendant
        if destination.paths.filter(path__overlap=Array(selected_elements.values_list('id', flat=True))).exists():
            raise ValidationError({'destination': [
                "'{}' is a child of one or more selected elements.".format(destination.id)
            ]})

        return data

    def save(self):
        corpus = self.validated_data['corpus_id']
        destination = self.validated_data['destination']
        move_selection(corpus_id=corpus.id, destination=destination, user_id=self.context['request'].user.id)


class SelectionParentCreateSerializer(serializers.Serializer, SelectionMixin):
    corpus_id = serializers.PrimaryKeyRelatedField(queryset=Corpus.objects.none())
    parent_id = serializers.PrimaryKeyRelatedField(queryset=Element.objects.none())

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):
            # Do not raise Error in order to create OpenAPI schema
            return
        corpora = Corpus.objects.writable(self.context['request'].user)
        self.fields['corpus_id'].queryset = corpora
        self.fields['parent_id'].queryset = Element.objects.filter(corpus__in=corpora).select_related('corpus')

    def validate(self, data):
        data = super().validate(data)
        corpus = data.get('corpus_id')
        parent = data.get('parent_id')

        if not settings.ARKINDEX_FEATURES['selection']:
            raise ValidationError('Selection is not available on this instance.')

        if not self.context['request'].user.selected_elements.filter(corpus=corpus).exists():
            raise ValidationError({'corpus_id': ['No elements from the specified corpus have been selected.']})

        if parent.corpus.id != corpus.id:
            raise ValidationError({'parent_id': ['An element cannot be added to a parent from another corpus.']})
        selected_elements = self.context['request'].user.selected_elements.filter(corpus=corpus.id)
        if selected_elements.filter(id=parent.id).exists():
            raise ValidationError({'parent_id': ['An element cannot be added as a parent of itself.']})

        # Assert parent is not a source's direct ancestor already
        if ElementPath.objects.filter(element__corpus_id=corpus.id, element__selections__user_id=self.context['request'].user.id, path__last=parent.id).exists():
            raise ValidationError({'parent_id': [
                f"'{parent.id}' is already a direct parent of one or more selected elements."
            ]})
        # Assert parent is not a source's descendant
        if parent.paths.filter(path__overlap=Array(selected_elements.values_list('id', flat=True))).exists():
            raise ValidationError({'parent_id': [
                f"'{parent.id}' is a child of one or more selected elements."
            ]})

        return data

    def save(self):
        corpus = self.validated_data['corpus_id']
        parent = self.validated_data['parent_id']
        add_parent_selection(corpus_id=corpus.id, parent=parent, user_id=self.context['request'].user.id)
