import json
import os
import sqlite3
from datetime import datetime, timezone
from textwrap import dedent
from unittest.mock import call, patch
from uuid import uuid4

from django.core import mail
from django.test import override_settings

from arkindex.documents.export import export_corpus
from arkindex.documents.models import (
    Classification,
    CorpusExportState,
    ElementPath,
    EntityLink,
    EntityType,
    MetaData,
    MetaType,
    Transcription,
    TranscriptionEntity,
)
from arkindex.images.models import Image, ImageServer
from arkindex.ponos.models import Artifact
from arkindex.process.models import Repository, WorkerType, WorkerVersion, WorkerVersionState
from arkindex.project.tests import FixtureTestCase

TABLE_NAMES = {
    'export_version',
    'classification',
    'element',
    'element_path',
    'entity',
    'entity_link',
    'entity_role',
    'entity_type',
    'image',
    'image_server',
    'metadata',
    'transcription',
    'transcription_entity',
    'worker_version',
}


class TestExport(FixtureTestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.repo = Repository.objects.get(url='http://my_repo.fake/workers/worker')
        cls.revision = cls.repo.revisions.first()
        # The fixtures have two worker versions, only one of them is used in existing elements
        cls.recognizer = cls.repo.workers.get(slug='reco').versions.get()

        # Retrieve a workerType for created workers
        cls.worker_type = WorkerType.objects.get(slug='dla')

    def _make_worker_version(self):
        return self.revision.versions.create(
            worker=self.repo.workers.create(slug=str(uuid4()), type=self.worker_type),
            configuration={},
            state=WorkerVersionState.Available,
            docker_image=Artifact.objects.first(),
        )

    @override_settings(PUBLIC_HOSTNAME='https://darkindex.lol')
    @patch('arkindex.documents.export.datetime')
    @patch('arkindex.documents.export.os.unlink')
    @patch('arkindex.project.aws.s3.Object')
    def test_export(self, s3_object_mock, unlink_mock, datetime_mock):
        datetime_mock.now.return_value = datetime(2042, 1, 1, 12, 34, 56, tzinfo=timezone.utc)

        element = self.corpus.elements.get(name='Volume 1')
        element.confidence = 0.5
        element.save()
        transcription = Transcription.objects.first()
        version = WorkerVersion.objects.get(worker__slug='reco')

        metadata_version = self._make_worker_version()
        element.metadatas.create(
            type=MetaType.Numeric,
            name='Score',
            value=1,
            worker_version=metadata_version,
        )

        element.classifications.create(
            ml_class=self.corpus.ml_classes.create(name='Blah'),
            confidence=.55555555,
        )

        location_type = EntityType.objects.get(name='location', corpus=self.corpus)
        person_type = EntityType.objects.get(name='person', corpus=self.corpus)

        entity1 = self.corpus.entities.create(
            name='Arrokuda',
            type=location_type,
            metas={'subtype': 'pokemon'},
        )
        entity2 = self.corpus.entities.create(
            name='Stonjourner',
            type=person_type,
            validated=True,
            moderator=self.superuser,
        )
        role = self.corpus.roles.create(
            parent_name='parent',
            child_name='child',
            parent_type=location_type,
            child_type=person_type,
        )
        role.links.create(parent=entity1, child=entity2)

        transcription.transcription_entities.create(
            entity=entity1,
            offset=1,
            length=1,
            worker_version=version,
        )

        transcription.transcription_entities.create(
            entity=entity2,
            offset=0,
            length=1,
            worker_version=version,
            confidence=0.42,
        )

        export = self.corpus.exports.create(user=self.user)

        export_corpus(export)

        export.refresh_from_db()
        self.assertEqual(export.state, CorpusExportState.Done)

        self.assertEqual(s3_object_mock().upload_file.call_count, 1)
        args, kwargs = s3_object_mock().upload_file.call_args
        self.assertCountEqual(kwargs.keys(), {'Callback', 'ExtraArgs'})
        self.assertTrue(callable(kwargs['Callback']))
        self.assertDictEqual(kwargs['ExtraArgs'], {
            'ContentType': 'application/vnd.sqlite3',
            'ContentDisposition': 'attachment; filename="unit-tests-20420101-123456.sqlite"'
        })
        self.assertEqual(len(args), 1)
        # Retrieve the database path from the S3 upload argument
        db_path = args[0]

        # We patched the file's removal to be able to check the DB after the job completes.
        # In some cases, unlink can be called multiple times due to tempfile's implementation of gettempdir,
        # which writes to a file then unlinks it.
        # We just ignore the call count, and assume the last call is ours.
        self.assertEqual(unlink_mock.call_args, call(db_path))

        db = sqlite3.connect(db_path)

        self.assertCountEqual(
            db.execute("SELECT name FROM sqlite_master WHERE type = 'table'").fetchall(),
            [(name, ) for name in TABLE_NAMES],
        )

        self.assertCountEqual(
            db.execute("SELECT version FROM export_version").fetchall(), [(5, )]
        )

        self.assertCountEqual(
            db.execute('SELECT id, display_name, url, max_width, max_height FROM image_server').fetchall(),
            [
                (
                    str(image_server.id),
                    image_server.display_name,
                    image_server.url,
                    image_server.max_width,
                    image_server.max_height
                )
                for image_server in ImageServer.objects.all()
            ]
        )

        self.assertCountEqual(
            db.execute('SELECT id, url, width, height, server_id FROM image').fetchall(),
            [
                (
                    str(image.id),
                    image.url,
                    image.width,
                    image.height,
                    str(image.server.id)
                )
                for image in Image.objects.all()
            ]
        )

        self.assertCountEqual(
            db.execute('SELECT id, name, slug, type, revision, repository_url FROM worker_version').fetchall(),
            [
                (
                    str(version.id),
                    version.worker.name,
                    version.worker.slug,
                    version.worker.type.slug,
                    version.revision.hash,
                    version.worker.repository.url
                ),
                (
                    str(metadata_version.id),
                    metadata_version.worker.name,
                    metadata_version.worker.slug,
                    metadata_version.worker.type.slug,
                    metadata_version.revision.hash,
                    metadata_version.worker.repository.url
                )
            ]
        )

        actual_rows = db.execute("""
            SELECT
                id,
                created,
                updated,
                name,
                type,
                rotation_angle,
                mirrored,
                confidence,
                worker_version_id,
                image_id,
                polygon
            FROM element
        """).fetchall()

        for i in range(len(actual_rows)):
            # Convert the row from a tuple to a list because we'll change it
            row = list(actual_rows[i])
            if row[-1] is not None:
                # Parse the polygons as JSON for easier comparison
                row[-1] = json.loads(row[-1])
            actual_rows[i] = row

        expected_rows = []
        for element in self.corpus.elements.all():
            row = [
                str(element.id),
                element.created.timestamp(),
                element.updated.timestamp(),
                element.name,
                element.type.slug,
                element.rotation_angle,
                int(element.mirrored),
                element.confidence
            ]

            if element.worker_version_id:
                row.append(str(element.worker_version_id))
            else:
                row.append(None)

            if element.polygon:
                row.append(str(element.image_id))
                row.append([
                    # coords returns a list of tuples of floats, we turn it into a list of lists of ints
                    [int(x), int(y)] for x, y in element.polygon.coords
                ])
            else:
                row.extend([None, None])

            expected_rows.append(row)

        self.assertCountEqual(actual_rows, expected_rows)

        self.assertCountEqual(
            db.execute("SELECT id, parent_id, child_id, ordering FROM element_path").fetchall(),
            [
                (
                    str(id),
                    str(parent_id),
                    str(child_id),
                    ordering
                )
                for id, parent_id, child_id, ordering
                in ElementPath.objects
                              .filter(element__corpus=self.corpus, path__last__isnull=False)
                              .values_list('id', 'path__last', 'element_id', 'ordering')
            ]
        )

        self.assertCountEqual(
            db.execute("SELECT id, element_id, text, confidence, orientation, worker_version_id FROM transcription").fetchall(),
            [
                (
                    str(transcription.id),
                    str(transcription.element_id),
                    transcription.text,
                    transcription.confidence,
                    transcription.orientation.value,
                    str(transcription.worker_version_id) if transcription.worker_version_id else None
                )
                for transcription in Transcription.objects.filter(element__corpus=self.corpus)
            ]
        )

        self.assertCountEqual(
            db.execute("SELECT id, element_id, class_name, state, moderator, confidence, high_confidence, worker_version_id FROM classification").fetchall(),
            [
                (
                    str(classification.id),
                    str(classification.element_id),
                    classification.ml_class.name,
                    classification.state.value,
                    classification.moderator.email if classification.moderator else None,
                    classification.confidence,
                    int(classification.high_confidence),
                    str(classification.worker_version_id) if classification.worker_version_id else None
                )
                for classification in Classification.objects.filter(element__corpus=self.corpus)
            ]
        )

        self.assertCountEqual(
            db.execute("SELECT id, element_id, name, type, value, entity_id, worker_version_id FROM metadata").fetchall(),
            [
                (
                    str(metadata.id),
                    str(metadata.element_id),
                    metadata.name,
                    metadata.type.value,
                    metadata.value,
                    str(metadata.entity_id) if metadata.entity_id else None,
                    str(metadata.worker_version_id) if metadata.worker_version_id else None
                )
                for metadata in MetaData.objects.filter(element__corpus=self.corpus)
            ]
        )

        self.assertCountEqual(
            db.execute("SELECT id, name, type_id, validated, moderator, metas, worker_version_id FROM entity").fetchall(),
            [
                (
                    str(entity.id),
                    entity.name,
                    str(entity.type.id),
                    int(entity.validated),
                    entity.moderator.email if entity.moderator else None,
                    json.dumps(entity.metas) if entity.metas else None,
                    str(entity.worker_version_id) if entity.worker_version_id else None,
                )
                for entity in self.corpus.entities.all()
            ]
        )

        self.assertCountEqual(
            db.execute("SELECT id, name, color FROM entity_type").fetchall(),
            [
                (
                    str(type.id),
                    type.name,
                    type.color
                )
                for type in self.corpus.entity_types.all()
            ]
        )

        self.assertCountEqual(
            db.execute("SELECT id, parent_name, child_name, parent_type_id, child_type_id FROM entity_role").fetchall(),
            [
                (
                    str(role.id),
                    role.parent_name,
                    role.child_name,
                    str(role.parent_type.id),
                    str(role.child_type.id),
                )
                for role in self.corpus.roles.all()
            ]
        )

        self.assertCountEqual(
            db.execute("SELECT id, parent_id, child_id, role_id FROM entity_link").fetchall(),
            [
                (
                    str(link.id),
                    str(link.parent_id),
                    str(link.child_id),
                    str(link.role_id)
                )
                for link in EntityLink.objects.filter(role__corpus=self.corpus)
            ]
        )

        self.assertCountEqual(
            db.execute("SELECT id, transcription_id, entity_id, offset, length, worker_version_id, confidence FROM transcription_entity").fetchall(),
            [
                (
                    str(transcription_entity.id),
                    str(transcription_entity.transcription_id),
                    str(transcription_entity.entity_id),
                    transcription_entity.offset,
                    transcription_entity.length,
                    str(transcription_entity.worker_version_id) if transcription_entity.worker_version_id else None,
                    transcription_entity.confidence,
                )
                for transcription_entity in TranscriptionEntity.objects.filter(entity__corpus=self.corpus)
            ]
        )

        expected_indexes = [
            # foreign_key columns: id, seq, table, from, to, on_update, on_delete, match
            # Cannot use a sqlite3.Row here because `foreign_key.from` would be a SyntaxError in Python
            (table_name, f'{table_name}_{foreign_key[3]}')
            for table_name in TABLE_NAMES
            for foreign_key in db.execute(f'PRAGMA foreign_key_list({table_name})').fetchall()
        ]

        self.assertCountEqual(
            db.execute("SELECT tbl_name, name FROM sqlite_master WHERE type = 'index' AND name NOT LIKE 'sqlite\\_autoindex\\_%' ESCAPE '\\'").fetchall(),
            expected_indexes
        )

        os.unlink(db_path)

        # Download URL is sent by email
        self.assertEqual(len(mail.outbox), 1)
        message = mail.outbox[0]
        self.assertEqual(message.subject, 'Arkindex project export completed')
        self.assertListEqual(message.to, [self.user.email])
        self.assertEqual(message.body, dedent(f"""
        Hello Test user,

        Your export for the Unit Tests project is completed.

        You can download it at:
        https://darkindex.lol/api/v1/export/{export.id}/

        --
        Arkindex

        """))

    @patch('arkindex.project.aws.s3.Object')
    def test_export_fail(self, s3_object_mock):
        s3_object_mock.return_value.upload_file.side_effect = ZeroDivisionError('i ran out of elbow oil')

        export = self.corpus.exports.create(user=self.user)
        with self.assertRaises(ZeroDivisionError):
            export_corpus(export)

        export.refresh_from_db()
        self.assertEqual(export.state, CorpusExportState.Failed)
        self.assertEqual(len(mail.outbox), 1)
        message = mail.outbox[0]
        self.assertEqual(message.subject, 'Arkindex project export failed')
        self.assertListEqual(message.to, [self.user.email])
        self.assertEqual(message.body, dedent("""
        Hello Test user,

        An error occurred while exporting the Unit Tests project.

        Error: i ran out of elbow oil

        Please try again or contact your system administrator.

        --
        Arkindex

        """))
