from django.core.exceptions import ObjectDoesNotExist
from drf_spectacular.authentication import TokenScheme
from drf_spectacular.contrib.rest_framework_simplejwt import SimpleJWTScheme
from rest_framework.authentication import TokenAuthentication
from rest_framework.exceptions import AuthenticationFailed

from arkindex.ponos.models import Agent, Task
from rest_framework_simplejwt.authentication import JWTAuthentication
from rest_framework_simplejwt.exceptions import InvalidToken
from rest_framework_simplejwt.settings import api_settings


class AgentUser(Agent):
    """
    A proxy model to implement the Django User interface on a Ponos agent.
    Allows Django REST Framework's usual permissions (like IsAuthenticated) to work.
    """

    is_staff: bool = False
    is_superuser: bool = False
    is_authenticated: bool = True
    is_anonymous: bool = False
    is_admin: bool = False
    is_active: bool = True
    is_agent: bool = True
    verified_email: bool = False

    @property
    def username(self) -> str:
        return str(self.id)

    def get_username(self) -> str:
        return self.username

    class Meta:
        proxy = True

    def save(self, *args, update_fields=None, **kwargs):
        """
        django.contrib.auth adds a 'user_logged_in' signal and a 'update_last_login' receiver,
        which tries to update the User.last_login field.
        The receiver is only added when the User model has a last_login field, which depends on
        the Django app integrating Ponos and not on Ponos itself, so we have to handle it.
        The receiver calls .save(update_fields=['last_login']), so we remove it.
        An empty `update_fields` will cause a save to abort silently.
        """
        if update_fields is not None and "last_login" in update_fields:
            update_fields = set(update_fields) - {"last_login"}
        return super().save(*args, update_fields=update_fields, **kwargs)


class AgentAuthentication(JWTAuthentication):
    """
    Allows authenticating as a Ponos agent using a JSON Web Token.
    """

    def get_user(self, validated_token):
        if api_settings.USER_ID_CLAIM not in validated_token:
            raise InvalidToken("Token does not hold agent information")
        try:
            return AgentUser.objects.get(id=validated_token[api_settings.USER_ID_CLAIM])
        except AgentUser.DoesNotExist:
            raise AuthenticationFailed("Agent not found")


class AgentAuthenticationExtension(SimpleJWTScheme):
    """
    drf_spectacular extension to make it recognize the AgentAuthentication.
    """

    target_class = "arkindex.ponos.authentication.AgentAuthentication"
    name = "agentAuth"


class TaskAuthentication(TokenAuthentication):
    keyword = 'Ponos'
    model = Task

    def authenticate_credentials(self, key):
        try:
            task = Task.objects.select_related('workflow__process__creator').get(token=key)
        except Task.DoesNotExist:
            # Same error message as the standard TokenAuthentication
            raise AuthenticationFailed('Invalid token.')

        # There is no Workflow.process_id, since the FK is on Process.workflow_id,
        # and accessing Workflow.process when there is no process causes an exception
        # instead of returning None.
        try:
            process = task.workflow.process
        except ObjectDoesNotExist:
            raise AuthenticationFailed('Task has no process.')

        if not process.creator_id or not process.creator.is_active:
            # Same error message as the standard TokenAuthentication
            raise AuthenticationFailed('User inactive or deleted.')

        # Must return a 2-tuple that will be set as (self.request.user, self.request.auth)
        return (process.creator, task)


class TaskAuthenticationExtension(TokenScheme):
    target_class = "arkindex.ponos.authentication.TaskAuthentication"
    name = "taskAuth"
    # The TokenScheme has a priority of -1 and matches both TokenAuthentication and its subclasses;
    # we set the priority to a higher number to make this extension match first, and disable
    # subclass matching so that this only applies to the TaskAuthentication.
    priority = 0
    match_subclasses = False
