from rest_framework.permissions import SAFE_METHODS

from arkindex.ponos.models import Task
from arkindex.process.models import Process
from arkindex.project.mixins import CorpusACLMixin, ProcessACLMixin
from arkindex.project.permissions import IsAuthenticated
from arkindex.users.models import Role


def require_agent_or_admin(request, view):
    return getattr(request.user, 'is_admin', False) or getattr(request.user, 'is_agent', False)


def require_task(request, view):
    return getattr(request.user, 'is_admin', False) or isinstance(request.auth, Task)


def require_agent_or_task(request, view):
    return (
        getattr(request.user, 'is_agent', False)
        or require_task(request, view)
    )


class IsAgent(IsAuthenticated):
    """
    Only allow Ponos agents and admins.
    """
    checks = IsAuthenticated.checks + (require_agent_or_admin, )


class IsAgentOrReadOnly(IsAgent):
    """
    Restricts write access to Ponos agents and admins,
    leave read access to everyone without any authentication.
    """
    allow_safe_methods = True


class IsAssignedAgentOrReadOnly(IsAgentOrReadOnly):
    """
    Restricts write access to Ponos agents and admins,
    and restricts write access on tasks to admins and agents assigned to them.
    Leave read access to everyone without any authentication.
    """

    def has_object_permission(self, request, view, obj) -> bool:
        if (
            isinstance(obj, Task)
            and not getattr(request.user, 'is_admin', False)
            and not request.user == obj.agent
        ):
            return request.method in SAFE_METHODS
        return super().has_object_permission(request, view, obj)


class IsAgentOrTaskAdmin(CorpusACLMixin, IsAuthenticated):
    """
    Permission to access a task with high privilege

    Allowed for admins, agents, creators of the process whose workflow contains the task,
    and users with an admin right on the process' corpus.
    """

    def has_object_permission(self, request, view, task):
        # Add request to attributes for the ACL mixin to work with self.user
        self.request = request

        return (
            require_agent_or_admin(request, view)
            or (
                task.workflow.process is not None
                and task.workflow.process.corpus_id is not None
                and self.has_admin_access(task.workflow.process.corpus)
            )
        )


class IsAgentOrTaskAdminOrReadOnly(ProcessACLMixin, IsAuthenticated):
    """
    Instance admins, agents, process admins, and the task itself are always allowed.
    For GET/HEAD, only a Guest level on the process is required for regular users.
    """

    def has_object_permission(self, request, view, task):
        # Allow agents and instance admins
        if require_agent_or_admin(request, view):
            return True

        # Allow a task to access itself
        if isinstance(request.auth, Task) and request.auth.id == task.id:
            return True

        # Add request to attributes for the ACL mixin to work with self.user
        self.request = request
        try:
            level = self.process_access_level(task.workflow.process)
        except Process.DoesNotExist:
            # Reject if the task has no process
            return False

        # Require *some* access to the process
        if level is None:
            return False

        # Require only a guest access for GET/HEAD
        if request.method in SAFE_METHODS:
            return level >= Role.Guest.value

        # Require admin access for other methods
        return level >= Role.Admin.value


class IsAgentOrArtifactAdmin(IsAgentOrTaskAdmin):
    """
    Permission to access an artifact with high privilege, based on
    access to the artifact's task through IsAgentOrTaskAdmin.
    """

    def has_object_permission(self, request, view, artifact):
        return super().has_object_permission(request, view, artifact.task)


class IsTask(IsAuthenticated):
    checks = (require_task, )


class IsAgentOrTask(IsAuthenticated):
    """
    Allow access to Ponos agents or tasks.
    """
    checks = (require_agent_or_task, )
