import base64
import hashlib
import random
import uuid
from io import BytesIO
from pathlib import Path
from unittest.mock import PropertyMock, call, patch

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.hashes import SHA256
from cryptography.hazmat.primitives.kdf.hkdf import HKDF
from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat, load_pem_private_key
from django.conf import settings
from django.test import override_settings
from django.urls import reverse
from django.utils import timezone
from rest_framework import status

from arkindex.documents.models import Corpus
from arkindex.ponos.api import timezone as api_tz
from arkindex.ponos.authentication import AgentUser
from arkindex.ponos.models import FINAL_STATES, GPU, Agent, Farm, Secret, State, Task, Workflow, encrypt
from arkindex.process.models import Process, ProcessMode
from arkindex.project.tests import FixtureAPITestCase
from arkindex.project.tools import build_public_key
from rest_framework_simplejwt.tokens import AccessToken, RefreshToken

RECIPE = """
env:
  top_env_variable: workflow_variable
  test: test_workflow
tasks:
  second-task:
    image: alpine
    command: echo hello
    parents:
    - first-task
    tags:
    - some-tag
  first-task:
    image: hello-world
"""
PONOS_PRIVATE_KEY = Path(__file__).absolute().parent / 'fixtures' / 'ponos.key'


# Helper to format a datetime as output by DRF
def str_date(d):
    return d.isoformat().replace("+00:00", "Z")


@override_settings(PONOS_LOG_TAIL=42)
class TestAPI(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.farm = Farm.objects.get()
        pubkey = build_public_key()
        cls.agent = AgentUser.objects.create(
            id=uuid.UUID(hashlib.md5(pubkey.encode("utf-8")).hexdigest()),
            farm=cls.farm,
            hostname="ghostname",
            cpu_cores=2,
            cpu_frequency=1e9,
            public_key=pubkey,
            ram_total=2e9,
            last_ping=timezone.now(),
        )
        cls.workflow = Workflow.objects.create(farm=cls.farm, recipe=RECIPE)
        cls.workflow.start()
        cls.task1, cls.task2 = cls.workflow.tasks.all()
        cls.gpu1 = cls.agent.gpus.create(
            id="108c6524-c63a-4811-bbed-9723d32a0688",
            name="GPU1",
            index=0,
            ram_total=2 * 1024 * 1024 * 1024,
        )
        cls.gpu2 = cls.agent.gpus.create(
            id="f30d1407-92bb-484b-84b0-0b8bae41ca91",
            name="GPU2",
            index=1,
            ram_total=8 * 1024 * 1024 * 1024,
        )

    def _build_workflow_response(self, response, **kwargs):
        """
        Return the serialization of the test workflow.
        Some parameters may be updated with kwargs.
        """
        self.task1.refresh_from_db()
        self.task2.refresh_from_db()
        self.workflow.refresh_from_db()
        data = {
            "id": str(self.workflow.id),
            "created": self.workflow.created.strftime("%G-%m-%dT%H:%M:%S.%fZ"),
            "finished": self.workflow.finished.strftime("%G-%m-%dT%H:%M:%S.%fZ")
            if self.workflow.finished
            else None,
            "state": self.workflow.state.value,
            "farm": {
                "id": str(self.workflow.farm.id),
                "name": self.workflow.farm.name,
            },
            "tasks": [
                {
                    "id": str(self.task1.id),
                    "run": 0,
                    "depth": 0,
                    "slug": "first-task",
                    "state": self.task1.state.value,
                    "parents": [],
                    "tags": [],
                    "shm_size": self.task1.shm_size,
                    "url": response.wsgi_request.build_absolute_uri(
                        reverse("api:task-details", args=[self.task1.id])
                    ),
                },
                {
                    "id": str(self.task2.id),
                    "run": 0,
                    "depth": 1,
                    "slug": "second-task",
                    "state": self.task2.state.value,
                    "parents": [str(self.task1.id)],
                    "tags": ["some-tag"],
                    "shm_size": self.task2.shm_size,
                    "url": response.wsgi_request.build_absolute_uri(
                        reverse("api:task-details", args=[self.task2.id])
                    ),
                },
            ],
        }
        data.update(kwargs)
        return data

    def test_workflow_details(self):
        self.client.force_login(self.superuser)
        with self.assertNumQueries(6):
            resp = self.client.get(
                reverse("api:workflow-details", args=[self.workflow.id]),
            )
            self.assertEqual(resp.status_code, status.HTTP_200_OK)
        data = resp.json()

        self.assertDictEqual(
            data, self._build_workflow_response(resp, state="unscheduled")
        )

    def test_partial_update_workflow_stopping(self):
        self.task1.state = State.Pending
        self.task1.save()
        self.client.force_login(self.superuser)
        resp = self.client.patch(
            reverse("api:workflow-details", kwargs={"pk": self.workflow.id}),
            {"state": State.Stopping.value},
        )
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        data = resp.json()

        self.assertDictEqual(
            data, self._build_workflow_response(resp, state="stopping")
        )
        self.assertEqual(self.task1.state.value, "stopping")
        self.assertEqual(self.task2.state.value, "stopped")

    def test_partial_update_workflow_stopping_finished(self):
        self.task1.state = State.Completed
        self.task1.save()
        self.task2.state = State.Completed
        self.task2.save()
        self.client.force_login(self.superuser)
        resp = self.client.patch(
            reverse("api:workflow-details", kwargs={"pk": self.workflow.id}),
            {"state": State.Stopping.value},
        )
        self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
        data = resp.json()
        self.assertEqual(data, {"state": ["Cannot stop a completed workflow"]})

    def test_partial_update_workflow_forbidden_fields(self):
        """
        Only workflow state can be updated
        """
        new_id = uuid.uuid4()
        new_farm_id = uuid.uuid4()
        self.client.force_login(self.superuser)
        resp = self.client.patch(
            reverse("api:workflow-details", kwargs={"pk": self.workflow.id}),
            {
                "id": new_id,
                "recipe": {},
                "updated": "2000-01-00T00:00:00.000000Z",
                "finished": "2000-01-00T00:00:00.000000Z",
                "tasks": [],
                "farm": new_farm_id,
            },
            format="json",
        )
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertDictEqual(resp.json(), self._build_workflow_response(resp))
        self.workflow.refresh_from_db()
        self.assertEqual(self.workflow.tasks.count(), 2)
        self.assertNotEqual(self.workflow.id, new_id)
        self.assertNotEqual(self.workflow.farm_id, new_farm_id)
        self.assertNotEqual(self.workflow.recipe, {})
        self.assertNotEqual(self.workflow.updated.isoformat(), '2000-01-01T00:00:00+00:00')
        self.assertNotEqual(self.workflow.finished.isoformat(), '2000-01-01T00:00:00+00:00')

    def test_partial_update_workflow_only_stopping(self):
        self.client.force_login(self.superuser)
        resp = self.client.patch(
            reverse("api:workflow-details", kwargs={"pk": self.workflow.id}),
            {"state": State.Completed.value},
        )
        self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
        data = resp.json()
        self.assertEqual(data, {"state": ["Can only change the state to 'stopping'"]})

    def test_update_workflow_stopping(self):
        self.task1.state = State.Pending
        self.task1.save()
        self.client.force_login(self.superuser)
        resp = self.client.put(
            reverse("api:workflow-details", kwargs={"pk": self.workflow.id}),
            {"state": State.Stopping.value},
        )
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        data = resp.json()

        self.assertDictEqual(
            data, self._build_workflow_response(resp, state="stopping")
        )
        self.assertEqual(self.task1.state.value, "stopping")
        self.assertEqual(self.task2.state.value, "stopped")

    def test_update_workflow_stopping_finished(self):
        self.task1.state = State.Completed
        self.task1.save()
        self.task2.state = State.Completed
        self.task2.save()
        self.client.force_login(self.superuser)
        resp = self.client.put(
            reverse("api:workflow-details", kwargs={"pk": self.workflow.id}),
            {"state": State.Stopping.value},
        )
        self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
        data = resp.json()
        self.assertEqual(data, {"state": ["Cannot stop a completed workflow"]})

    def test_update_workflow_forbidden_fields(self):
        """
        Only workflow state can be updated
        """
        self.task1.state = State.Pending
        self.task1.save()

        new_id = uuid.uuid4()
        new_farm_id = uuid.uuid4()
        self.client.force_login(self.superuser)
        resp = self.client.put(
            reverse("api:workflow-details", kwargs={"pk": self.workflow.id}),
            {
                "id": new_id,
                "recipe": {},
                "state": "stopping",
                "updated": "2000-01-00T00:00:00.000000Z",
                "finished": "2000-01-00T00:00:00.000000Z",
                "tasks": [],
                "farm": new_farm_id,
            },
            format="json",
        )
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertDictEqual(resp.json(), self._build_workflow_response(resp))

        self.workflow.refresh_from_db()
        self.assertEqual(self.workflow.tasks.count(), 2)
        self.assertNotEqual(self.workflow.id, new_id)
        self.assertNotEqual(self.workflow.farm_id, new_farm_id)
        self.assertNotEqual(self.workflow.recipe, {})
        self.assertNotEqual(self.workflow.updated.isoformat(), '2000-01-01T00:00:00+00:00')
        self.assertIsNone(self.workflow.finished)

        # Only the state was updated to Stopping
        self.task1.refresh_from_db()
        self.assertEqual(self.task1.state, State.Stopping)
        self.task2.refresh_from_db()
        self.assertEqual(self.task2.state, State.Stopped)

    def test_update_workflow_only_stopping(self):
        self.client.force_login(self.superuser)
        resp = self.client.put(
            reverse("api:workflow-details", kwargs={"pk": self.workflow.id}),
            {"state": State.Completed.value},
        )
        self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
        data = resp.json()
        self.assertEqual(data, {"state": ["Can only change the state to 'stopping'"]})

    @patch("arkindex.ponos.aws.s3")
    @patch("arkindex.ponos.models.s3")
    def test_task_details(self, s3_mock, aws_s3_mock):
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.return_value = {
            "Body": BytesIO(b"Failed successfully")
        }
        aws_s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"

        resp = self.client.get(reverse("api:task-details", args=[self.task1.id]))
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        data = resp.json()

        self.assertDictEqual(
            data,
            {
                "id": str(self.task1.id),
                "run": 0,
                "depth": 0,
                "slug": "first-task",
                "state": "unscheduled",
                "parents": [],
                "tags": [],
                "logs": "Failed successfully",
                "url": resp.wsgi_request.build_absolute_uri(
                    reverse("api:task-details", args=[self.task1.id])
                ),
                "full_log": "http://somewhere",
                "extra_files": {},
                "agent": None,
                "gpu": None,
                "shm_size": None,
            },
        )

        self.assertEqual(s3_mock.Object.call_count, 1)
        self.assertEqual(s3_mock.Object().get.call_count, 1)
        self.assertEqual(s3_mock.Object().get.call_args, call(Range="bytes=-42"))
        self.assertEqual(aws_s3_mock.meta.client.generate_presigned_url.call_count, 1)
        self.assertEqual(
            aws_s3_mock.meta.client.generate_presigned_url.call_args,
            call("get_object", Params={"Bucket": "ponos", "Key": "somelog"}),
        )

    def test_task_definition(self):
        with self.assertNumQueries(3):
            response = self.client.get(
                reverse("api:task-definition", args=[self.task1.id]),
                HTTP_AUTHORIZATION=f"Bearer {self.agent.token.access_token}",
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            'id': str(self.task1.id),
            'slug': 'first-task',
            'parents': [],
            'image': 'hello-world',
            'command': None,
            'shm_size': None,
            'env': {
                'test': 'test_workflow',
                'top_env_variable': 'workflow_variable',
                'ARKINDEX_TASK_TOKEN': self.task1.token,
            },
            'has_docker_socket': False,
            'image_artifact_url': None,
            'agent_id': None,
            'gpu_id': None,
            'workflow_id': str(self.workflow.id),
            'extra_files': {},
            's3_logs_put_url': None,
        })

    def test_task_definition_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.get(reverse("api:task-definition", args=[self.task1.id]))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_task_definition_requires_agent_or_admin(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.get(reverse("api:task-definition", args=[self.task1.id]))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_task_definition_image_artifact(self):
        self.task1.image_artifact = self.task2.artifacts.create(
            path='path/to/image.tar',
            size=1234567,
            content_type='application/x-tar',
        )
        self.task1.save()

        with self.assertNumQueries(4):
            response = self.client.get(
                reverse("api:task-definition", args=[self.task1.id]),
                HTTP_AUTHORIZATION=f"Bearer {self.agent.token.access_token}",
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            'id': str(self.task1.id),
            'slug': 'first-task',
            'parents': [],
            'image': 'hello-world',
            'command': None,
            'shm_size': None,
            'env': {
                'test': 'test_workflow',
                'top_env_variable': 'workflow_variable',
                'ARKINDEX_TASK_TOKEN': self.task1.token,
            },
            'has_docker_socket': False,
            'image_artifact_url': 'http://testserver' + reverse(
                'api:task-artifact-download',
                kwargs={
                    'pk': str(self.task2.id),
                    'path': 'path/to/image.tar',
                },
            ),
            'agent_id': None,
            'gpu_id': None,
            'workflow_id': str(self.workflow.id),
            'extra_files': {},
            's3_logs_put_url': None,
        })

    def test_task_definition_shm_size(self):
        self.task1.shm_size = '128g'
        self.task1.save()

        with self.assertNumQueries(3):
            response = self.client.get(
                reverse("api:task-definition", args=[self.task1.id]),
                HTTP_AUTHORIZATION=f"Bearer {self.agent.token.access_token}",
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            'id': str(self.task1.id),
            'slug': 'first-task',
            'parents': [],
            'image': 'hello-world',
            'command': None,
            'shm_size': '128g',
            'env': {
                'test': 'test_workflow',
                'top_env_variable': 'workflow_variable',
                'ARKINDEX_TASK_TOKEN': self.task1.token,
            },
            'has_docker_socket': False,
            'image_artifact_url': None,
            'agent_id': None,
            'gpu_id': None,
            'workflow_id': str(self.workflow.id),
            'extra_files': {},
            's3_logs_put_url': None,
        })

    @patch("arkindex.ponos.aws.s3")
    @patch("arkindex.ponos.models.s3")
    @patch("arkindex.ponos.serializers.timezone")
    def test_task_final_updates_workflow_finished(
        self, timezone_mock, s3_mock, aws_s3_mock
    ):
        """
        Updating a task to a final state should update the workflow's completion date
        if it causes the whole workflow to be finished
        """
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.return_value = {
            "Body": BytesIO(b"Failed successfully")
        }
        aws_s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"

        expected_datetime = timezone.datetime(3000, 1, 1, 12).astimezone()
        timezone_mock.now.return_value = expected_datetime

        self.task1.agent = self.agent
        self.task1.state = State.Completed
        self.task1.save()
        self.task2.agent = self.agent
        self.task2.state = State.Running
        self.task2.save()

        for state in FINAL_STATES:
            with self.subTest(state=state):
                self.workflow.finished = None
                self.workflow.save()

                resp = self.client.patch(
                    reverse("api:task-details", args=[self.task2.id]),
                    data={"state": state.value},
                    HTTP_AUTHORIZATION=f"Bearer {self.agent.token.access_token}",
                )
                self.assertEqual(resp.status_code, status.HTTP_200_OK)

                self.task2.refresh_from_db()
                self.assertEqual(self.task2.state, state)

                self.workflow.refresh_from_db()
                self.assertEqual(self.workflow.finished, expected_datetime)

    @patch("arkindex.ponos.signals.task_failure.send_robust")
    @patch("arkindex.ponos.aws.s3")
    @patch("arkindex.ponos.models.s3")
    def test_task_final_not_completed_triggers_signal(
        self, s3_mock, aws_s3_mock, send_mock
    ):
        """
        Updating a task to a final state that is not `State.Completed`
        should trigger the task_failure signal
        """
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.return_value = {
            "Body": BytesIO(b"Failed successfully")
        }
        aws_s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"

        self.task1.agent = self.agent
        self.task1.save()

        states_expecting_signal = {State.Error, State.Failed, State.Stopped}

        for state in State:
            with self.subTest(state=state):
                resp = self.client.patch(
                    reverse("api:task-details", args=[self.task1.id]),
                    data={"state": state.value},
                    HTTP_AUTHORIZATION=f"Bearer {self.agent.token.access_token}",
                )
                self.assertEqual(resp.status_code, status.HTTP_200_OK)

                if state in states_expecting_signal:
                    self.assertEqual(send_mock.call_count, 1)
                    _, kwargs = send_mock.call_args
                    self.assertDictEqual(kwargs, {"task": self.task1})
                else:
                    self.assertFalse(send_mock.called)

                send_mock.reset_mock()

    @patch("django.contrib.auth")
    @patch("arkindex.ponos.aws.s3")
    @patch("arkindex.ponos.models.s3")
    def test_task_update_gpu(self, s3_mock, aws_s3_mock, auth_mock):
        auth_mock.signals.user_logged_in = None
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.return_value = {
            "Body": BytesIO(b"Some nice GPU task output...")
        }
        aws_s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"

        # All GPUs are available
        self.assertFalse(GPU.objects.filter(task__isnull=False).exists())

        # The task is running on a GPU
        self.task1.gpu = self.agent.gpus.first()
        self.task1.agent = self.agent
        self.task1.state = State.Running
        self.task1.save()

        # One GPU is assigned
        self.assertTrue(GPU.objects.filter(task__isnull=False).exists())

        resp = self.client.get(reverse("api:task-details", args=[self.task1.id]))
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        data = resp.json()

        # Last ping is updated in the API call
        self.agent.refresh_from_db()

        self.assertDictEqual(
            data,
            {
                "id": str(self.task1.id),
                "run": 0,
                "depth": 0,
                "extra_files": {},
                "slug": "first-task",
                "state": "running",
                "parents": [],
                "shm_size": None,
                "tags": [],
                "logs": "Some nice GPU task output...",
                "url": resp.wsgi_request.build_absolute_uri(
                    reverse("api:task-details", args=[self.task1.id])
                ),
                "full_log": "http://somewhere",
                "agent": {
                    "cpu_cores": 2,
                    "cpu_frequency": 1000000000,
                    "cpu_load": None,
                    "farm": {"id": str(self.agent.farm_id), "name": "Wheat farm"},
                    "gpus": [
                        {
                            "id": "108c6524-c63a-4811-bbed-9723d32a0688",
                            "index": 0,
                            "name": "GPU1",
                            "ram_total": 2147483648,
                        },
                        {
                            "id": "f30d1407-92bb-484b-84b0-0b8bae41ca91",
                            "index": 1,
                            "name": "GPU2",
                            "ram_total": 8589934592,
                        },
                    ],
                    "hostname": "ghostname",
                    "id": str(self.agent.id),
                    "last_ping": str_date(self.agent.last_ping),
                    "ram_load": None,
                    "ram_total": 2000000000,
                },
                "gpu": {
                    "id": "108c6524-c63a-4811-bbed-9723d32a0688",
                    "index": 0,
                    "name": "GPU1",
                    "ram_total": 2147483648,
                },
            },
        )

        # Now let's complete that task, and check that the GPU is not assigned anymore
        resp = self.client.patch(
            reverse("api:task-details", args=[self.task1.id]),
            data={"state": "completed"},
            HTTP_AUTHORIZATION=f'Bearer {self.agent.token.access_token}',
        )
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        data = resp.json()

        self.assertDictEqual(
            data,
            {
                "id": str(self.task1.id),
                "run": 0,
                "depth": 0,
                "extra_files": {},
                "slug": "first-task",
                "state": "completed",
                "parents": [],
                "shm_size": None,
                "tags": [],
                "logs": "",
                "url": resp.wsgi_request.build_absolute_uri(
                    reverse("api:task-details", args=[self.task1.id])
                ),
                "full_log": "http://somewhere",
                "agent": {
                    "cpu_cores": 2,
                    "cpu_frequency": 1000000000,
                    "cpu_load": None,
                    "farm": {"id": str(self.agent.farm_id), "name": "Wheat farm"},
                    "gpus": [
                        {
                            "id": "108c6524-c63a-4811-bbed-9723d32a0688",
                            "index": 0,
                            "name": "GPU1",
                            "ram_total": 2147483648,
                        },
                        {
                            "id": "f30d1407-92bb-484b-84b0-0b8bae41ca91",
                            "index": 1,
                            "name": "GPU2",
                            "ram_total": 8589934592,
                        },
                    ],
                    "hostname": "ghostname",
                    "id": str(self.agent.id),
                    "last_ping": str_date(self.agent.last_ping),
                    "ram_load": None,
                    "ram_total": 2000000000,
                },
                # GPU is now un-assigned
                "gpu": None,
            },
        )

        # All GPUs are available
        self.assertFalse(GPU.objects.filter(task__isnull=False).exists())

    @patch('arkindex.ponos.models.Task.short_logs', new_callable=PropertyMock)
    def test_update_task_from_agent(self, short_logs_mock):
        short_logs_mock.return_value = ""
        self.assertEqual(self.task1.state, State.Unscheduled)
        self.task1.agent = self.agent
        self.task1.save()

        with self.assertNumQueries(8):
            resp = self.client.put(
                reverse("api:task-details", args=[self.task1.id]),
                data={"state": State.Running.value},
                HTTP_AUTHORIZATION=f"Bearer {self.agent.token.access_token}",
            )
            self.assertEqual(resp.status_code, status.HTTP_200_OK)

        data = resp.json()
        self.assertEqual(data["id"], str(self.task1.id))
        self.assertEqual(data["state"], State.Running.value)
        self.task1.refresh_from_db()
        self.assertEqual(self.task1.state, State.Running)

    def test_update_running_task_state_stopping(self):
        self.task1.state = State.Running
        self.task1.save()
        self.client.force_login(self.superuser)

        with self.assertNumQueries(4):
            resp = self.client.put(
                reverse("api:task-update", args=[self.task1.id]),
                data={"state": State.Stopping.value},
            )
            self.assertEqual(resp.status_code, status.HTTP_200_OK)

        self.assertDictEqual(resp.json(), {
            'id': str(self.task1.id),
            'state': State.Stopping.value,
        })
        self.task1.refresh_from_db()
        self.assertEqual(self.task1.state, State.Stopping)

    def test_update_non_running_task_state_stopping(self):
        states = list(State)
        states.remove(State.Running)
        self.client.force_login(self.superuser)

        for state in states:
            with self.subTest(state=state):
                self.task1.state = state
                self.task1.save()

                with self.assertNumQueries(3):
                    resp = self.client.put(
                        reverse("api:task-update", args=[self.task1.id]),
                        data={"state": State.Stopping.value},
                    )
                    self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertListEqual(resp.json(), ["You can only stop a 'Running' task"])

                self.task1.refresh_from_db()
                self.assertEqual(self.task1.state, state)

    def test_update_final_task_state_pending(self):
        self.task1.state = State.Completed
        self.task1.agent = self.agent
        self.task1.save()
        self.client.force_login(self.superuser)

        with self.assertNumQueries(4):
            resp = self.client.put(
                reverse("api:task-update", args=[self.task1.id]),
                data={"state": State.Pending.value},
            )
            self.assertEqual(resp.status_code, status.HTTP_200_OK)

        self.task1.refresh_from_db()
        self.assertEqual(self.task1.state, State.Pending)
        self.assertIsNone(self.task1.agent)

    def test_update_non_final_task_state_pending(self):
        states = set(State) - set(FINAL_STATES)
        self.client.force_login(self.superuser)

        for state in states:
            with self.subTest(state=state):
                self.task1.state = state
                self.task1.save()

                with self.assertNumQueries(3):
                    resp = self.client.put(
                        reverse("api:task-update", args=[self.task1.id]),
                        data={"state": State.Pending.value},
                    )
                    self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)

                self.assertListEqual(
                    resp.json(),
                    ["You can only restart a task with a state equal to 'Completed', 'Failed', 'Error' or 'Stopped'."],
                )
                self.task1.refresh_from_db()
                self.assertEqual(self.task1.state, state)

    @patch('arkindex.ponos.models.Task.short_logs', new_callable=PropertyMock)
    def test_partial_update_task_from_agent(self, short_logs_mock):
        short_logs_mock.return_value = ""
        self.assertEqual(self.task1.state, State.Unscheduled)
        self.task1.agent = self.agent
        self.task1.save()

        with self.assertNumQueries(8):
            resp = self.client.patch(
                reverse("api:task-details", args=[self.task1.id]),
                data={"state": State.Running.value},
                HTTP_AUTHORIZATION=f"Bearer {self.agent.token.access_token}",
            )
            self.assertEqual(resp.status_code, status.HTTP_200_OK)

        data = resp.json()
        self.assertEqual(data["id"], str(self.task1.id))
        self.assertEqual(data["state"], State.Running.value)
        self.task1.refresh_from_db()
        self.assertEqual(self.task1.state, State.Running)

    def test_partial_update_running_task_state_stopping(self):
        self.task1.state = State.Running
        self.task1.save()
        self.client.force_login(self.superuser)

        with self.assertNumQueries(4):
            resp = self.client.patch(
                reverse("api:task-update", args=[self.task1.id]),
                data={"state": State.Stopping.value},
            )
            self.assertEqual(resp.status_code, status.HTTP_200_OK)

        self.assertDictEqual(resp.json(), {
            'id': str(self.task1.id),
            'state': State.Stopping.value,
        })
        self.task1.refresh_from_db()
        self.assertEqual(self.task1.state, State.Stopping)

    def test_partial_update_non_running_task_state_stopping(self):
        states = list(State)
        states.remove(State.Running)
        self.task1.save()
        self.client.force_login(self.superuser)

        for state in states:
            with self.subTest(state=state):
                self.task1.state = state
                self.task1.save()

                with self.assertNumQueries(3):
                    resp = self.client.patch(
                        reverse("api:task-update", args=[self.task1.id]),
                        data={"state": State.Stopping.value},
                    )
                    self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertListEqual(resp.json(), ["You can only stop a 'Running' task"])

                self.task1.refresh_from_db()
                self.assertEqual(self.task1.state, state)

    def test_partial_update_final_task_state_pending(self):
        self.task1.state = State.Completed
        self.task1.agent = self.agent
        self.task1.save()
        self.client.force_login(self.superuser)

        with self.assertNumQueries(4):
            resp = self.client.patch(
                reverse("api:task-update", args=[self.task1.id]),
                data={"state": State.Pending.value},
            )
            self.assertEqual(resp.status_code, status.HTTP_200_OK)

        self.task1.refresh_from_db()
        self.assertEqual(self.task1.state, State.Pending)
        self.assertIsNone(self.task1.agent)

    def test_partial_update_non_final_task_state_pending(self):
        states = set(State) - set(FINAL_STATES)
        self.task1.save()
        self.client.force_login(self.superuser)

        for state in states:
            with self.subTest(state=state):
                self.task1.state = state
                self.task1.save()

                with self.assertNumQueries(3):
                    resp = self.client.patch(
                        reverse("api:task-update", args=[self.task1.id]),
                        data={"state": State.Pending.value},
                    )
                    self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)

                self.assertListEqual(
                    resp.json(),
                    ["You can only restart a task with a state equal to 'Completed', 'Failed', 'Error' or 'Stopped'."],
                )
                self.task1.refresh_from_db()
                self.assertEqual(self.task1.state, state)

    @patch("arkindex.ponos.aws.s3")
    @patch("arkindex.ponos.models.s3")
    def test_task_logs_unicode_error(self, s3_mock, aws_s3_mock):
        """
        Ensure the Task.short_logs property is able to handle sliced off Unicode characters.
        Since we fetch the latest logs from S3 using `Range: bytes=-N`, sometimes the characters
        can have a missing byte and cause a UnicodeDecodeError, such as U+00A0 (non-breaking space)
        or U+00A9 (copyright symbol).
        """
        with self.assertRaises(UnicodeDecodeError):
            b"\xa0Failed successfully".decode("utf-8")

        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.return_value = {
            "Body": BytesIO(b"\xa0Failed successfully")
        }
        aws_s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"

        resp = self.client.get(reverse("api:task-details", args=[self.task1.id]))
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        data = resp.json()

        self.assertEqual(data["logs"], "�Failed successfully")

    @patch("arkindex.ponos.serializers.check_agent_key")
    @patch("arkindex.ponos.serializers.timezone")
    def test_agent_create(self, timezone_mock, check_mock):
        check_mock.return_value = True
        timezone_mock.now.return_value = timezone.datetime(2000, 1, 1, 12).astimezone()

        resp = self.client.post(
            reverse("api:agent-register"),
            {
                "hostname": "toastname",
                "cpu_cores": 42,
                "cpu_frequency": 1337e6,
                "ram_total": 16e9,
                "farm": str(self.farm.id),
                "public_key": build_public_key(),
                "derivation": "{:032x}".format(random.getrandbits(128)),
                "gpus": [
                    {
                        "id": "b6b8d7c1-c6bd-4de6-ae92-866a270be36f",
                        "name": "A",
                        "index": 0,
                        "ram_total": 512 * 1024 * 1024,
                    },
                    {
                        "id": "14fe053a-8014-46a3-ad7b-99fa4389d74c",
                        "name": "B",
                        "index": 1,
                        "ram_total": 2 * 1024 * 1024 * 1024,
                    },
                ],
            },
            format="json",
        )

        self.assertEqual(resp.status_code, status.HTTP_201_CREATED)
        self.assertEqual(check_mock.call_count, 1)
        data = resp.json()
        agent = Agent.objects.get(id=data["id"])
        self.assertEqual(agent.hostname, "toastname")
        self.assertEqual(agent.cpu_cores, 42)
        self.assertEqual(agent.cpu_frequency, 1337e6)
        self.assertEqual(agent.gpus.count(), 2)
        gpu_a = agent.gpus.get(name="A")
        gpu_b = agent.gpus.get(name="B")
        self.assertEqual(gpu_a.index, 0)
        self.assertEqual(gpu_b.index, 1)
        self.assertEqual(gpu_a.ram_total, 512 * 1024 * 1024)
        self.assertEqual(gpu_b.ram_total, 2 * 1024 * 1024 * 1024)

        self.assertIn("access_token", data)
        self.assertIn("refresh_token", data)
        del data["access_token"]
        del data["refresh_token"]

        self.assertDictEqual(
            data,
            {
                "id": str(agent.id),
                "hostname": "toastname",
                "cpu_cores": 42,
                "cpu_frequency": int(1337e6),
                "farm": str(self.farm.id),
                "ram_total": 16_000_000_000,
                "cpu_load": None,
                "ram_load": None,
                "last_ping": "2000-01-01T12:00:00Z",
                "gpus": [
                    {
                        "id": "b6b8d7c1-c6bd-4de6-ae92-866a270be36f",
                        "index": 0,
                        "name": "A",
                        "ram_total": 512 * 1024 * 1024,
                    },
                    {
                        "id": "14fe053a-8014-46a3-ad7b-99fa4389d74c",
                        "index": 1,
                        "name": "B",
                        "ram_total": 2 * 1024 * 1024 * 1024,
                    },
                ],
            },
        )

    @patch("arkindex.ponos.serializers.check_agent_key")
    @patch("arkindex.ponos.serializers.timezone")
    def test_agent_update(self, timezone_mock, check_mock):
        check_mock.return_value = True
        timezone_mock.now.return_value = timezone.datetime(2000, 1, 1, 12).astimezone()

        # Still a POST, but on an existing agent
        resp = self.client.post(
            reverse("api:agent-register"),
            {
                "hostname": self.agent.hostname,
                "cpu_cores": 12,
                "cpu_frequency": 1e9,
                "ram_total": 32e9,
                "farm": str(self.farm.id),
                "public_key": self.agent.public_key,
                "derivation": "{:032x}".format(random.getrandbits(128)),
                "gpus": [
                    {
                        "id": "deadbeef-c6bd-4de6-ae92-866a270be36f",
                        "name": "new gpu",
                        "index": 0,
                        "ram_total": 32 * 1024 * 1024 * 1024,
                    },
                ],
            },
            format="json",
        )

        self.assertEqual(resp.status_code, status.HTTP_201_CREATED)
        self.assertEqual(check_mock.call_count, 1)
        data = resp.json()
        self.agent.refresh_from_db()
        self.assertEqual(self.agent.hostname, "ghostname")
        self.assertEqual(self.agent.cpu_cores, 12)
        self.assertEqual(self.agent.cpu_frequency, 1e9)
        self.assertEqual(self.agent.gpus.count(), 1)
        gpu = self.agent.gpus.get(name="new gpu")
        self.assertEqual(gpu.index, 0)
        self.assertEqual(gpu.ram_total, 32 * 1024 * 1024 * 1024)
        self.assertEqual(gpu.id, uuid.UUID("deadbeef-c6bd-4de6-ae92-866a270be36f"))

        self.assertIn("access_token", data)
        self.assertIn("refresh_token", data)
        del data["access_token"]
        del data["refresh_token"]

        self.assertDictEqual(
            data,
            {
                "cpu_cores": 12,
                "cpu_frequency": 1000000000,
                "cpu_load": None,
                "farm": str(self.farm.id),
                "gpus": [
                    {
                        "id": "deadbeef-c6bd-4de6-ae92-866a270be36f",
                        "index": 0,
                        "name": "new gpu",
                        "ram_total": 34359738368,
                    }
                ],
                "hostname": "ghostname",
                "id": str(self.agent.id),
                "last_ping": "2000-01-01T12:00:00Z",
                "ram_load": None,
                "ram_total": 32000000000,
            },
        )

    @patch("arkindex.ponos.serializers.check_agent_key")
    @patch("arkindex.ponos.serializers.timezone")
    def test_agent_create_existing_gpu(self, timezone_mock, check_mock):
        check_mock.return_value = True
        timezone_mock.now.return_value = timezone.datetime(2000, 1, 1, 12).astimezone()

        resp = self.client.post(
            reverse("api:agent-register"),
            {
                "hostname": "toastname",
                "cpu_cores": 42,
                "cpu_frequency": 1337e6,
                "ram_total": 16e9,
                "farm": str(self.farm.id),
                "public_key": build_public_key(),
                "derivation": "{:032x}".format(random.getrandbits(128)),
                "gpus": [
                    {
                        "id": str(self.gpu1.id),
                        "name": "A",
                        "index": 0,
                        "ram_total": 512 * 1024 * 1024,
                    },
                    {
                        "id": str(self.gpu2.id),
                        "name": "B",
                        "index": 1,
                        "ram_total": 2 * 1024 * 1024 * 1024,
                    },
                ],
            },
            format="json",
        )
        self.assertEqual(resp.status_code, status.HTTP_201_CREATED)
        self.assertEqual(check_mock.call_count, 1)
        data = resp.json()

        self.gpu1.refresh_from_db()
        self.gpu2.refresh_from_db()
        self.assertEqual(self.gpu1.name, "A")
        self.assertEqual(self.gpu2.name, "B")
        self.assertEqual(self.gpu1.ram_total, 512 * 1024**2)
        self.assertEqual(self.gpu2.ram_total, 2 * 1024**3)

        new_agent = Agent.objects.get(id=data["id"])
        self.assertEqual(self.gpu1.agent, new_agent)
        self.assertEqual(self.gpu2.agent, new_agent)

        # Existing agent no longer has any assigned GPUs, since the new agent stole them
        self.assertFalse(self.agent.gpus.exists())

    @patch("arkindex.ponos.serializers.check_agent_key")
    @patch("arkindex.ponos.serializers.timezone")
    def test_agent_update_existing_gpu(self, timezone_mock, check_mock):
        check_mock.return_value = True
        timezone_mock.now.return_value = timezone.datetime(2000, 1, 1, 12).astimezone()
        self.assertTrue(self.agent.gpus.count(), 2)

        # Still a POST, but on an existing agent
        resp = self.client.post(
            reverse("api:agent-register"),
            {
                "hostname": self.agent.hostname,
                "cpu_cores": 12,
                "cpu_frequency": 1e9,
                "ram_total": 32e9,
                "farm": str(self.farm.id),
                "public_key": self.agent.public_key,
                "derivation": "{:032x}".format(random.getrandbits(128)),
                "gpus": [
                    {
                        "id": "108c6524-c63a-4811-bbed-9723d32a0688",
                        "name": "GPU1",
                        "index": 0,
                        "ram_total": 2 * 1024 * 1024 * 1024,
                    },
                    {
                        "id": "f30d1407-92bb-484b-84b0-0b8bae41ca91",
                        "name": "GPU2",
                        "index": 1,
                        "ram_total": 8 * 1024 * 1024 * 1024,
                    },
                ],
            },
            format="json",
        )

        self.assertEqual(resp.status_code, status.HTTP_201_CREATED)

        self.agent.refresh_from_db()
        self.assertTrue(self.agent.gpus.count(), 2)

    @override_settings(PONOS_PRIVATE_KEY=PONOS_PRIVATE_KEY)
    def test_agent_create_bad_seed(self):
        # Build keys
        server_public_key = ec.generate_private_key(
            ec.SECP384R1(),
            default_backend(),
        ).public_key()

        agent_private_key = ec.generate_private_key(
            ec.SECP384R1(),
            default_backend(),
        )
        agent_public_bytes = agent_private_key.public_key().public_bytes(
            Encoding.PEM,
            PublicFormat.SubjectPublicKeyInfo,
        )

        # Add 1 to the farm's seed
        wrong_seed = "{:064x}".format(int(self.farm.seed, 16) + 1)

        # Perform derivation with the wrong seed
        shared_key = agent_private_key.exchange(ec.ECDH(), server_public_key)
        derived_key = HKDF(
            algorithm=SHA256(),
            backend=default_backend(),
            length=32,
            salt=None,
            info=wrong_seed.encode("utf-8"),
        ).derive(shared_key)

        resp = self.client.post(
            reverse("api:agent-register"),
            {
                "hostname": "toastname",
                "cpu_cores": 42,
                "cpu_frequency": 1337e6,
                "ram_total": 16e9,
                "farm": str(self.farm.id),
                "public_key": agent_public_bytes.decode("utf-8"),
                "derivation": base64.b64encode(derived_key).decode("utf-8"),
                "gpus": [],
            },
            format="json",
        )
        self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            resp.json(), {"non_field_errors": ["Key verification failed"]}
        )

    def test_agent_actions_requires_token(self):
        resp = self.client.get(reverse("api:agent-actions"))
        self.assertEqual(resp.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_agent_actions_no_user(self):
        self.client.force_login(self.superuser)
        try:
            resp = self.client.get(reverse("api:agent-actions"))
            self.assertEqual(resp.status_code, status.HTTP_401_UNAUTHORIZED)
        finally:
            self.client.logout()

    def test_agent_actions_query_params_required(self):
        resp = self.client.get(
            reverse("api:agent-actions"),
            HTTP_AUTHORIZATION=f'Bearer {self.agent.token.access_token}',
        )
        self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            resp.json(),
            {
                "cpu_load": ["This query parameter is required."],
                "ram_load": ["This query parameter is required."],
            },
        )

    def test_agent_actions_query_params_validation(self):
        resp = self.client.get(
            reverse("api:agent-actions"),
            HTTP_AUTHORIZATION=f'Bearer {self.agent.token.access_token}',
            data={"cpu_load": "high", "ram_load": "low"},
        )
        self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            resp.json(),
            {
                "cpu_load": ["A valid number is required."],
                "ram_load": ["A valid number is required."],
            },
        )

    def test_agent_non_pending_actions(self):
        """
        Only pending tasks may be retrieved as new actions
        """
        self.workflow.tasks.update(state=State.Error)
        with self.assertNumQueries(7):
            resp = self.client.get(
                reverse("api:agent-actions"),
                HTTP_AUTHORIZATION=f'Bearer {self.agent.token.access_token}',
                data={"cpu_load": 0.9, "ram_load": 0.49},
            )
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertDictEqual(resp.json(), {"actions": []})

    def test_agent_no_stealing(self):
        """
        An agent may not take another agent's tasks
        """
        self.workflow.tasks.update(agent=self.agent, state=State.Pending)
        pubkey = build_public_key()
        agent2 = AgentUser.objects.create(
            id=uuid.UUID(hashlib.md5(pubkey.encode("utf-8")).hexdigest()),
            farm=self.farm,
            hostname="agentorange",
            cpu_cores=2,
            cpu_frequency=1e9,
            public_key=pubkey,
            ram_total=2e9,
            last_ping=timezone.now(),
        )
        with self.assertNumQueries(7):
            resp = self.client.get(
                reverse("api:agent-actions"),
                HTTP_AUTHORIZATION=f'Bearer {agent2.token.access_token}',
                data={"cpu_load": 0.9, "ram_load": 0.49},
            )
            self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertDictEqual(resp.json(), {"actions": []})

    def test_agent_actions(self):
        """
        Agent may retrieve one tasks using the API due to its resources limitations
        """
        self.workflow.tasks.update(state=State.Pending)
        now = timezone.now()

        with patch.object(api_tz, "now") as api_now_mock:
            api_now_mock.return_value = now
            with self.assertNumQueries(11):
                resp = self.client.get(
                    reverse("api:agent-actions"),
                    HTTP_AUTHORIZATION="Bearer {}".format(
                        self.agent.token.access_token
                    ),
                    data={"cpu_load": 0.9, "ram_load": 0.49},
                )
        self.assertEqual(resp.status_code, status.HTTP_200_OK)

        self.agent.refresh_from_db()
        self.assertEqual(self.agent._estimate_new_tasks_cost(), 0.99)

        self.assertDictEqual(
            resp.json(),
            {
                "actions": [
                    {
                        "action": "start_task",
                        "task_id": str(self.task2.id),
                    }
                ]
            },
        )
        # Agent load and last ping attributes have been updated
        self.assertEqual(
            (self.agent.cpu_load, self.agent.ram_load, self.agent.last_ping),
            (0.9, 0.49, now),
        )

    def test_artifacts_list(self):
        url = reverse("api:task-artifacts", args=[self.task1.id])

        # Not accessible when logged out
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

        # Accessible but empty when logged in
        resp = self.client.get(
            url,
            HTTP_AUTHORIZATION=f'Bearer {self.agent.token.access_token}',
        )
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        data = resp.json()
        self.assertListEqual(data, [])

        # Add a few artifacts
        xxx = self.task1.artifacts.create(
            path="path/to/xxx", content_type="application/json", size=123
        )
        demo = self.task1.artifacts.create(
            path="demo.txt", content_type="text/plain", size=12
        )
        self.task2.artifacts.create(path="demo.txt", content_type="text/plain", size=1)

        resp = self.client.get(
            url,
            HTTP_AUTHORIZATION=f'Bearer {self.agent.token.access_token}',
        )
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        data = resp.json()

        # We only get the 2 artifacts on task1, ordered by path
        self.assertListEqual(
            data,
            [
                {
                    "content_type": "text/plain",
                    "created": str_date(demo.created),
                    "id": str(demo.id),
                    "path": "demo.txt",
                    "s3_put_url": None,
                    "size": 12,
                    "updated": str_date(demo.updated),
                    "url": f"http://testserver/api/v1/task/{self.task1.id}/artifact/demo.txt",
                },
                {
                    "content_type": "application/json",
                    "created": str_date(xxx.created),
                    "id": str(xxx.id),
                    "path": "path/to/xxx",
                    "s3_put_url": None,
                    "size": 123,
                    "updated": str_date(xxx.updated),
                    "url": f"http://testserver/api/v1/task/{self.task1.id}/artifact/path/to/xxx",
                },
            ],
        )

    def test_list_artifacts_process_creator(self):
        Process.objects.create(
            creator=self.user,
            corpus=self.corpus,
            mode=ProcessMode.Workers,
            workflow=self.workflow,
        )
        artifact = self.task1.artifacts.create(
            path="demo.txt",
            content_type="text/plain",
            size=12,
        )
        self.client.force_login(self.user)

        with self.assertNumQueries(7):
            response = self.client.get(reverse('api:task-artifacts', kwargs={'pk': self.task1.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertListEqual(response.json(), [
            {
                'id': str(artifact.id),
                'created': artifact.created.isoformat().replace('+00:00', 'Z'),
                'updated': artifact.updated.isoformat().replace('+00:00', 'Z'),
                'content_type': 'text/plain',
                'path': 'demo.txt',
                'size': 12,
                's3_put_url': None,
                'url': f'http://testserver/api/v1/task/{self.task1.id}/artifact/demo.txt',
            }
        ])

    @override_settings(PONOS_ARTIFACT_MAX_SIZE=99999)
    def test_artifact_creation(self):

        # No artifacts in DB at first
        self.assertFalse(self.task1.artifacts.exists())

        # Create an artifact through the API
        url = reverse("api:task-artifacts", args=[self.task1.id])
        resp = self.client.post(
            url,
            data={"path": "some/path.txt", "content_type": "text/plain", "size": 1000},
            HTTP_AUTHORIZATION=f'Bearer {self.agent.token.access_token}',
        )
        self.assertEqual(resp.status_code, status.HTTP_201_CREATED)

        # Check response has a valid S3 put URL, without matching the parameters in the querystring
        data = resp.json()
        s3_put_url = data.get("s3_put_url")
        self.assertIsNotNone(s3_put_url)
        del data["s3_put_url"]
        self.assertTrue(
            s3_put_url.startswith(
                f"http://somewhere/ponos-artifacts/{self.workflow.id}/{self.task1.id}/some/path.txt"
            )
        )

        # An artifact has been created
        artifact = self.task1.artifacts.get()

        self.assertDictEqual(
            resp.json(),
            {
                "content_type": "text/plain",
                "created": str_date(artifact.created),
                "id": str(artifact.id),
                "path": "some/path.txt",
                "size": 1000,
                "updated": str_date(artifact.updated),
                "url": f"http://testserver/api/v1/task/{self.task1.id}/artifact/some/path.txt",
            },
        )

        # Creating another artifact with the same path will fail cleanly
        resp = self.client.post(
            url,
            data={
                "path": "some/path.txt",
                "content_type": "text/plain",
                "size": 10243,
            },
            HTTP_AUTHORIZATION=f'Bearer {self.agent.token.access_token}',
        )
        self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            resp.json(), {"path": ["An artifact with this path already exists"]}
        )

    @override_settings()
    def test_create_artifact_size_limits(self):
        params = [
            (None, -42, "Ensure this value is greater than or equal to 1."),
            (None, 0, "Ensure this value is greater than or equal to 1."),
            (
                None,
                5 * 1024**3 + 1,
                "Ensure this value is less than or equal to 5368709120.",
            ),
            (123456789000, -42, "Ensure this value is greater than or equal to 1."),
            (123456789000, 0, "Ensure this value is greater than or equal to 1."),
            (
                123456789000,
                987654321000,
                "Ensure this value is less than or equal to 123456789000.",
            ),
        ]

        for size_setting, size, expected_error in params:
            with self.subTest(size=size):
                settings.PONOS_ARTIFACT_MAX_SIZE = size_setting
                url = reverse("api:task-artifacts", args=[self.task1.id])
                resp = self.client.post(
                    url,
                    data={
                        "path": "some/path.txt",
                        "content_type": "text/plain",
                        "size": size,
                    },
                    HTTP_AUTHORIZATION=f"Bearer {self.agent.token.access_token}",
                )
                self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertEqual(resp.json(), {"size": [expected_error]})

    @override_settings(PONOS_ARTIFACT_MAX_SIZE=99999)
    def test_create_artifact_task(self):
        Process.objects.create(
            creator=self.user,
            mode=ProcessMode.Repository,
            workflow=self.workflow,
        )

        with self.assertNumQueries(5):
            resp = self.client.post(
                reverse("api:task-artifacts", args=[self.task1.id]),
                data={
                    "path": "some/path.txt",
                    "content_type": "text/plain",
                    "size": 1000,
                },
                HTTP_AUTHORIZATION=f"Ponos {self.task1.token}",
            )
            self.assertEqual(resp.status_code, status.HTTP_201_CREATED, resp.json())

        data = resp.json()
        s3_put_url = data.pop("s3_put_url")
        self.assertTrue(
            s3_put_url.startswith(
                f"http://somewhere/ponos-artifacts/{self.workflow.id}/{self.task1.id}/some/path.txt"
            )
        )

        artifact = self.task1.artifacts.get()
        self.assertDictEqual(
            resp.json(),
            {
                "content_type": "text/plain",
                "created": str_date(artifact.created),
                "id": str(artifact.id),
                "path": "some/path.txt",
                "size": 1000,
                "updated": str_date(artifact.updated),
                "url": f"http://testserver/api/v1/task/{self.task1.id}/artifact/some/path.txt",
            },
        )

    def test_artifact_download(self):

        # Add a new artifact
        self.task1.artifacts.create(
            path="path/to/file.json",
            content_type="application/json",
            size=42,
        )

        # Request to get the real download link will fail as anonymous
        url = reverse(
            "api:task-artifact-download", args=[self.task1.id, "path/to/file.json"]
        )
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

        # Request to a missing artifact will lead to a 404
        bad_url = reverse(
            "api:task-artifact-download", args=[self.task1.id, "nope.xxx"]
        )
        resp = self.client.get(
            bad_url,
            HTTP_AUTHORIZATION=f'Bearer {self.agent.token.access_token}',
        )
        self.assertEqual(resp.status_code, status.HTTP_404_NOT_FOUND)

        # Valid request will redirect to the S3 url
        resp = self.client.get(
            url,
            HTTP_AUTHORIZATION=f'Bearer {self.agent.token.access_token}',
        )
        self.assertEqual(resp.status_code, status.HTTP_302_FOUND)

        # Check the S3 download link
        self.assertTrue(resp.has_header("Location"))
        self.assertTrue(
            resp["Location"].startswith(
                f"http://somewhere/ponos-artifacts/{self.workflow.id}/{self.task1.id}/path/to/file.json"
            )
        )

    def test_task_create_empty_body(self):
        response = self.client.post(reverse("api:task-create"))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {
                "image": ["This field is required."],
                "parents": ["This field is required."],
                "slug": ["This field is required."],
                "workflow_id": ["This field is required."],
            },
        )

    def test_task_create_no_parent(self):
        response = self.client.post(
            reverse("api:task-create"),
            data={
                "workflow_id": str(self.workflow.id),
                "slug": "test_task",
                "image": "registry.gitlab.com/test",
                "parents": [],
            },
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(), {"parents": ["This list may not be empty."]}
        )

    def test_task_create_distinct_workflows_on_parents(self):
        workflow2 = Workflow.objects.create(farm=self.farm, recipe=RECIPE)
        task3 = workflow2.tasks.create(
            run=0,
            depth=1,
            slug="task_parent",
            image="registry.gitlab.com/test",
        )

        response = self.client.post(
            reverse("api:task-create"),
            data={
                "workflow_id": str(self.workflow.id),
                "slug": "test_task",
                "image": "registry.gitlab.com/test",
                "parents": [str(self.task1.id), str(self.task2.id), str(task3.id)],
            },
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {
                "non_field_errors": [
                    "All parents must be in the same workflow as the child task"
                ]
            },
        )

    def test_task_create_distinct_runs_on_parents(self):
        task3 = self.workflow.tasks.create(
            run=1,
            depth=1,
            slug="task_parent",
            image="registry.gitlab.com/test",
        )

        response = self.client.post(
            reverse("api:task-create"),
            data={
                "workflow_id": str(self.workflow.id),
                "slug": "test_task",
                "image": "registry.gitlab.com/test",
                "parents": [str(self.task1.id), str(self.task2.id), str(task3.id)],
            },
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {
                "non_field_errors": [
                    "All parents must have the same run in the given workflow"
                ]
            },
        )

    def test_task_create_duplicate(self):
        self.workflow.tasks.create(
            run=0,
            depth=3,
            slug="sibling",
            image="registry.gitlab.com/test",
        )

        response = self.client.post(
            reverse("api:task-create"),
            data={
                "workflow_id": str(self.workflow.id),
                "slug": "sibling",
                "image": "registry.gitlab.com/test",
                "parents": [str(self.task1.id), str(self.task2.id)],
            },
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {
                "non_field_errors": [
                    "A task with the `sibling` slug already exists in run 0."
                ]
            },
        )

    def test_task_create(self):
        task3 = self.workflow.tasks.create(
            run=0,
            depth=3,
            slug="task_parent",
            image="registry.gitlab.com/test",
        )

        with self.assertNumQueries(12):
            response = self.client.post(
                reverse("api:task-create"),
                data={
                    "workflow_id": str(self.workflow.id),
                    "slug": "test_task",
                    "image": "registry.gitlab.com/test",
                    "parents": [str(self.task1.id), str(self.task2.id), str(task3.id)],
                    "command": "echo Test",
                    "env": {"test": "test", "test2": "test2"},
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        new_task = self.workflow.tasks.get(slug='test_task')

        self.assertDictEqual(
            response.json(),
            {
                "id": str(new_task.id),
                "workflow_id": str(self.workflow.id),
                "slug": "test_task",
                "parents": [str(self.task1.id), str(self.task2.id), str(task3.id)],
                "image": "registry.gitlab.com/test",
                "command": "echo Test",
                "env": {
                    "test": "test",
                    "top_env_variable": "workflow_variable",
                    "test2": "test2",
                    "ARKINDEX_TASK_TOKEN": new_task.token,
                },
                "run": 0,
                "depth": 4,
                "has_docker_socket": False,
            },
        )

    def test_task_create_has_docker_socket_true(self):
        task3 = self.workflow.tasks.create(
            run=0,
            depth=3,
            slug="task_parent",
            image="registry.gitlab.com/test",
        )

        with self.assertNumQueries(12):
            response = self.client.post(
                reverse("api:task-create"),
                data={
                    "workflow_id": str(self.workflow.id),
                    "slug": "test_task",
                    "image": "registry.gitlab.com/test",
                    "parents": [str(self.task1.id), str(self.task2.id), str(task3.id)],
                    "command": "echo Test",
                    "env": {"test": "test", "test2": "test2"},
                    "has_docker_socket": True,
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        new_task = self.workflow.tasks.get(slug='test_task')

        self.assertDictEqual(
            response.json(),
            {
                "id": str(new_task.id),
                "workflow_id": str(self.workflow.id),
                "slug": "test_task",
                "parents": [str(self.task1.id), str(self.task2.id), str(task3.id)],
                "image": "registry.gitlab.com/test",
                "command": "echo Test",
                "env": {
                    "test": "test",
                    "top_env_variable": "workflow_variable",
                    "test2": "test2",
                    "ARKINDEX_TASK_TOKEN": new_task.token,
                },
                "run": 0,
                "depth": 4,
                "has_docker_socket": True,
            },
        )

    def test_retrieve_secret_requires_auth(self):
        """
        Only agents or tasks may access a secret details
        """
        with self.assertNumQueries(0):
            response = self.client.get(
                reverse("api:secret-details", kwargs={"name": "abc"})
            )
            self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)
        self.assertDictEqual(
            response.json(), {"detail": "Authentication credentials were not provided."}
        )

    def test_retrieve_secret_not_found(self):
        """
        A 404 should be raised when no secret match query name
        """
        response = self.client.get(
            reverse(
                "api:secret-details", kwargs={"name": "the_most_important_secret"}
            ),
            HTTP_AUTHORIZATION=f'Bearer {self.agent.token.access_token}',
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertDictEqual(response.json(), {"detail": "Not found."})

    @override_settings(PONOS_PRIVATE_KEY=PONOS_PRIVATE_KEY)
    def test_agent_retrieve_secret(self):
        """
        Agent should be able to retrieve a secret content as cleartext
        """
        account_name = "bank_account/0001/private"
        secret = Secret.objects.create(
            name=account_name,
            nonce=b"1337" * 4,
            content=encrypt(b"1337" * 4, "1337$"),
        )
        self.assertEqual(secret.content, b"\xc1\x81\xc0\xceo")

        with self.assertNumQueries(2):
            response = self.client.get(
                reverse("api:secret-details", kwargs={"name": account_name}),
                HTTP_AUTHORIZATION=f'Bearer {self.agent.token.access_token}',
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(
            response.json(),
            {
                "id": str(secret.id),
                "name": account_name,
                "content": "1337$",
            },
        )

    @override_settings(PONOS_PRIVATE_KEY=PONOS_PRIVATE_KEY)
    def test_retrieve_secret_forbid_users(self):
        account_name = "bank_account/0001/private"
        secret = Secret.objects.create(
            name=account_name,
            nonce=b"1337" * 4,
            content=encrypt(b"1337" * 4, "1337$"),
        )
        self.assertEqual(secret.content, b"\xc1\x81\xc0\xceo")

        for user in [self.user, self.superuser]:
            with self.subTest(user=user):
                self.client.force_login(user)
                with self.assertNumQueries(0):
                    response = self.client.get(reverse("api:secret-details", kwargs={"name": account_name}))
                    self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    @override_settings(PONOS_PRIVATE_KEY=PONOS_PRIVATE_KEY)
    def test_task_retrieve_secret_requires_process(self):
        account_name = "bank_account/0001/private"
        secret = Secret.objects.create(
            name=account_name,
            nonce=b"1337" * 4,
            content=encrypt(b"1337" * 4, "1337$"),
        )
        self.assertEqual(secret.content, b"\xc1\x81\xc0\xceo")

        with self.assertNumQueries(1):
            response = self.client.get(
                reverse("api:secret-details", kwargs={"name": account_name}),
                HTTP_AUTHORIZATION=f'Ponos {self.task1.token}',
            )
            self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

        self.assertDictEqual(
            response.json(),
            {'detail': 'Task has no process.'},
        )

    @override_settings(PONOS_PRIVATE_KEY=PONOS_PRIVATE_KEY)
    def test_task_retrieve_secret_requires_active_user(self):
        account_name = "bank_account/0001/private"
        secret = Secret.objects.create(
            name=account_name,
            nonce=b"1337" * 4,
            content=encrypt(b"1337" * 4, "1337$"),
        )
        self.assertEqual(secret.content, b"\xc1\x81\xc0\xceo")

        Process.objects.create(
            creator=self.user,
            mode=ProcessMode.Repository,
            workflow=self.workflow,
        )
        self.user.is_active = False
        self.user.save()

        with self.assertNumQueries(1):
            response = self.client.get(
                reverse("api:secret-details", kwargs={"name": account_name}),
                HTTP_AUTHORIZATION=f'Ponos {self.task1.token}',
            )
            self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

        self.assertDictEqual(
            response.json(),
            {'detail': 'User inactive or deleted.'},
        )

    @override_settings(PONOS_PRIVATE_KEY=PONOS_PRIVATE_KEY)
    def test_task_retrieve_secret(self):
        account_name = "bank_account/0001/private"
        secret = Secret.objects.create(
            name=account_name,
            nonce=b"1337" * 4,
            content=encrypt(b"1337" * 4, "1337$"),
        )
        self.assertEqual(secret.content, b"\xc1\x81\xc0\xceo")

        corpus = Corpus.objects.create(name='Test corpus')
        corpus.processes.create(
            mode=ProcessMode.Workers,
            creator=self.user,
            workflow=self.workflow,
        )

        with self.assertNumQueries(2):
            response = self.client.get(
                reverse("api:secret-details", kwargs={"name": account_name}),
                HTTP_AUTHORIZATION=f'Ponos {self.task1.token}',
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(
            response.json(),
            {
                "id": str(secret.id),
                "name": account_name,
                "content": "1337$",
            },
        )

    def test_list_agents_state(self):
        """
        Lists agents from all farms with their status
        """
        # Add tasks with different states on the agent
        self.agent.tasks.bulk_create(
            [
                Task(
                    run=0,
                    depth=0,
                    workflow=self.workflow,
                    slug=state.value,
                    state=state,
                    agent=self.agent,
                )
                for state in State
            ]
        )
        self.client.force_login(self.superuser)
        with self.assertNumQueries(6):
            response = self.client.get(reverse("api:agents-state"))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        data = response.json()
        self.assertEqual(data["count"], 1)
        agent_state = data["results"][0]
        del agent_state["last_ping"]
        self.assertDictEqual(
            agent_state,
            {
                "id": str(self.agent.id),
                "active": True,
                "cpu_cores": 2,
                "cpu_frequency": 1000000000,
                "cpu_load": None,
                "farm": {
                    "id": str(self.farm.id),
                    "name": "Wheat farm",
                },
                "gpus": [
                    {
                        "id": "108c6524-c63a-4811-bbed-9723d32a0688",
                        "index": 0,
                        "name": "GPU1",
                        "ram_total": 2147483648,
                    },
                    {
                        "id": "f30d1407-92bb-484b-84b0-0b8bae41ca91",
                        "index": 1,
                        "name": "GPU2",
                        "ram_total": 8589934592,
                    },
                ],
                "hostname": "ghostname",
                "ram_load": None,
                "ram_total": 2000000000,
                "running_tasks_count": 1,
            },
        )

    def test_retrieve_agent_details(self):
        """
        The view returns an agents with its details and associated running tasks
        """
        # Add tasks with different states on the agent
        Task.objects.bulk_create(
            [
                Task(
                    run=0,
                    depth=0,
                    workflow=self.workflow,
                    slug=state.value,
                    state=state,
                    agent=self.agent,
                )
                for state in State
            ]
        )
        running_task = self.agent.tasks.get(state=State.Running)
        self.client.force_login(self.superuser)
        with self.assertNumQueries(7):
            response = self.client.get(
                reverse("api:agent-details", kwargs={"pk": str(self.agent.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        data = response.json()
        del data["last_ping"]
        self.assertDictEqual(
            data,
            {
                "id": str(self.agent.id),
                "active": True,
                "cpu_cores": 2,
                "cpu_frequency": 1000000000,
                "cpu_load": None,
                "farm": {
                    "id": str(self.farm.id),
                    "name": "Wheat farm",
                },
                "gpus": [
                    {
                        "id": "108c6524-c63a-4811-bbed-9723d32a0688",
                        "index": 0,
                        "name": "GPU1",
                        "ram_total": 2147483648,
                    },
                    {
                        "id": "f30d1407-92bb-484b-84b0-0b8bae41ca91",
                        "index": 1,
                        "name": "GPU2",
                        "ram_total": 8589934592,
                    },
                ],
                "hostname": "ghostname",
                "ram_load": None,
                "ram_total": 2000000000,
                "running_tasks": [
                    {
                        "id": str(running_task.id),
                        "run": 0,
                        "depth": 0,
                        "parents": [],
                        "slug": "running",
                        "state": "running",
                        "shm_size": None,
                        "tags": [],
                        "url": response.wsgi_request.build_absolute_uri(
                            reverse("api:task-details", args=[running_task.id])
                        ),
                    }
                ],
            },
        )

    def test_list_farms(self):
        """
        Any user is able to list farms basic information
        """
        barley_farm = Farm.objects.create(name="Barley")
        with self.assertNumQueries(2):
            response = self.client.get(reverse("api:farm-list"))
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(
            response.json(),
            {
                "count": 2,
                "number": 1,
                "previous": None,
                "next": None,
                "results": [
                    {"id": str(barley_farm.id), "name": "Barley"},
                    {"id": str(self.farm.id), "name": "Wheat farm"},
                ],
            },
        )

    @override_settings(PONOS_PRIVATE_KEY=PONOS_PRIVATE_KEY)
    def test_public_key(self):
        expected = load_pem_private_key(
            PONOS_PRIVATE_KEY.read_bytes(),
            password=None,
            backend=default_backend(),
        ).public_key().public_bytes(
            Encoding.PEM,
            PublicFormat.SubjectPublicKeyInfo,
        )

        with self.assertNumQueries(0):
            response = self.client.get(reverse('api:public-key'))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(response.content, expected)

    def test_agent_refresh(self):
        with self.assertNumQueries(0):
            response = self.client.post(
                reverse('api:agent-token-refresh'),
                {'refresh': str(self.agent.token)},
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        data = response.json()
        self.assertCountEqual(list(data.keys()), ['access', 'refresh'])
        access_token = AccessToken(data['access'])
        refresh_token = RefreshToken(data['refresh'])
        access_token.verify()
        refresh_token.verify()
        self.assertEqual(access_token.payload['agent_id'], str(self.agent.id))
        self.assertEqual(refresh_token.payload['agent_id'], str(self.agent.id))
