import logging
from collections import defaultdict
from datetime import timedelta
from textwrap import dedent
from uuid import UUID

from django.conf import settings
from django.core.mail import send_mail
from django.db import transaction
from django.db.models import Avg, CharField, Count, DurationField, F, Max, Min, Q, Value
from django.db.models.functions import Cast, Coalesce, Concat, Greatest, Now, NullIf
from django.db.models.query import Prefetch
from django.shortcuts import get_object_or_404
from django.template.loader import render_to_string
from django.utils.functional import cached_property
from drf_spectacular.utils import (
    OpenApiExample,
    OpenApiParameter,
    PolymorphicProxySerializer,
    extend_schema,
    extend_schema_view,
)
from rest_framework import permissions, status
from rest_framework.exceptions import NotFound, PermissionDenied, ValidationError
from rest_framework.generics import (
    CreateAPIView,
    DestroyAPIView,
    GenericAPIView,
    ListAPIView,
    ListCreateAPIView,
    RetrieveAPIView,
    RetrieveDestroyAPIView,
    RetrieveUpdateAPIView,
    RetrieveUpdateDestroyAPIView,
    UpdateAPIView,
)
from rest_framework.response import Response
from rest_framework.views import APIView

from arkindex.documents.models import Corpus, Element
from arkindex.ponos.authentication import TaskAuthentication
from arkindex.ponos.models import STATES_ORDERING, State, Task
from arkindex.ponos.permissions import IsTask
from arkindex.process.models import (
    ActivityState,
    DataFile,
    GitRef,
    GitRefType,
    Process,
    ProcessMode,
    Repository,
    Revision,
    Worker,
    WorkerActivity,
    WorkerActivityState,
    WorkerConfiguration,
    WorkerRun,
    WorkerType,
    WorkerVersion,
    WorkerVersionGPUUsage,
    WorkerVersionState,
)
from arkindex.process.providers import GitProvider
from arkindex.process.serializers.files import DataFileCreateSerializer, DataFileSerializer
from arkindex.process.serializers.git import ExternalRepositorySerializer, RevisionSerializer
from arkindex.process.serializers.imports import (
    ApplyProcessTemplateSerializer,
    CreateImportTranskribusErrorResponseSerializer,
    CreateProcessTemplateSerializer,
    ElementsWorkflowSerializer,
    FilesProcessSerializer,
    ImportTranskribusSerializer,
    ProcessElementLightSerializer,
    ProcessElementSerializer,
    ProcessListSerializer,
    ProcessSerializer,
    StartProcessSerializer,
)
from arkindex.process.serializers.ingest import BucketSerializer, S3ImportSerializer
from arkindex.process.serializers.training import StartTrainingSerializer
from arkindex.process.serializers.worker_runs import WorkerRunEditSerializer, WorkerRunSerializer
from arkindex.process.serializers.workers import (
    RepositorySerializer,
    WorkerActivitySerializer,
    WorkerConfigurationExistsErrorSerializer,
    WorkerConfigurationListSerializer,
    WorkerConfigurationSerializer,
    WorkerSerializer,
    WorkerStatisticsSerializer,
    WorkerTypeSerializer,
    WorkerVersionCreateSerializer,
    WorkerVersionSerializer,
)
from arkindex.project.aws import get_ingest_resource
from arkindex.project.fields import ArrayRemove
from arkindex.project.mixins import (
    ConflictAPIException,
    CorpusACLMixin,
    ProcessACLMixin,
    RepositoryACLMixin,
    SelectionMixin,
    WorkerACLMixin,
)
from arkindex.project.pagination import CustomCursorPagination
from arkindex.project.permissions import IsVerified, IsVerifiedOrReadOnly
from arkindex.project.tools import PercentileCont, RTrimChr
from arkindex.project.triggers import process_delete
from arkindex.training.models import ModelVersionState
from arkindex.users.models import OAuthCredentials, Role, Scope
from arkindex.users.utils import get_max_level

logger = logging.getLogger(__name__)


@extend_schema_view(
    get=extend_schema(
        operation_id='ListProcesses',
        tags=['process'],
        parameters=[
            OpenApiParameter(
                'corpus',
                type=UUID,
                description="Filter processes by corpus ID",
                required=False,
            ),
            OpenApiParameter(
                'created',
                type=bool,
                description='Restrict to or exclude processes you created',
                required=False,
            ),
            OpenApiParameter(
                'mode',
                enum=[mode.value for mode in ProcessMode],
                description="Filter processes by mode. When this filter is unset, the `local` mode is excluded.",
                required=False,
            ),
            OpenApiParameter(
                'state',
                enum=[state.value for state in State],
                description='Filter processes by workflow state',
                required=False,
            ),
            OpenApiParameter(
                'id',
                description='Filter processes by beginning of UUID',
                required=False,
            ),
            OpenApiParameter(
                'name',
                description='Filter processes whose name contains the given string (case insensitive)',
                required=False,
            ),
            OpenApiParameter(
                'with_workflow',
                type=bool,
                description='Restrict to or exclude processes with workflows. '
                            'This has no effect when filtering by the `template` or `local` modes.',
                required=False,
            )
        ]
    )
)
class ProcessList(ProcessACLMixin, ListAPIView):
    """
    List all visible processes.
    """
    permission_classes = (IsVerified, )
    serializer_class = ProcessListSerializer

    def get_queryset(self):
        filters = Q()

        if 'corpus' in self.request.query_params:
            corpus_id = self.request.query_params['corpus']
            try:
                corpus_id = UUID(corpus_id)
            except (AttributeError, ValueError):
                raise ValidationError({'corpus': ["'{}' is not a valid UUID".format(corpus_id)]})
            # No supplementary validation is required on the corpus ID filter
            filters &= Q(corpus=corpus_id)

        # Default process_mode variable for compatibility later
        process_mode = None
        if 'mode' in self.request.query_params:
            try:
                process_mode = ProcessMode(self.request.query_params['mode'])
                filters &= Q(mode=process_mode)
            except ValueError:
                raise ValidationError({
                    'mode': ["Mode '{}' does not exist".format(self.request.query_params['mode'])]
                })
        else:
            # Exclude local processes by default
            filters &= ~Q(mode=ProcessMode.Local)

        # When listing template or local processes, the workflow filter makes no sense
        if process_mode not in (ProcessMode.Template, ProcessMode.Local):
            if 'with_workflow' in self.request.query_params:
                with_workflow = self.request.query_params['with_workflow']
                filters &= Q(workflow__isnull=bool(with_workflow and with_workflow.lower() in ('false', '0')))

        if 'created' in self.request.query_params:
            created = self.request.query_params['created']
            creator_filter = Q(creator_id=self.request.user.id)
            if created.lower() in ('false', '0'):
                creator_filter = ~creator_filter
            filters &= creator_filter

        if 'id' in self.request.query_params:
            filters &= Q(id__startswith=self.request.query_params['id'])

        if 'name' in self.request.query_params:
            filters &= Q(name__icontains=self.request.query_params['name'])

        qs = self.readable_processes \
            .filter(filters) \
            .prefetch_related('workflow__tasks')
        # Order workflow by completion date when available, or by date of last updated task in workflow
        qs = qs.annotate(
            last_date=Greatest(Max('workflow__tasks__updated'), 'updated'),
            date_order=Coalesce('workflow__finished', 'last_date'),
            last_run=Max('workflow__tasks__run')
        )

        state_value = self.request.query_params.get('state')
        if state_value:
            try:
                state = State(self.request.query_params['state'])
            except ValueError:
                raise ValidationError({'state': [f"State '{state_value}' does not exist"]})
            # Filter out processes which have a task with an incompatible state on their last run
            excluding_states = STATES_ORDERING[:STATES_ORDERING.index(state)]
            excluded_processes = qs.filter(
                Q(workflow__tasks__run=F('last_run')),
                Q(workflow__tasks__state__in=excluding_states)
            )
            # Keep non excluded processes matching the state on their last run tasks
            state_query = Q(workflow__tasks__run=F('last_run')) & Q(workflow__tasks__state=state)
            if state == State.Unscheduled:
                # Handle a workflow with no task as unscheduled
                state_query |= Q(workflow__tasks__isnull=True)
            qs = qs.filter(state_query).exclude(id__in=excluded_processes.values('id')).distinct()

        return qs.order_by('-date_order')


class ProcessDetails(ProcessACLMixin, RetrieveAPIView):
    """
    Retrieve a process
    """
    permission_classes = (IsVerified, )
    serializer_class = ProcessSerializer
    queryset = Process.objects \
        .select_related('corpus', 'workflow') \
        .prefetch_related('workflow__tasks') \
        .annotate(last_run=Max('workflow__tasks__run'))

    def get_object(self):
        if not hasattr(self, '_process'):
            self._process = super().get_object()
        return self._process

    def get_serializer_context(self):
        context = super().get_serializer_context()
        if self.request and 'pk' in self.kwargs:
            context['process'] = self.get_object()
        return context

    def check_object_permissions(self, request, process):
        super().check_object_permissions(request, process)

        access_level = self.process_access_level(process)
        required_access = Role.Admin.value
        if request.method in permissions.SAFE_METHODS:
            # A guest role allow to retrieve information on a process
            required_access = Role.Guest.value

        if not access_level:
            raise NotFound
        if access_level < required_access:
            raise PermissionDenied(detail='You do not have a sufficient access level to this process.')

        if request.method == 'DELETE' and process.state == State.Running:
            raise ValidationError({'__all__': ['Cannot delete a workflow while it is running']})


@extend_schema(tags=['process'])
@extend_schema_view(
    get=extend_schema(description='Retrieve details of a process.\n\nRequires a **guest** access.'),
    patch=extend_schema(description='Partially update a process.\n\nRequires an **admin** access.'),
    put=extend_schema(description='Update dependencies of a process.\n\nRequires an **admin** access.'),
    delete=extend_schema(
        description=(
            'Delete a process.\n\n'
            'Cannot be used on currently running processes. '
            'Requires an **admin** access to the process.\n\n'
            'Triggers an asynchronous deletion of the process if it has linked '
            'activities and return a **HTTP_202_ACCEPTED**.'
        ),
        responses={
            202: None,
            204: None,
        },
    ),
)
class ProcessEdit(ProcessDetails, UpdateAPIView, DestroyAPIView):
    def destroy(self, request, *args, **kwargs):
        # If the process has no activity, deletes it directly. Triggers an async deletion otherwise.
        instance = self.get_object()
        if instance.activity_state == ActivityState.Pending:
            raise ValidationError({'__all__': ['Please wait activities to be initialized before deleting this process']})
        if instance.activities.exists():
            process_delete(instance, delete_activity=WorkerActivityState.Queued)
            return Response(status=status.HTTP_202_ACCEPTED)
        instance.delete()
        return Response(status=status.HTTP_204_NO_CONTENT)


@extend_schema(tags=['process'], deprecated=True)
@extend_schema_view(
    get=extend_schema(operation_id='RetrieveDataImport', description='Obsolete endpoint maintained for backward-compatibility purposes. Use **RetrieveProcess** instead.\n\nRetrieve details of a process.\nRequires a **guest** access.'),
)
class DataImportDetails(ProcessDetails):
    pass


class ProcessRetry(ProcessACLMixin, GenericAPIView):
    """
    Retry a process. Can only be used on processes with Error, Failed, Stopped or Completed states.\n\n
    Requires an **admin** access to the process.
    """
    permission_classes = (IsVerified, )
    serializer_class = ProcessSerializer
    queryset = Process.objects \
        .select_related('corpus', 'workflow') \
        .prefetch_related('workflow__tasks') \
        .annotate(last_run=Max('workflow__tasks__run'))

    def check_object_permissions(self, request, process):
        super().check_object_permissions(request, process)

        access_level = self.process_access_level(process)
        if not access_level:
            raise NotFound
        if access_level < Role.Admin.value:
            raise PermissionDenied(detail='You do not have an admin access to this process.')

        # process.state can cause new SQL queries to be run, so we access it just once
        state = process.state
        # Allow 'retrying' a process that has no Ponos workflow (that has never been started)
        if process.workflow is not None and state in (State.Unscheduled, State.Pending):
            raise ValidationError({'__all__': ['This workflow is already pending']})
        elif state == State.Running:
            raise ValidationError({'__all__': ['This workflow is already running']})
        elif state == State.Stopping:
            raise ValidationError({'__all__': ['This workflow is stopping']})

    @extend_schema(
        operation_id='RetryProcess',
        tags=['process'],
        request=None,
    )
    def post(self, request, *args, **kwargs):
        process = self.get_object()
        process.retry()
        serializer = self.get_serializer(process)
        return Response(serializer.data)


@extend_schema(
    responses={201: ProcessSerializer},
    tags=['files'],
)
class FilesProcess(CreateAPIView):
    """
    Start a process from one or more uploaded files.
    """

    permission_classes = (IsVerified, )
    serializer_class = FilesProcessSerializer

    def create(self, *args, **kwargs):
        super().create(*args, **kwargs)
        return Response(
            status=status.HTTP_201_CREATED,
            data=ProcessSerializer(self.process, context={'request': self.request}).data,
        )

    def perform_create(self, serializer):
        files = serializer.validated_data['files']
        corpus = files[0].corpus
        mode = serializer.validated_data['mode']
        # Serializer validation codes returns an element for 'folder_id'
        folder = serializer.validated_data.get('folder_id')
        folder_type = serializer.validated_data.get('folder_type')
        element_type = serializer.validated_data['element_type']

        if folder and folder.corpus != corpus:
            # The files' corpus is already validated as writable
            raise ValidationError({'__all__': ['Element and files are in different corpora']})

        # Start a process with thumbnails generation
        self.process = corpus.processes.create(
            creator=self.request.user,
            mode=mode,
            element=folder,
            folder_type=folder_type,
            element_type=element_type,
        )
        self.process.files.set(files)
        self.process.start(thumbnails=True)


@extend_schema_view(
    post=extend_schema(
        operation_id='CreateElementsWorkflow',
        tags=['process'],
        responses={201: ProcessSerializer},
    )
)
class CorpusWorkflow(SelectionMixin, CorpusACLMixin, CreateAPIView):
    """
    Create a distributed workflow from elements of an Arkindex corpus.\n\n
    Requires an **admin** access to the corpus.
    """
    permission_classes = (IsVerified, )
    serializer_class = ElementsWorkflowSerializer

    @transaction.atomic
    def create(self, request, pk=None, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        corpus, element, process_name, name_contains, element_type, selection, load_children, use_cache = map(
            lambda key: serializer.validated_data.pop(key, None),
            (
                'corpus',
                'element',
                'name',
                'element_name_contains',
                'element_type',
                'selection',
                'load_children',
                'use_cache',
            )
        )

        process = corpus.processes.create(
            creator=self.request.user,
            mode=ProcessMode.Workers,
            name=process_name,
            name_contains=name_contains,
            element_type=element_type,
            element=element,
            load_children=load_children,
            use_cache=use_cache,
        )

        # Ensure process elements do exist
        if selection:
            elements = self.get_selection(corpus.id)
            if not elements.exists():
                raise ValidationError({
                    '__all__': ['No element match those filters.']
                })
            process.elements.set(elements)

        return Response(
            status=status.HTTP_201_CREATED,
            data=ProcessSerializer(
                process,
                context=self.get_serializer_context()
            ).data,
        )


class StartProcess(CorpusACLMixin, APIView):
    """
    Start a process, used to build a Workflow with Workers.\n\n
    Requires an **admin** access to the corpus of this process.
    """
    permission_classes = (IsVerified, )
    # For OpenAPI type discovery
    queryset = Process.objects.none()

    @extend_schema(
        operation_id='StartProcess',
        tags=['process'],
        request=StartProcessSerializer,
        responses=ProcessSerializer
    )
    def post(self, request, pk=None, **kwargs):

        qs = Process.objects \
            .select_related('corpus') \
            .filter(corpus_id__isnull=False) \
            .prefetch_related(Prefetch('worker_runs', queryset=WorkerRun.objects.select_related('version', 'model_version')))

        process = get_object_or_404(qs, pk=self.kwargs['pk'])

        if not self.has_admin_access(process.corpus):
            raise PermissionDenied(detail='You do not have an admin access to the corpus of this process.')

        if process.mode != ProcessMode.Workers or process.workflow is not None:
            raise ValidationError(
                {'__all__': ['Only a Process with Workers mode and not already launched can be started later on']})

        serializer = StartProcessSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        data = serializer.validated_data

        errors = defaultdict(list)
        # Use process.worker_runs.all() to access the (prefetched) worker_runs to avoid new SQL queries
        # The related version have also been prefetched
        if len(list(process.worker_runs.all())) > 0:
            if data.get('use_gpu') and (not any(item.version.gpu_usage != WorkerVersionGPUUsage.Disabled for item in process.worker_runs.all())):
                errors['use_gpu'] = 'The process is configured to use GPU, but does not include any workers that support GPU usage.'

            # Check if a worker run has no model version but version.model_usage = True
            missing_model_versions = []
            unavailable_versions = []
            has_unavailable_model_versions = False

            for worker_run in process.worker_runs.all():
                if worker_run.version.model_usage and worker_run.model_version_id is None:
                    missing_model_versions.append(worker_run.version.worker.name)
                if worker_run.version.state != WorkerVersionState.Available or worker_run.version.docker_image_id is None:
                    unavailable_versions.append(worker_run.version.id)
                if worker_run.model_version_id and worker_run.model_version.state != ModelVersionState.Available:
                    has_unavailable_model_versions = True

            if len(missing_model_versions) > 0:
                errors['model_version'].append(f"The following workers require a model version and none was set: {missing_model_versions}")
            if has_unavailable_model_versions:
                errors['model_version'].append('This process contains one or more unavailable model versions and cannot be started.')
            if len(unavailable_versions) > 0:
                errors['version'] = 'This process contains one or more unavailable worker versions and cannot be started.'

        else:
            if data.get('worker_activity'):
                errors['worker_activity'] = 'The process must have workers attached to handle their activity.'
            if data.get('use_cache'):
                errors['use_cache'] = 'The process must have workers attached to use cached results.'
            if data.get('use_gpu'):
                errors['use_gpu'] = 'The process must have workers attached to use GPUs.'
            if not data.get('thumbnails'):
                errors['__all__'] = ['The process must either use thumbnail generation or have worker runs.']
        if errors:
            raise ValidationError(errors)

        process.start(**serializer.validated_data)

        return Response(
            status=status.HTTP_200_OK,
            data=ProcessSerializer(
                process,
                context={'request': request}
            ).data,
        )


@extend_schema(tags=['files'])
class DataFileList(CorpusACLMixin, ListAPIView):
    """
    List uploaded files in a corpus
    """
    permission_classes = (IsVerified, )
    serializer_class = DataFileSerializer
    # Tell the OpenAPI schema generator to believe this view is a list
    action = 'List'
    # For OpenAPI type discovery
    queryset = DataFile.objects.none()

    def get_queryset(self):
        return DataFile.objects.filter(corpus=self.get_corpus(self.kwargs['pk']), trashed=False)


@extend_schema(tags=['files'])
@extend_schema_view(
    get=extend_schema(description="Get an uploaded file's metadata"),
    patch=extend_schema(description="Update a datafile's status"),
    put=extend_schema(description="Update a datafile's status"),
    delete=extend_schema(description='Delete an uploaded file')
)
class DataFileRetrieve(CorpusACLMixin, RetrieveUpdateDestroyAPIView):
    """
    Get one file
    """
    permission_classes = (IsVerified, )
    serializer_class = DataFileSerializer

    def get_queryset(self):
        # We need to use the default database to avoid stale read
        # when a datafile is updated immediately after it's created on a small upload
        return DataFile.objects.using('default').filter(
            corpus__in=Corpus.objects.readable(self.request.user)
        )

    def perform_update(self, serializer):
        if not self.has_write_access(serializer.instance.corpus):
            raise PermissionDenied
        return super().perform_update(serializer)

    def perform_destroy(self, instance):
        if not self.has_write_access(instance.corpus):
            raise PermissionDenied

        # The file is simply marked as trashed here and will be deleted asynchronously through a cron job
        instance.trashed = True
        instance.save()


@extend_schema_view(post=extend_schema(operation_id='CreateDataFile', tags=['files']))
class DataFileCreate(CreateAPIView):
    """
    Create a Datafile.
    In case of success, a signed URI is returned to upload file content directly to a remote server.
    """

    permission_classes = (IsVerified, )
    serializer_class = DataFileCreateSerializer


@extend_schema(exclude=True)
class GitRepositoryImportHook(APIView):
    """
    This endpoint is intended as a webhook for Git repository hosting applications like GitLab.
    """

    def post(self, request, pk=None, **kwargs):
        repo = get_object_or_404(Repository, id=pk)
        if not repo.enabled:
            raise PermissionDenied(detail='No credentials available for this repository.')

        try:
            repo.provider_class(credentials=repo.credentials).handle_webhook(repo, request)
        except Exception as e:
            user = repo.credentials.user
            # Notifying the user by mail that there was an error during webhook handling
            sent = send_mail(
                subject='An error occurred during Git hook handling',
                message=render_to_string(
                    'webhook_error.html',
                    context={
                        'user': user,
                        'url': repo.url,
                        'error': e.args[0],
                    },
                    request=request,
                ),
                from_email=None,
                recipient_list=[user.email],
                fail_silently=True,
            )
            if sent == 0:
                logger.error(f'Failed to send webhook error email to {user.email}')

            raise

        return Response(status=status.HTTP_204_NO_CONTENT)


@extend_schema_view(
    get=extend_schema(
        tags=['repos'],
        operation_id='ListRepositories',
        description='List connected repositories',
    )
)
class RepositoryList(RepositoryACLMixin, ListAPIView):
    """
    List connected repositories
    """
    permission_classes = (IsVerified, )
    serializer_class = RepositorySerializer

    def get_queryset(self):
        return self.readable_repositories \
            .annotate(authorized_users=Count('memberships')) \
            .select_related('credentials') \
            .prefetch_related('workers__type') \
            .order_by('url')


@extend_schema_view(
    get=extend_schema(
        operation_id='ListExternalRepositories',
        parameters=[
            OpenApiParameter(
                'search',
                description='Optional query terms to filter repositories',
            )
        ],
        tags=['repos'],
    ),
    post=extend_schema(
        operation_id='CreateExternalRepository',
        description='Using the given OAuth credentials, this links an external Git repository '
                    'to Arkindex, connects a push hook and starts an initial process.',
        responses={201: ProcessSerializer},
        tags=['repos'],
    )
)
class AvailableRepositoriesList(ListCreateAPIView):
    """
    List repositories associated to user OAuth credentials

    Using the given OAuth credentials ID, this uses the Git hosting
    application API's search feature to look for a repository matching
    the given query. Without a query, returns a full list.
    """
    permission_classes = (IsVerified, )
    pagination_class = None
    serializer_class = ExternalRepositorySerializer
    queryset = OAuthCredentials.objects.none()

    def get_queryset(self):
        cred = get_object_or_404(OAuthCredentials, user=self.request.user, id=self.kwargs['pk'])
        provider = cred.git_provider_class(credentials=cred)
        repos = provider.list_repos(query=self.request.query_params.get('search'))
        return map(provider.to_dict, repos)

    @transaction.atomic
    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        user = self.request.user
        cred = get_object_or_404(OAuthCredentials, user=user, id=self.kwargs['pk'])
        provider = cred.git_provider_class(credentials=cred)
        repo = provider.create_repo(
            request=self.request,
            id=serializer.validated_data.get('id')
        )
        # Add an admin membership to the repository creator
        repo.memberships.create(user=user, level=Role.Admin.value)

        latest_commit_sha = provider.get_latest_commit_sha(repo)
        rev, _ = provider.get_or_create_revision(repo, latest_commit_sha)

        process = GitProvider.start_imports(provider, rev)

        # Return the serialized process
        return Response(
            status=status.HTTP_201_CREATED,
            data=ProcessSerializer(process, context={'request': self.request}).data,
        )


@extend_schema(
    tags=['repos'],
    description='Retrieve a repository',
)
@extend_schema_view(
    delete=extend_schema(description="Delete a repository")
)
class RepositoryRetrieve(RepositoryACLMixin, RetrieveDestroyAPIView):
    """
    Retrieve a repository
    """
    permission_classes = (IsVerified, )
    serializer_class = RepositorySerializer

    def get_queryset(self):
        return self.readable_repositories.annotate(authorized_users=Count('memberships'))

    def check_object_permissions(self, request, repo):
        if self.request.method not in permissions.SAFE_METHODS and not self.has_admin_access(repo):
            raise PermissionDenied(detail='You do not have admin access to this repository.')


@extend_schema(
    tags=['repos'],
    description='Retrieve a revision',
)
class RevisionRetrieve(RepositoryACLMixin, RetrieveAPIView):
    """
    Retrieve a revision
    """
    permission_classes = (IsVerified, )
    serializer_class = RevisionSerializer

    def get_queryset(self):
        return Revision.objects.filter(repo__in=self.executable_repositories)


@extend_schema(tags=['repos'])
@extend_schema_view(
    get=extend_schema(
        description='List workers for which the user has an execution access.',
        parameters=[
            OpenApiParameter(
                'repository_id',
                type=UUID,
                description='Filter workers by repository ID.',
                required=False,
            ),
            OpenApiParameter(
                'name',
                type=str,
                description='Filter workers by name.',
                required=False,
            )
        ]
    ),
    post=extend_schema(
        description=(
            'Create a worker.\n\n'
            'The user must be **internal** to perform this request.'
        )
    ),
)
class WorkerList(WorkerACLMixin, ListCreateAPIView):
    """
    List workers executable by the request user.
    Create a worker.
    """
    permission_classes = (IsVerified, )
    serializer_class = WorkerSerializer
    queryset = Worker.objects.none()

    def filter_queryset(self, qs):
        filters = Q()

        repo_id = self.request.query_params.get('repository_id')
        if repo_id:
            try:
                repo_id = UUID(repo_id)
            except (TypeError, ValueError):
                raise ValidationError({'repository_id': ['Invalid UUID']})
            filters &= Q(repository_id=repo_id)

        name_filter = self.request.query_params.get('name')
        if name_filter:
            filters &= Q(name__icontains=name_filter)

        worker_type = self.request.query_params.get('type')
        if worker_type:
            try:
                worker_type_id = UUID(worker_type)
                filters &= Q(type_id=worker_type_id)
            except ValueError:
                # If the given type is not a worker_type id, it might be a worker_type__slug
                try:
                    worker_type = WorkerType.objects.get(slug=worker_type)
                except WorkerType.DoesNotExist:
                    raise ValidationError({'type': ['No registered worker type with that slug.']})
                filters &= Q(type_id=worker_type.id)

        return super().filter_queryset(qs.filter(filters))

    def get_queryset(self):
        """
        List executable workers (i.e. via a direct write right or a write right on its repository).
        """
        return self.executable_workers \
            .order_by('name', 'id') \
            .select_related('type') \
            .distinct()

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        repo_id = serializer.validated_data['repository_id']
        repo = get_object_or_404(Repository, pk=repo_id)

        access_right = get_max_level(self.request.user, repo)
        if not access_right:
            raise NotFound
        elif access_right < Role.Admin.value:
            raise PermissionDenied(detail='You do not have admin access to this repository.')

        # Get or create a WorkerType with given slug
        worker_type = serializer.validated_data['type']
        worker, created = repo.workers.get_or_create(
            slug=serializer.validated_data['slug'],
            defaults={
                'name': serializer.validated_data['name'],
                'type': worker_type,
            }
        )

        reponse_status = status.HTTP_201_CREATED if created else status.HTTP_200_OK

        return Response(WorkerSerializer(worker).data, status=reponse_status)


@extend_schema(
    tags=['repos'],
    description='List available worker types on an Arkindex instance.',
)
class WorkerTypesList(ListAPIView):
    """
    List available worker types on instance
    """
    serializer_class = WorkerTypeSerializer
    queryset = WorkerType.objects.all().order_by('display_name')


@extend_schema(tags=['repos'])
@extend_schema_view(
    get=extend_schema(
        description=(
            'List versions for a given worker ID with their revision and associated git references.\n\n'
            'Requires an **execution** access to the worker or its repository.'
        ),
        parameters=[
            OpenApiParameter(
                'mode',
                type=str,
                default='complete',
                description=('If set to `simple`, exclude versions other than master/main or with no tag')
            )
        ],
    ),
    post=extend_schema(
        description=(
            'Create a version for a given worker ID, revision ID and JSON configuration.\n\n'
            'The user must be **internal** to perform this request.'
        )
    ),
)
class WorkerVersionList(WorkerACLMixin, ListCreateAPIView):
    """
    List versions, their revision and associated git references for a given worker UUID

    Create a WorkerVersion instance for a worker and a revision with a given JSON configuration,
    return existing query if a workerVersion already exists for this worker and this revision
    """
    permission_classes = (IsVerified, )
    serializer_class = WorkerVersionCreateSerializer
    queryset = WorkerVersion.objects.none()

    @property
    def simple_mode(self):
        return self.request.query_params.get('mode', 'complete').lower() == 'simple'

    def get_queryset(self):
        worker = get_object_or_404(
            Worker.objects.select_related('repository'),
            pk=self.kwargs['pk']
        )

        if not self.has_execution_access(worker):
            raise PermissionDenied(detail='You do not have an execution access to this worker.')

        filters = Q()
        if self.simple_mode:
            # Limit output to versions with tags or master/main branches
            filters = Q(revision__refs__type=GitRefType.Tag) | Q(revision__refs__type=GitRefType.Branch, revision__refs__name__in=["master", "main"])

        return worker.versions \
            .filter(filters) \
            .prefetch_related('revision__repo', 'revision__refs', 'worker__type') \
            .distinct() \
            .order_by('-revision__created')

    def create(self, request, *args, **kwargs):
        if not isinstance(request.auth, Task):
            raise PermissionDenied(detail='Only Ponos tasks can create a new version from a worker.')

        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        # We need to use the default database to avoid stale read
        # when creating a worker version on a newly created worker
        worker = get_object_or_404(
            Worker.objects.using('default'),
            pk=self.kwargs['pk']
        )

        revision = serializer.validated_data['revision_id']
        if worker.repository_id != revision.repo_id:
            raise ValidationError({
                'revision': ['The revision must be part of the same repository as the worker.']
            })

        version, created = WorkerVersion.objects.get_or_create(
            worker=worker,
            revision=revision,
            defaults={
                'configuration': serializer.validated_data['configuration'],
                'gpu_usage': serializer.validated_data['gpu_usage'],
                'model_usage': serializer.validated_data['model_usage'],
            }
        )

        reponse_status = status.HTTP_201_CREATED if created else status.HTTP_200_OK

        return Response(WorkerVersionSerializer(version).data, status=reponse_status)


@extend_schema_view(
    get=extend_schema(
        operation_id='ListCorpusWorkerVersions',
        tags=['ml'],
    )
)
class CorpusWorkerVersionList(CorpusACLMixin, ListAPIView):
    """
    List worker versions used by any ML result in a given corpus.

    Guest access is required on private corpora. No specific rights are required on the workers.
    """
    permission_classes = (IsVerifiedOrReadOnly, )
    serializer_class = WorkerVersionSerializer
    # For OpenAPI type discovery
    queryset = WorkerVersion.objects.none()

    @cached_property
    def corpus(self):
        return get_object_or_404(self.readable_corpora, pk=self.kwargs['pk'])

    def get_queryset(self):
        return (
            self.corpus.worker_versions
            .select_related(
                'worker',
                'revision',
            )
            .order_by(
                'worker__name',
                'revision__hash',
            )
            .prefetch_related(
                'revision__repo',
                'revision__refs',
                'worker__repository',
                'worker__type',
            )
        )


@extend_schema(tags=['repos'])
class WorkerVersionRetrieve(RetrieveUpdateAPIView):
    """
    Retrieve a specific worker version. No authentication is required.
    """
    permission_classes = (IsVerifiedOrReadOnly, )
    serializer_class = WorkerVersionSerializer
    queryset = WorkerVersion.objects.select_related('worker').all()

    def check_object_permissions(self, request, instance):
        """
        Allow any user to retrieve a worker version without authentication
        This behavior allows to retrieve a version information on public resources

        Only Ponos tasks are allowed to update a worker version
        """
        if self.request.method in permissions.SAFE_METHODS:
            return
        super().check_object_permissions(request, instance)
        if not isinstance(self.request.auth, Task):
            raise PermissionDenied(detail='Only Ponos tasks are allowed to update a worker version.')


@extend_schema(
    tags=['repos'],
    description='Retrieve a worker',
)
class WorkerRetrieve(WorkerACLMixin, RetrieveAPIView):
    permission_classes = (IsVerified, )
    serializer_class = WorkerSerializer
    queryset = Worker.objects.all()

    def get_object(self):
        queryset = self.filter_queryset(self.get_queryset()).prefetch_related('repository')
        worker = get_object_or_404(queryset, pk=self.kwargs['pk'])
        if not self.has_read_access(worker):
            raise PermissionDenied(detail='You do not have a guest access to this worker.')
        super().check_object_permissions(self.request, worker)
        return worker


@extend_schema(tags=['repos'])
@extend_schema_view(
    get=extend_schema(
        description=(
            'List configurations for a given worker ID.\n\n'
            'Requires a **read** access to the worker or its repository.'
        ),
        parameters=[
            OpenApiParameter(
                'archived',
                type=bool,
                description='Filter configurations by their `archived` state. '
                            'By default, the list only displays enabled (non-archived) configurations.',
                required=False,
                default=False,
            )
        ]
    ),
    post=extend_schema(
        description=(
            'Create a configuration for a given worker ID, name and JSON configuration.\n\n'
            'Requires an **admin** access to the worker or its repository.'
        ),
        responses={
            201: WorkerConfigurationSerializer,
            400: WorkerConfigurationExistsErrorSerializer
        },
        examples=[OpenApiExample(
            status_codes=['400'],
            response_only=True,
            name="config-exists",
            value={'id': ["01fd00d9-1020-4d84-85f2-d0785e2c5b4a"], 'configuration': ["A worker configuration with this configuration already exists for this worker."]},
            description="An error where a worker configuration with this configuration already exists, including the existing worker configuration's UUID."
        )]
    ),
)
class WorkerConfigurationList(WorkerACLMixin, ListCreateAPIView):
    permission_classes = (IsVerified, )
    serializer_class = WorkerConfigurationListSerializer
    queryset = WorkerConfiguration.objects.none()

    @cached_property
    def worker(self):
        worker = get_object_or_404(
            Worker.objects.select_related('repository'),
            pk=self.kwargs['pk']
        )
        if self.request.method in permissions.SAFE_METHODS and not self.has_read_access(worker):
            raise PermissionDenied(detail='You do not have a guest access to this worker.')
        if self.request.method not in permissions.SAFE_METHODS and not self.has_execution_access(worker):
            raise PermissionDenied(detail='You do not have contributor access to this worker.')
        return worker

    def get_serializer_context(self):
        context = super().get_serializer_context()
        if 'pk' in self.kwargs:
            context['worker'] = self.worker
        return context

    def get_queryset(self):
        return self.worker.configurations.order_by('name')

    def filter_queryset(self, queryset):
        # Consider any value that does not look like a false, including the empty string (`?archived` alone),
        # to be true, but consider a missing parameter to be false.
        archived = self.request.query_params.get('archived', 'false').lower() not in ('false', '0')
        return queryset.filter(archived=archived)


@extend_schema_view(
    get=extend_schema(description=dedent(
        """
        Retrieve details of a worker configuration.

        Requires a **read** access to the worker or its repository.
        """
    )),
    put=extend_schema(description=dedent(
        """
        Update a worker configuration.

        Requires a **write** access to the worker or its repository.
        """
    )),
    patch=extend_schema(description=dedent(
        """
        Update parts of a worker configuration.

        Requires a **write** access to the worker or its repository.
        """
    ))
)
@extend_schema(tags=['repos'])
class WorkerConfigurationRetrieve(WorkerACLMixin, RetrieveUpdateAPIView):
    permission_classes = (IsVerified, )
    serializer_class = WorkerConfigurationSerializer

    def get_queryset(self):
        return WorkerConfiguration.objects.select_related('worker__repository')

    def check_object_permissions(self, request, obj):
        if request.method in permissions.SAFE_METHODS:
            if not self.has_read_access(obj.worker):
                raise PermissionDenied(detail='You do not have a guest access to this worker.')
        else:
            if not self.has_execution_access(obj.worker):
                raise PermissionDenied(detail='You do not have a contributor access to this worker.')

        super().check_object_permissions(request, obj)


@extend_schema(tags=['process'])
@extend_schema_view(
    get=extend_schema(description=(
        'List worker runs for a given process ID.\n\n'
        'Requires an **admin** access to the process project.'
    )),
    post=extend_schema(description=(
        'Create a worker run for a given process ID.\n\n'
        'Requires an **admin** access to the process project and a **contributor** access to the worker.'
    )),
)
class WorkerRunList(WorkerACLMixin, ListCreateAPIView):
    """
    List worker runs for a given process UUID.
    Create a run on a process.
    """
    permission_classes = (IsVerified, )
    serializer_class = WorkerRunSerializer
    queryset = WorkerRun.objects.none()

    @cached_property
    def process(self):
        process = get_object_or_404(
            Process.objects.filter(corpus_id__isnull=False).select_related('corpus'),
            pk=self.kwargs['pk']
        )

        if not self.has_access(process.corpus, Role.Admin.value):
            raise PermissionDenied(detail='You do not have an admin access to the corpus of this process.')

        return process

    def get_queryset(self):
        return self.process.worker_runs \
            .select_related(
                'version__worker__repository',
                'version__worker__type',
                'version__revision__repo',
                'model_version__model',
            ) \
            .prefetch_related(Prefetch(
                'version__revision__refs',
                queryset=GitRef.objects.select_related('repository')
            )) \
            .order_by('id')

    def perform_create(self, serializer):
        # TODO: Aggregate errors using a defaultdict(list) and send them all at once instead of raising early, per best practices
        # https://redmine.teklia.com/projects/arkindex/wiki/Arkindex_API_best_practices#Error-responses
        try:
            version = (
                WorkerVersion.objects
                # Required for the WorkerACLMixin
                .select_related('worker__repository')
                .only('id', 'state', 'worker__id', 'worker__public', 'worker__repository__id')
                .get(id=serializer.validated_data['version_id'])
            )
        except WorkerVersion.DoesNotExist:
            raise ValidationError({'worker_version_id': ['This version does not exist.']})

        # User must have admin access to the process project and a contributor access to the worker run
        if not self.has_execution_access(version.worker):
            raise ValidationError({'worker_version_id': ['You do not have an execution access to this version.']})

        if self.process.worker_runs.filter(version_id=serializer.validated_data['version_id']).exists():
            raise ValidationError({'worker_version_id': ['A WorkerRun with this version already exists on this process.']})

        if version.state != WorkerVersionState.Available:
            raise ValidationError({'worker_version_id': ['This worker version is not in an Available state.']})

        configuration = serializer.validated_data.pop('configuration_id', None)
        if configuration and configuration.worker_id != version.worker.id:
            raise ValidationError({'configuration_id': ['The configuration must be part of the same worker.']})

        if self.process.mode != ProcessMode.Workers:
            raise ValidationError({'process': ['Import mode must be Workers']})

        if self.process.workflow_id is not None:
            raise ValidationError({'__all__': ["Cannot create a WorkerRun on a Process that has already started"]})

        serializer.save(process=self.process, configuration=configuration)


@extend_schema(tags=['process'])
@extend_schema_view(
    get=extend_schema(
        operation_id='RetrieveWorkerRun',
        description='Retrieve details of a worker run.\n\nRequires an **admin** access to the process project.'
    ),
    patch=extend_schema(
        operation_id='PartialUpdateWorkerRun',
        description='Partially update a worker run.\n\nRequires an **admin** access to the process project.'
    ),
    put=extend_schema(
        operation_id='UpdateWorkerRun',
        description='Update dependencies of a worker run.\n\nRequires an **admin** access to the process project.'
    ),
    delete=extend_schema(
        operation_id='DestroyWorkerRun',
        description='Delete a worker run.\n\nRequires an **admin** access to the process project.'
    ),
)
class WorkerRunDetails(CorpusACLMixin, RetrieveUpdateDestroyAPIView):
    """
    Retrieve or edit a worker run i.e. update its dependencies.
    Requires an admin access to the process project.
    """
    permission_classes = (IsVerified, )
    serializer_class = WorkerRunEditSerializer

    def get_object(self):
        if not hasattr(self, '_worker_run'):
            self._worker_run = super().get_object()
        return self._worker_run

    def get_queryset(self):
        # Use default DB to avoid a race condition checking process workflow
        return WorkerRun.objects \
            .filter(process__corpus_id__isnull=False) \
            .using('default') \
            .select_related('version__worker__type', 'configuration', 'process__workflow', 'process__corpus', 'version__revision__repo')

    def get_serializer_context(self):
        context = super().get_serializer_context()
        if 'model_version_id' in self.request.data:
            context['model_usage'] = self.get_object().version.model_usage
        return context

    def check_object_permissions(self, request, worker_run):
        if not self.has_admin_access(worker_run.process.corpus):
            raise PermissionDenied(detail='You do not have an admin access to the process project.')

        # Updating a worker run is not possible once the process is started
        if request.method not in permissions.SAFE_METHODS and worker_run.process.workflow_id is not None:
            raise ValidationError({'__all__': ["Cannot update a WorkerRun on a Process that has already started"]})

        super().check_object_permissions(request, worker_run)

    def perform_destroy(self, instance):
        # Remove references to this instance from its children parents field
        instance.process.worker_runs.filter(parents__contains=[instance.id]).update(parents=ArrayRemove('parents', instance.id))
        return super().perform_destroy(instance)

    def perform_update(self, serializer):
        worker = serializer.instance.version.worker
        configuration = serializer.validated_data.get('configuration_id', None)
        if configuration and configuration.worker_id != worker.id:
            raise ValidationError({'configuration_id': ['The configuration must be part of the same worker.']})
        super().perform_update(serializer)


@extend_schema_view(post=extend_schema(
    operation_id='CreateImportTranskribus',
    tags=['process'],
    responses={
        201: ProcessSerializer,
        400: CreateImportTranskribusErrorResponseSerializer
    },
    examples=[OpenApiExample(
        status_codes=['400'],
        response_only=True,
        name="user-permission",
        value={'collection_id': "User user@example.com is not a member of the collection 1"},
        description="An error where the user is not a member of the collection."
    )]
))
class ImportTranskribus(CreateAPIView):
    """
    Start a data import from Transkribus email and collection ID.
    """

    permission_classes = (IsVerified, )
    serializer_class = ImportTranskribusSerializer

    def create(self, *args, **kwargs):
        if not settings.ARKINDEX_FEATURES['transkribus']:
            raise ValidationError(['Transkribus import is unavailable due to the transkribus feature being disabled.'])
        super().create(*args, **kwargs)
        return Response(
            status=status.HTTP_201_CREATED,
            data=ProcessSerializer(self.process, context={'request': self.request}).data,
        )

    def perform_create(self, serializer):
        collection_id = serializer.validated_data['collection_id']

        # Create corpus
        corpus = Corpus.objects.create(
            name=f"Transkribus collection n°{collection_id}",
        )
        corpus.memberships.create(user=self.request.user, level=Role.Admin.value)
        corpus.create_default_transkribus_types()

        # Start a process with thumbnails generation
        self.process = corpus.processes.create(
            creator=self.request.user,
            mode=ProcessMode.Transkribus,
            collection_id=collection_id
        )
        self.process.start(thumbnails=True)


@extend_schema_view(get=extend_schema(
    operation_id='ListProcessElements',
    parameters=[
        OpenApiParameter(
            'with_image',
            type=bool,
            default=False,
            description='Include zone and image information in the response.',
        )
    ],
    responses=PolymorphicProxySerializer(
        component_name='ListProcessElementsResponse',
        serializers={
            True: ProcessElementSerializer,
            False: ProcessElementLightSerializer,
        },
        resource_type_field_name='with_image',
    ),
    tags=['process']
))
class ListProcessElements(CorpusACLMixin, ListAPIView):
    """
    List all elements for a process with workers.\n\n
    Requires an **admin** access to the process corpus.
    """
    pagination_class = CustomCursorPagination
    permission_classes = (IsVerified, )
    # For OpenAPI type discovery
    queryset = Element.objects.none()

    @property
    def with_image(self):
        return self.request.query_params.get('with_image', 'false').lower() in ('true', '1')

    @cached_property
    def process(self):
        return get_object_or_404(
            # Avoid stale read on newly created processes
            Process.objects.using('default').filter(corpus_id__isnull=False),
            Q(pk=self.kwargs['pk'])
        )

    def get_serializer_class(self):
        if self.with_image:
            return ProcessElementSerializer
        else:
            return ProcessElementLightSerializer

    def get_queryset(self):
        if not self.has_admin_access(self.process.corpus):
            raise PermissionDenied(detail='You do not have an admin access to the corpus of this process.')

        try:
            queryset = self.process.list_elements()
        except AssertionError as e:
            raise ValidationError({'__all__': [str(e)]})

        if not self.with_image:
            return queryset.values('id', 'type_id', 'name', 'confidence')

        return queryset.annotate(
            # Build the image URL by concatenating the server's URL to the image's path
            # Server URLs might end in a slash, but not all the time,
            # so we strip any trailing slashes and add our own.
            # This uses NullIf as this concatenation will return '/' if there is no zone on the element.
            image_url=NullIf(
                Concat(
                    RTrimChr('image__server__url', Value('/')),
                    Value('/'),
                    Cast('image__path', CharField()),
                ),
                Value('/'),
            )
        ).values(
            'id',
            'type_id',
            'name',
            'confidence',
            'image_id',
            'image__width',
            'image__height',
            'polygon',
            'image_url',
            'rotation_angle',
            'mirrored',
        )


class UpdateWorkerActivity(GenericAPIView):
    """
    Allow a Ponos task to update an element's state
    Only allow defined evolutions of the element's state
    """
    authentication_classes = (TaskAuthentication, )
    permission_classes = (IsTask, )
    serializer_class = WorkerActivitySerializer
    queryset = WorkerActivity.objects.none()

    @cached_property
    def allowed_transitions(self):
        # Defines a list of allowed previous states for any transition
        queued = WorkerActivityState.Queued.value
        started = WorkerActivityState.Started.value
        error = WorkerActivityState.Error.value
        processed = WorkerActivityState.Processed.value
        return {
            queued: [error],
            started: [
                queued,
                # If a worker failed for some reason to update an activity to a processed or error state,
                # retrying would cause the new worker to try to update from started to started.
                started,
                error
            ],
            error: [started],
            processed: [started],
        }

    @extend_schema(
        tags=['ml'],
        responses={
            200: WorkerActivitySerializer,
            409: None,
        },
        operation_id='UpdateWorkerActivity',
        description=dedent("""
            Updates the activity of a worker version on an element.

            Only a Ponos task authenticated via task authentication may perform this request.

            A **HTTP_409_CONFLICT** is returned in case the body is valid but the update failed.
        """).strip(),
    )
    def put(self, request, *args, **kwarg):
        """
        Update a worker activity with two database requests.
        If the group element_id, worker_version, configuration_id matches no existing
        activity, the update count is 0.
        If the new state is disallowed, the update count is 0 too.
        """
        serializer = self.get_serializer(data=request.data, partial=False)
        serializer.is_valid(raise_exception=True)

        worker_version_id = self.kwargs['pk']
        element_id = serializer.validated_data['element_id']
        state = serializer.validated_data['state'].value
        process_id = serializer.validated_data['process_id']

        # Between zero and one worker run can match the filter due to the DB constraint
        worker_run = WorkerRun.objects.filter(process_id=process_id, version_id=worker_version_id).first()
        configuration_id = worker_run.configuration_id if worker_run else None

        # Between zero and one worker activity can match the filter due to the DB constraint
        activity_filter = Q(
            worker_version_id=worker_version_id,
            element_id=element_id,
            configuration_id=configuration_id,
            process_id=process_id,
            state__in=self.allowed_transitions[state]
        )

        fields_to_update = {'state': state}

        # If we are switching to started, we need to restrict to activities that either are not started,
        # or have not been updated for an hour, and also update the `started` timestamp.
        if state == WorkerActivityState.Started.value:
            activity_filter &= (
                ~Q(state=WorkerActivityState.Started)
                | Q(updated__lte=Now() - timedelta(seconds=settings.WORKER_ACTIVITY_TIMEOUT))
            )
            fields_to_update['started'] = Now()

        activity = WorkerActivity.objects.filter(activity_filter)

        update_count = activity.update(**fields_to_update)

        if not update_count:
            # As no row has been updated the provided data was in conflict with the actual state
            raise ConflictAPIException(
                {
                    '__all__': [
                        'Either this worker activity does not exist, '
                        'is assigned to another process, '
                        f"or updating the state to '{state}' is forbidden."
                    ]
                }
            )

        # Update other activities
        if state == WorkerActivityState.Error.value:
            WorkerActivity.objects.filter(
                process_id=process_id,
                element_id=element_id,
                state=WorkerActivityState.Queued,
            ).update(state=WorkerActivityState.Error)

        return Response(serializer.data)


class WorkerActivityBase(ListAPIView):
    permission_classes = (IsVerified, )
    serializer_class = WorkerStatisticsSerializer
    pagination_class = None
    queryset = WorkerActivity.objects \
        .values('worker_version_id', 'configuration_id') \
        .annotate(
            average_processed_time=Avg(
                F('updated') - F('started'),
                filter=Q(state=WorkerActivityState.Processed, started__isnull=False),
            ),
            max_processed_time=Max(
                F('updated') - F('started'),
                filter=Q(state=WorkerActivityState.Processed, started__isnull=False),
            ),
            min_processed_time=Min(
                F('updated') - F('started'),
                filter=Q(state=WorkerActivityState.Processed, started__isnull=False),
            ),
            median_processed_time=PercentileCont(
                0.5,
                order_by=F('updated') - F('started'),
                filter=Q(state=WorkerActivityState.Processed, started__isnull=False),
                output_field=DurationField(),
            ),
        ) \
        .annotate(
            # Add the counts per state in a separate annotate() call, because otherwise the F('started')
            # on the processed time annotations would be interpreted as being the started activities count,
            # not the started datetime, causing an error: "'<CombinedExpression: F(updated) - F(started)>' is an aggregate"
            **{
                state.value: Count('id', filter=Q(state=state.value))
                for state in WorkerActivityState
            }
        )


@extend_schema_view(
    get=extend_schema(
        operation_id='CorpusWorkersActivity',
        tags=['process']
    )
)
class CorpusWorkersActivity(CorpusACLMixin, WorkerActivityBase):
    """
    Retrieve corpus wise statistics about the activity of all its worker processes.\n
    Requires a **guest** access.
    """

    def filter_queryset(self, queryset):
        corpus = self.get_corpus(self.kwargs['corpus'], role=Role.Guest)
        return queryset.filter(element_id__in=corpus.elements.values('id'))


@extend_schema_view(
    get=extend_schema(
        operation_id='ProcessWorkersActivity',
        tags=['process']
    )
)
class ProcessWorkersActivity(ProcessACLMixin, WorkerActivityBase):
    """
    Retrieve process statistics about the activity of its workers.\n
    Requires a **guest** access.
    """

    def filter_queryset(self, queryset):
        process = get_object_or_404(Process.objects.all(), pk=self.kwargs['pk'])
        access_level = self.process_access_level(process)
        if not access_level:
            raise NotFound
        if access_level < Role.Guest.value:
            raise PermissionDenied(detail='You do not have a guest access to this process.')

        return queryset.filter(process_id=process.id)


@extend_schema_view(
    get=extend_schema(
        operation_id='ListWorkerActivities',
        tags=['process'],
        parameters=[
            OpenApiParameter(
                'process_id',
                type=UUID,
                description='Filter worker activities by process ID',
                required=False,
            ),
            OpenApiParameter(
                'state',
                enum=[state.value for state in WorkerActivityState],
                description='Filter worker activities by state',
                required=False,
            ),
        ]
    )
)
class WorkerActivityList(CorpusACLMixin, ProcessACLMixin, ListAPIView):
    """
    List worker activities in a corpus.

    Requires guest access to the corpus.
    """
    permission_classes = (IsVerified, )
    serializer_class = WorkerActivitySerializer

    @cached_property
    def corpus(self):
        return self.get_corpus(self.kwargs['corpus'])

    def get_queryset(self):
        return WorkerActivity.objects.filter(element__corpus=self.corpus).order_by('id')

    def filter_queryset(self, queryset):
        errors = {}

        process_id = None
        if 'process_id' in self.request.query_params:
            try:
                process_id = UUID(self.request.query_params['process_id'])
            except (TypeError, ValueError):
                errors['process_id'] = ['Process ID should be an UUID.']

        if process_id:
            try:
                # This uses a select_related because the process access level check looks for the corpus
                process = Process.objects.select_related('corpus').get(id=process_id, corpus=self.corpus)
            except Process.DoesNotExist:
                errors['process_id'] = ['This process does not exist.']
            else:
                access_level = self.process_access_level(process)
                if not access_level or access_level < Role.Guest.value:
                    errors['process_id'] = ['You do not have a guest access to this process.']
                queryset = queryset.filter(process=process)

        if 'state' in self.request.query_params:
            try:
                state = WorkerActivityState(self.request.query_params['state'])
            except (TypeError, ValueError):
                errors['state'] = [f"{self.request.query_params['state']!r} is not a valid WorkerActivity state"]
            else:
                queryset = queryset.filter(state=state)

        if errors:
            raise ValidationError(errors)

        return queryset


@extend_schema_view(post=extend_schema(
    operation_id='CreateProcessTemplate',
    responses={201: ProcessSerializer},
    tags=['process']
))
class CreateProcessTemplate(ProcessACLMixin, WorkerACLMixin, CreateAPIView):
    """
    Create a process template all elements for a process with workers.\n\n
    """
    permission_classes = (IsVerified, )
    serializer_class = CreateProcessTemplateSerializer

    def get_queryset(self):
        return Process.objects \
            .prefetch_related(Prefetch('worker_runs', queryset=WorkerRun.objects.select_related('version__worker__type')))

    def check_object_permissions(self, request, template):
        access_level = self.process_access_level(template)
        if not access_level or access_level < Role.Contributor.value:
            raise PermissionDenied(detail='You do not have a contributor access to this process.')

        if not template.worker_runs.all():
            raise ValidationError({'template': 'This process does not have any worker runs.'})

        for worker_run in template.worker_runs.all():
            if not self.has_execution_access(worker_run.version.worker):
                raise PermissionDenied(detail='You do not have an execution access to every worker of this template.')

    def get_serializer_context(self):
        context = super().get_serializer_context()
        context['template'] = self.get_object()
        return context

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        self.perform_create(serializer)
        headers = self.get_success_headers(serializer.data)
        return Response(
            status=status.HTTP_201_CREATED,
            data=ProcessSerializer(serializer.instance, context={'request': self.request}).data, headers=headers
        )


@extend_schema_view(post=extend_schema(
    operation_id='ApplyProcessTemplate',
    parameters=[
        OpenApiParameter(
            'id',
            type=UUID,
            location=OpenApiParameter.PATH,
            description='ID of the template to apply',
            required=True,
        ),
    ],
    responses={200: ProcessSerializer, 403: None},
    tags=['process'],
))
class ApplyProcessTemplate(ProcessACLMixin, WorkerACLMixin, CreateAPIView):
    """
    Apply a template on a given process.

    Requires a **Contributor** access to both the template and the target process. The target process must have no previously applied template.
    """
    permission_classes = (IsVerified, )
    # For OpenAPI type discovery
    serializer_class = ApplyProcessTemplateSerializer

    def get_queryset(self):
        return Process.objects \
            .filter(mode=ProcessMode.Template) \
            .prefetch_related(Prefetch('worker_runs', queryset=WorkerRun.objects.select_related('version__worker__type')))

    def check_object_permissions(self, request, template):
        access_level = self.process_access_level(template)
        if not access_level or access_level < Role.Contributor.value:
            raise PermissionDenied(detail='You do not have a contributor access to this process.')

        for worker_run in template.worker_runs.all():
            if not self.has_execution_access(worker_run.version.worker):
                raise PermissionDenied(detail='You do not have an execution access to this worker.')

    def get_serializer_context(self):
        context = super().get_serializer_context()
        context['template'] = self.get_object()
        return context

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data, user=request.user)
        serializer.is_valid(raise_exception=True)
        self.perform_create(serializer)
        process = serializer.instance
        return Response(
            status=status.HTTP_200_OK,
            data=ProcessSerializer(process, context={'request': self.request}).data
        )


@extend_schema_view(
    delete=extend_schema(
        operation_id='ClearProcess',
        tags=['process'],
        responses={
            204: None,
        },
    )
)
class ClearProcess(ProcessACLMixin, DestroyAPIView):
    """
    Remove all templates and/or workers from a (not started) process
    """
    permission_classes = (IsVerified, )
    queryset = Process.objects.all()

    def check_object_permissions(self, request, process):
        super().check_object_permissions(request, process)

        access_level = self.process_access_level(process)
        required_access = Role.Admin.value

        if not access_level:
            raise NotFound
        if access_level < required_access:
            raise PermissionDenied(detail='You do not have a sufficient access level to this process.')
        if process.workflow_id is not None:
            raise ValidationError({'__all__': ['A process can only be cleared before getting started.']})

    def destroy(self, request, *args, **kwargs):
        process = self.get_object()
        process.clear()
        return Response(status=status.HTTP_204_NO_CONTENT)


@extend_schema(tags=['ingest'])
class BucketList(ListAPIView):
    """
    List the S3 buckets that are available for ingestion.
    """

    permission_classes = (IsVerified, )
    scopes = (Scope.S3Ingest, )
    pagination_class = None
    serializer_class = BucketSerializer

    def get_queryset(self):
        resource = get_ingest_resource()
        # Listing buckets does not use any pagination at all,
        # so this will make a single API query once.
        buckets = list(resource.buckets.all())

        # Add fake Bucket instances if any extra buckets were specified in the settings
        # Note that duplicate bucket names are not checked
        if settings.INGEST_EXTRA_BUCKETS:
            buckets += [
                resource.Bucket(name=name)
                for name in settings.INGEST_EXTRA_BUCKETS
            ]

        # Most S3 API implementations sort buckets by name,
        # but that is not specified anywhere, so we sort manually
        return sorted(buckets, key=lambda bucket: bucket.name)


@extend_schema(tags=['ingest'])
class S3ImportCreate(CreateAPIView):
    """
    Create a new S3 import.

    Requires a verified email and the `s3_ingest` scope.
    """

    permission_classes = (IsVerified, )
    scopes = (Scope.S3Ingest, )
    serializer_class = S3ImportSerializer


@extend_schema_view(
    post=extend_schema(
        operation_id='CreateTrainingProcess',
        tags=['training'],
        responses={201: ProcessSerializer},
    )
)
class StartTraining(CreateAPIView):
    """
    Start a Machine Learning training process
    """
    permission_classes = (IsVerified, )
    serializer_class = StartTrainingSerializer

    def create(self, request, *args, **kwargs):
        # Create and start the training process, then serialize it with the base ProcessSerializer
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        self.perform_create(serializer)
        process = serializer.instance
        return Response(
            status=status.HTTP_201_CREATED,
            data=ProcessSerializer(
                process,
                context=self.get_serializer_context()
            ).data,
        )
