from django.test import override_settings
from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus
from arkindex.images.models import ImageServer
from arkindex.process.models import Process, ProcessMode, WorkerVersion
from arkindex.project.tests import FixtureTestCase
from arkindex.users.models import Scope


class TestCreateS3Import(FixtureTestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.import_worker_version = WorkerVersion.objects.get(worker__slug='file_import')

    def test_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.post(reverse('api:s3-import-create'), {
                'corpus_id': str(self.corpus.id),
                'folder_type': 'volume',
                'element_type': 'page',
                'bucket_name': 'blah',
                'prefix': 'a/b/c',
            })
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_requires_verified(self):
        self.user.user_scopes.create(scope=Scope.S3Ingest)
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.post(reverse('api:s3-import-create'), {
                'corpus_id': str(self.corpus.id),
                'folder_type': 'volume',
                'element_type': 'page',
                'bucket_name': 'blah',
                'prefix': 'a/b/c',
            })
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_requires_scope(self):
        self.assertFalse(self.user.user_scopes.exists())
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(reverse('api:s3-import-create'), {
                'corpus_id': str(self.corpus.id),
                'folder_type': 'volume',
                'element_type': 'page',
                'bucket_name': 'blah',
                'prefix': 'a/b/c',
            })
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_invalid(self):
        self.user.user_scopes.create(scope=Scope.S3Ingest)
        self.client.force_login(self.user)
        private_corpus = Corpus.objects.create(name='nope')
        private_element = private_corpus.elements.create(type=private_corpus.types.create(slug='something'), name="can't touch this")
        cases = [
            (
                {},
                {
                    'bucket_name': ['This field is required.'],
                    'corpus_id': ['This field is required.'],
                }
            ),
            (
                {'corpus_id': '', 'element_id': '', 'bucket_name': '', 'folder_type': '', 'element_type': ''},
                {
                    'bucket_name': ['This field may not be blank.'],
                    'corpus_id': ['This field may not be null.'],
                },
            ),
            (
                {'corpus_id': 'lol', 'element_id': 'lol', 'folder_type': 'a', 'element_type': 'a', 'bucket_name': 'a'},
                {
                    'bucket_name': ['Ensure this field has at least 3 characters.'],
                    'corpus_id': ['“lol” is not a valid UUID.'],
                    'element_id': ['“lol” is not a valid UUID.'],
                },
            ),
            (
                {'corpus_id': str(self.corpus.id), 'bucket_name': 'a' * 100, 'prefix': '/' * 2000},
                {
                    'bucket_name': ['Ensure this field has no more than 63 characters.'],
                    'prefix': ['Ensure this field has no more than 1024 characters.'],
                },
            ),
            (
                {'corpus_id': str(self.corpus.id), 'bucket_name': 'aaa', 'folder_type': 'not_found', 'element_type': 'not_found'},
                {
                    'element_type': [f"Corpus {self.corpus.id} does not have a 'not_found' type, or it is a folder type"],
                    'folder_type': [f"Corpus {self.corpus.id} does not have a 'not_found' type, or it is not a folder type"],
                },
            ),
            (
                {'corpus_id': str(private_corpus.id), 'bucket_name': 'bucket_nem'},
                {'corpus_id': [f'Invalid pk "{private_corpus.id}" - object does not exist.']},
            ),
            (
                {'corpus_id': str(self.corpus.id), 'element_id': str(private_element.id), 'bucket_name': 'lol'},
                {'element_id': [f'Invalid pk "{private_element.id}" - object does not exist.']},
            ),
        ]
        for request, expected in cases:
            with self.subTest(request=request):
                response = self.client.post(reverse('api:s3-import-create'), request)
                self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(response.json(), expected)

    @override_settings(
        PONOS_RECIPE={},
        ARKINDEX_TASKS_IMAGE='arkindex-tasks-image',
        INGEST_IMAGESERVER_ID=999,
        INGEST_S3_ENDPOINT='http://s3.null.teklia.com',
        INGEST_S3_REGION=None,
        INGEST_S3_ACCESS_KEY='🔑',
        INGEST_S3_SECRET_KEY='its-secret-i-wont-tell-you',
    )
    def test_create(self):
        self.user.user_scopes.create(scope=Scope.S3Ingest)
        self.client.force_login(self.user)
        ImageServer.objects.create(id=999, display_name='Ingest image server', url='https://dev.null.teklia.com')
        element = self.corpus.elements.get(name='Volume 1')

        with self.assertNumQueries(28), self.settings(IMPORTS_WORKER_VERSION=str(self.import_worker_version.id)):
            response = self.client.post(reverse('api:s3-import-create'), {
                'corpus_id': str(self.corpus.id),
                'element_id': str(element.id),
                'folder_type': 'volume',
                'element_type': 'page',
                'bucket_name': 'blah',
                'prefix': 'a/b/c',
            })
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        process = Process.objects.get(id=data['id'])

        self.assertDictEqual(data, {'id': str(process.id)})
        self.assertEqual(process.mode, ProcessMode.S3)
        self.assertEqual(process.creator, self.user)
        self.assertEqual(process.corpus_id, self.corpus.id)
        self.assertEqual(process.element_id, element.id)
        self.assertEqual(process.folder_type, self.corpus.types.get(slug='volume'))
        self.assertEqual(process.element_type, self.corpus.types.get(slug='page'))
        self.assertEqual(process.bucket_name, 'blah')
        self.assertEqual(process.prefix, 'a/b/c')

        worker_run = process.worker_runs.get()
        self.assertEqual(worker_run.version, self.import_worker_version)
        self.assertListEqual(worker_run.parents, [])
        self.assertIsNone(worker_run.configuration_id)
        self.assertIsNone(worker_run.model_version_id)

        self.assertIsNotNone(process.workflow_id)
        task = process.workflow.tasks.get()
        self.assertEqual(task.slug, 'import')
        self.assertEqual(task.image, 'arkindex-tasks-image')
        self.assertEqual(task.command, f'python -m arkindex_tasks.import_s3 --corpus={self.corpus.id} '
                                       '--bucket=blah --folder-type=volume --page-type=page '
                                       '--iiif-base-url=https://dev.null.teklia.com --prefix=a/b/c '
                                       f'--element={element.id}')
        self.assertDictEqual(task.env, {
            'ARKINDEX_CORPUS_ID': str(self.corpus.id),
            'ARKINDEX_PROCESS_ID': str(process.id),
            'ARKINDEX_TASK_TOKEN': task.token,
            'ARKINDEX_WORKER_RUN_ID': str(worker_run.id),
            'INGEST_S3_ENDPOINT': 'http://s3.null.teklia.com',
            'INGEST_S3_ACCESS_KEY': '🔑',
            'INGEST_S3_SECRET_KEY': 'its-secret-i-wont-tell-you',
        })

    @override_settings(
        PONOS_RECIPE={},
        ARKINDEX_TASKS_IMAGE='arkindex-tasks-image',
        INGEST_IMAGESERVER_ID=999,
        INGEST_S3_ENDPOINT='http://s3.null.teklia.com',
        INGEST_S3_REGION=None,
        INGEST_S3_ACCESS_KEY='🔑',
        INGEST_S3_SECRET_KEY='its-secret-i-wont-tell-you',
        IMPORTS_WORKER_VERSION='aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa',
    )
    def test_defaults(self):
        self.user.user_scopes.create(scope=Scope.S3Ingest)
        self.client.force_login(self.user)
        self.corpus.types.create(slug='folder', display_name='Folder', folder=True)
        ImageServer.objects.create(id=999, display_name='Ingest image server', url='https://dev.null.teklia.com')

        with self.assertNumQueries(26), self.settings(IMPORTS_WORKER_VERSION=str(self.import_worker_version.id)):
            response = self.client.post(reverse('api:s3-import-create'), {
                'corpus_id': str(self.corpus.id),
                'bucket_name': 'blah',
            })
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        process = Process.objects.get(id=data['id'])

        self.assertDictEqual(data, {'id': str(process.id)})
        self.assertEqual(process.mode, ProcessMode.S3)
        self.assertEqual(process.creator, self.user)
        self.assertEqual(process.corpus_id, self.corpus.id)
        self.assertIsNone(process.element_id)
        self.assertEqual(process.folder_type, self.corpus.types.get(slug='folder'))
        self.assertEqual(process.element_type, self.corpus.types.get(slug='page'))
        self.assertEqual(process.bucket_name, 'blah')
        self.assertIsNone(process.prefix)
        self.assertIsNotNone(process.workflow_id)

        worker_run = process.worker_runs.get()
        self.assertEqual(worker_run.version, self.import_worker_version)
        self.assertListEqual(worker_run.parents, [])
        self.assertIsNone(worker_run.configuration_id)
        self.assertIsNone(worker_run.model_version_id)

        task = process.workflow.tasks.get()
        self.assertEqual(task.slug, 'import')
        self.assertEqual(task.image, 'arkindex-tasks-image')
        self.assertEqual(task.command, f'python -m arkindex_tasks.import_s3 --corpus={self.corpus.id} '
                                       '--bucket=blah --folder-type=folder --page-type=page '
                                       '--iiif-base-url=https://dev.null.teklia.com')
        self.assertDictEqual(task.env, {
            'ARKINDEX_CORPUS_ID': str(self.corpus.id),
            'ARKINDEX_PROCESS_ID': str(process.id),
            'ARKINDEX_TASK_TOKEN': task.token,
            'ARKINDEX_WORKER_RUN_ID': str(worker_run.id),
            'INGEST_S3_ENDPOINT': 'http://s3.null.teklia.com',
            'INGEST_S3_ACCESS_KEY': '🔑',
            'INGEST_S3_SECRET_KEY': 'its-secret-i-wont-tell-you',
        })
