from unittest.mock import patch

from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus
from arkindex.process.models import DataFile, Process, ProcessMode
from arkindex.project.aws import S3FileStatus
from arkindex.project.tests import FixtureAPITestCase


class TestDataFileApi(FixtureAPITestCase):
    """
    Test datafile creation and upload to S3
    """

    def setUp(self):
        super().setUp()
        self.df = DataFile.objects.create(
            name='test.pdf',
            size=42,
            content_type='application/pdf',
            corpus=self.corpus
        )

    def build_file_create_request(self, name=None, size=None, corpus=None, content_type='text/plain'):
        return {
            'name': name or 'some text',
            'size': size or 1,
            'content_type': content_type,
            'corpus': corpus or str(self.corpus.id),
        }

    def test_create_df_requires_login(self):
        request = self.build_file_create_request()
        response = self.client.post(reverse('api:file-create'), request)
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_create_df_corpus_acl(self):
        self.client.force_login(self.user)
        private_corpus = Corpus.objects.create(name='private')
        request = self.build_file_create_request(corpus=str(private_corpus.id))
        response = self.client.post(reverse('api:file-create'), request)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    @patch('arkindex.project.aws.s3.meta.client.generate_presigned_url')
    def test_create_datafile(self, s3_presigned_url_mock):
        self.client.force_login(self.user)
        s3_presigned_url_mock.return_value = 'http://s3/upload_put_url'
        request = self.build_file_create_request()
        response = self.client.post(reverse('api:file-create'), request)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        self.assertIn('id', data)
        df = DataFile.objects.get(id=data['id'])
        self.assertDictEqual(
            data,
            {
                'id': str(df.id),
                'name': df.name,
                'size': df.size,
                'content_type': df.content_type,
                'status': df.status.value,
                's3_url': df.s3_url,
                's3_put_url': df.s3_put_url,
                'corpus': str(df.corpus.id),
            }
        )
        self.assertListEqual(
            [df.name, df.size, df.s3_put_url],
            ['some text', 1, 'http://s3/upload_put_url']
        )

    @patch('arkindex.project.aws.s3.Object')
    def test_check_uploaded_datafile(self, s3_object):
        s3_object().content_length = 42
        s3_object().content_type = 'application/pdf'

        self.client.force_login(self.user)
        response = self.client.patch(
            reverse('api:file-retrieve', kwargs={'pk': self.df.id}),
            {'status': 'checked'},
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.df.refresh_from_db()
        self.assertDictEqual(
            response.json(),
            {
                'id': str(self.df.id),
                'name': self.df.name,
                'content_type': self.df.content_type,
                'size': self.df.size,
                'status': self.df.status.value,
                's3_url': None,
            }
        )
        self.assertEqual(self.df.status, S3FileStatus.Checked)
        self.assertEqual(self.df.content_type, 'application/pdf')
        self.assertEqual(self.df.size, 42)

    @patch('arkindex.project.aws.s3.Object')
    def test_set_error_status_on_failure(self, s3_object):
        s3_object().content_type = 'wrong/type'
        s3_object().content_length = -1
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse('api:file-retrieve', kwargs={'pk': self.df.id}),
            {'status': 'checked'},
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.df.refresh_from_db()
        self.assertEqual(self.df.status, S3FileStatus.Error)

    @patch('arkindex.project.aws.s3.Object')
    def test_check_even_if_already_checked(self, s3_object):
        self.df.status = S3FileStatus.Checked
        self.df.save()
        s3_object().content_type = 'wrong/type'
        s3_object().content_length = -1
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse('api:file-retrieve', kwargs={'pk': self.df.id}),
            {'status': 'checked'},
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.df.refresh_from_db()
        self.assertEqual(self.df.status, S3FileStatus.Error)

    def test_retrieve_datafile_s3_url(self):
        self.client.force_login(self.user)
        response = self.client.get(
            reverse('api:file-retrieve', kwargs={'pk': self.df.id}),
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertIsNone(data['s3_url'])

    @patch('arkindex.project.aws.s3.meta.client.generate_presigned_url')
    def test_retrieve_datafile_s3_url_admin(self, gen_url_mock):
        """
        The S3 GET URL of a DataFile can be retrieved when the user is an instance administrator,
        for backwards compatibility with workers that were using the internal user.
        """
        gen_url_mock.return_value = 'http://somewhere'
        self.client.force_login(self.superuser)

        response = self.client.get(
            reverse('api:file-retrieve', kwargs={'pk': self.df.id}),
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['s3_url'], 'http://somewhere')

    @patch('arkindex.project.aws.s3.meta.client.generate_presigned_url')
    def test_retrieve_datafile_s3_url_task(self, gen_url_mock):
        """
        Ponos task authentication allows access to the S3 URL, to let file imports download files.
        """
        gen_url_mock.return_value = 'http://somewhere'
        process = Process.objects.create(
            corpus=self.corpus,
            creator=self.user,
            mode=ProcessMode.Files,
        )
        process.files.add(self.df)
        process.start()
        task = process.workflow.tasks.get()

        response = self.client.get(
            reverse('api:file-retrieve', kwargs={'pk': self.df.id}),
            HTTP_AUTHORIZATION=f'Ponos {task.token}',
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['s3_url'], 'http://somewhere')

    def test_rename_datafile(self):
        self.client.force_login(self.user)
        self.assertEqual(self.df.name, 'test.pdf')
        response = self.client.patch(
            reverse('api:file-retrieve', kwargs={'pk': self.df.id}),
            {'name': 'taste.pdf'},
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.df.refresh_from_db()
        self.assertEqual(self.df.name, 'taste.pdf')

    def test_create_datafile_large_content_type(self):
        """
        Some Manifests may have really long content type
        Example: https://adore.ugent.be/IIIF/manifests/archive.ugent.be%3ACC0726BC-F6FA-11E7-8F64-7FDC0AD9BE4D
        """
        self.client.force_login(self.user)
        request = self.build_file_create_request(
            name="IIIF manifest",
            content_type='application/ld+json;profile="http://iiif.io/api/presentation/2/context.json";charset=utf-8',
        )
        response = self.client.post(reverse('api:file-create'), request)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        self.assertIn('id', data)
        self.assertEqual(data["content_type"], 'application/ld+json;profile="http://iiif.io/api/presentation/2/context.json";charset=utf-8')
