import uuid

from django.urls import reverse
from rest_framework import status

from arkindex.ponos.models import Workflow
from arkindex.process.models import (
    GitRefType,
    ProcessMode,
    Repository,
    Revision,
    Worker,
    WorkerType,
    WorkerVersion,
    WorkerVersionGPUUsage,
    WorkerVersionState,
)
from arkindex.process.utils import get_default_farm_id
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Right, Role, User

RECIPE = '''
tasks:
  first:
    image: alpine
  second:
    parents:
      - first
    image: hello-world
'''


class TestWorkersWorkerVersions(FixtureAPITestCase):
    """
    Test workers and worker versions endpoints
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.creds = cls.user.credentials.get()
        cls.repo = cls.creds.repos.get(url='http://my_repo.fake/workers/worker')
        # User is the creator of the repository
        cls.repo.memberships.create(user=cls.user, level=Role.Admin.value)
        cls.rev = cls.repo.revisions.get()

        cls.worker_type_dla = WorkerType.objects.get(slug='dla')

        cls.worker_reco = Worker.objects.get(slug='reco')
        cls.worker_dla = Worker.objects.get(slug='dla')
        cls.worker_gpu = Worker.objects.get(slug='worker-gpu')
        cls.worker_generic = Worker.objects.get(slug='generic')
        cls.worker_file_import = Worker.objects.get(slug='file_import')

        cls.rev2 = Revision.objects.create(
            hash='1234',
            message='commit message',
            author='Bob',
            repo=cls.repo,
        )

        process = cls.rev.processes.create(mode=ProcessMode.Repository, creator=cls.user)
        process.start()
        cls.task = process.workflow.tasks.get()

    def setUp(self):
        super().setUp()
        self.version_1 = WorkerVersion.objects.get(worker=self.worker_reco)
        self.version_1.gpu_usage = WorkerVersionGPUUsage.Disabled
        self.version_2 = self.worker_dla.versions.get()

    # Tests on get_query_set for WorkerList
    def test_workers_list_requires_login(self):
        response = self.client.get(reverse('api:workers-list'))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_workers_list(self):
        """
        User is able to list workers on the repository as he has an admin access
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.get(reverse('api:workers-list'))
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.maxDiff = None
        self.assertDictEqual(response.json(), {
            'count': 5,
            'next': None,
            'number': 1,
            'previous': None,
            'results': [
                {
                    'id': str(self.worker_dla.id),
                    'repository_id': str(self.repo.id),
                    'name': 'Document layout analyser',
                    'slug': 'dla',
                    'type': 'dla',
                },
                {
                    'id': str(self.worker_file_import.id),
                    'repository_id': str(self.repo.id),
                    'name': 'File import',
                    'slug': 'file_import',
                    'type': 'import',
                },
                {
                    'id': str(self.worker_generic.id),
                    'repository_id': str(self.repo.id),
                    'name': 'Generic worker with a Model',
                    'slug': 'generic',
                    'type': 'recognizer',
                },
                {
                    'id': str(self.worker_reco.id),
                    'repository_id': str(self.repo.id),
                    'name': 'Recognizer',
                    'slug': 'reco',
                    'type': 'recognizer',
                },
                {
                    'id': str(self.worker_gpu.id),
                    'repository_id': str(self.repo.id),
                    'name': 'Worker requiring a GPU',
                    'slug': 'worker-gpu',
                    'type': 'worker',
                },
            ]
        })

    def test_workers_list_filter_type_slug(self):
        """
        User is able to filter workers on the repository by worker type slug
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.get(reverse('api:workers-list'), {'type': 'dla'})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'count': 1,
            'next': None,
            'number': 1,
            'previous': None,
            'results': [
                {
                    'id': str(self.worker_dla.id),
                    'repository_id': str(self.repo.id),
                    'name': 'Document layout analyser',
                    'slug': 'dla',
                    'type': 'dla',
                }
            ]
        })

    def test_workers_list_filter_invalid_type_slug(self):
        """
        Raise when trying to filter workers with an invalid type slug
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.get(reverse('api:workers-list'), {'type': 'invalid-slug'})
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"type": ['No registered worker type with that slug.']})

    def test_workers_list_filter_type_id(self):
        """
        User is able to filter workers on the repository by worker type id
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.get(reverse('api:workers-list'), {'type': self.worker_type_dla.id})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'count': 1,
            'next': None,
            'number': 1,
            'previous': None,
            'results': [
                {
                    'id': str(self.worker_dla.id),
                    'repository_id': str(self.repo.id),
                    'name': 'Document layout analyser',
                    'slug': 'dla',
                    'type': 'dla',
                }
            ]
        })

    def test_workers_list_requires_contributor(self):
        """
        User is not able to list workers with a guest access
        """
        self.repo.memberships.filter(user=self.user).update(level=Role.Guest.value)
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.get(reverse('api:workers-list'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'count': 0,
            'next': None,
            'number': 1,
            'previous': None,
            'results': []
        })

    def test_workers_list_public_attribute(self):
        """
        Any authenticated user is able to list a worker with the public attribute
        """
        user = User.objects.create_user('user42@test.fr', 'abcd')
        user.verified_email = True
        user.save()
        self.worker_reco.public = True
        self.worker_reco.save()
        self.client.force_login(user)
        with self.assertNumQueries(7):
            response = self.client.get(reverse('api:workers-list'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'count': 1,
            'next': None,
            'number': 1,
            'previous': None,
            'results': [{
                'id': str(self.worker_reco.id),
                'repository_id': str(self.repo.id),
                'name': 'Recognizer',
                'slug': 'reco',
                'type': 'recognizer',
            }]
        })

    def test_workers_list_name_filter(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.get(reverse('api:workers-list'), {'name': 'layout'})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'count': 1,
            'next': None,
            'number': 1,
            'previous': None,
            'results': [
                {
                    'id': str(self.worker_dla.id),
                    'repository_id': str(self.worker_dla.repository.id),
                    'name': 'Document layout analyser',
                    'slug': 'dla',
                    'type': 'dla',
                }
            ]
        })

    def test_workers_list_filter_repository(self):
        repo2 = Repository.objects.create(
            url='http://gitlab/repo2',
            hook_token='hook-token2',
            credentials=self.creds,
        )
        worker_2 = Worker.objects.create(
            repository=repo2,
            name='Worker 2',
            slug='worker_2',
            type=self.worker_type_dla
        )
        worker_2.memberships.create(user=self.user, level=Role.Contributor.value)

        self.client.force_login(self.user)
        response = self.client.get(reverse('api:workers-list'), {'repository_id': str(repo2.id)})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'count': 1,
            'next': None,
            'number': 1,
            'previous': None,
            'results': [
                {
                    'id': str(worker_2.id),
                    'repository_id': str(repo2.id),
                    'name': 'Worker 2',
                    'slug': 'worker_2',
                    'type': self.worker_type_dla.slug
                }
            ]
        })

    def test_workers_retrieve_required_login(self):
        response = self.client.get(reverse('api:worker-retrieve', kwargs={'pk': str(self.worker_reco.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_workers_retrieve(self):
        """
        A user is able to retrieve a worker if he has a guest access on it or its repository
        """
        repo2 = Repository.objects.create(
            url='http://gitlab/repo2',
            hook_token='hook-token2',
            credentials=self.creds,
        )
        worker_2 = repo2.workers.create(name='Worker 2', slug='worker_2', type=self.worker_type_dla)
        repo2.memberships.create(user=self.user, level=Role.Guest.value)

        self.client.force_login(self.user)
        with self.assertNumQueries(9):
            response = self.client.get(reverse('api:worker-retrieve', kwargs={'pk': str(worker_2.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'id': str(worker_2.id),
            'repository_id': str(repo2.id),
            'name': worker_2.name,
            'slug': worker_2.slug,
            'type': worker_2.type.slug,
        })

    def test_worker_create_requires_login(self):
        response = self.client.post(reverse('api:workers-list'))
        with self.assertNumQueries(0):
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(response.json(), {'detail': 'Authentication credentials were not provided.'})

    def test_workers_retrieve_no_rights(self):
        new_user = User.objects.create(email='new@test.fr', verified_email=True)
        self.client.force_login(new_user)
        with self.assertNumQueries(9):
            response = self.client.get(reverse('api:worker-retrieve', kwargs={'pk': str(self.worker_reco.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(response.json(), {'detail': 'You do not have a guest access to this worker.'})

    def test_worker_create_repository_not_uid(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse('api:workers-list'),
                data={
                    'repository_id': 'repo',
                    'name': 'Worker post',
                    'slug': 'worker_post',
                    'type': self.worker_type_dla.slug
                }
            )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'repository_id': ['Must be a valid UUID.']})

    def test_workers_create_unexisting_repository(self):
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse('api:workers-list'),
                data={
                    'repository_id': uuid.uuid4(),
                    'name': 'Worker post',
                    'slug': 'worker_post',
                    'type': self.worker_type_dla.slug
                },
                HTTP_AUTHORIZATION=f'Ponos {self.task.token}',
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(response.json(), {'detail': 'Not found.'})

    def test_workers_create_required_fields(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.post(reverse('api:workers-list'))

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'repository_id': ['This field is required.'],
            'name': ['This field is required.'],
            'slug': ['This field is required.'],
            'type': ['This field is required.']
        })

    def test_workers_create_worker(self):
        """
        User with an admin right on a repository is able to create a worker
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(11):
            response = self.client.post(
                reverse('api:workers-list'),
                data={
                    'repository_id': str(self.repo.id),
                    'name': 'Worker post',
                    'slug': 'worker_post',
                    'type': self.worker_type_dla.slug
                }
            )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        self.assertNotEqual(data['id'], str(self.worker_reco.id))
        self.assertEqual(data['name'], 'Worker post')
        self.assertEqual(data['slug'], 'worker_post')
        self.assertEqual(data['type'], self.worker_type_dla.slug)

    def test_workers_create_ponos_task_rights(self):
        """
        Ponos tasks are not allowed to create workers when they do not have a right on the repository
        """
        Right.objects.filter(
            content_id__in=(self.repo.id, *self.repo.workers.values_list('id', flat=True)),
            user=self.user
        ).delete()
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse('api:workers-list'),
                data={
                    'repository_id': str(self.repo.id),
                    'name': 'Worker post',
                    'slug': 'worker_post',
                    'type': self.worker_type_dla.slug
                },
                HTTP_AUTHORIZATION=f'Ponos {self.task.token}',
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_workers_post_return_existing_worker(self):
        self.client.force_login(self.user)
        # A worker named Worker 1 already exists with the same slug on this repo
        with self.assertNumQueries(9):
            response = self.client.post(
                reverse('api:workers-list'),
                data={
                    'repository_id': str(self.repo.id),
                    'name': 'Worker Test',
                    'slug': 'reco',
                    'type': self.worker_type_dla.slug
                }
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['id'], str(self.worker_reco.id))
        self.assertEqual(data['name'], 'Recognizer')
        self.assertEqual(data['slug'], 'reco')
        self.assertEqual(data['type'], 'recognizer')

    def test_workers_create_worker_new_type(self):
        """
        User with an admin right on a repository is able to create a worker with a non-existing worker type
        """
        self.client.force_login(self.user)
        new_worker_type_slug = 'newType'
        self.assertFalse(WorkerType.objects.filter(slug=new_worker_type_slug).exists())
        with self.assertNumQueries(14):
            response = self.client.post(
                reverse('api:workers-list'),
                data={
                    'repository_id': str(self.repo.id),
                    'name': 'Worker post',
                    'slug': 'worker_post',
                    'type': 'newType'
                }
            )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        self.assertTrue(WorkerType.objects.filter(slug=new_worker_type_slug).exists())
        self.assertNotEqual(data['id'], str(self.worker_reco.id))
        self.assertEqual(data['name'], 'Worker post')
        self.assertEqual(data['slug'], 'worker_post')
        self.assertEqual(data['type'], 'newType')

    # Tests on get_query_set for WorkerVersionList
    def test_versions_list_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.get(reverse('api:worker-versions', kwargs={'pk': str(self.worker_reco.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_versions_list(self):
        self.client.force_login(self.user)
        last_revision = self.repo.revisions.create(hash='42', message='A second version', author='Michel')
        last_version = self.worker_reco.versions.create(
            revision=last_revision,
            configuration={'a': 'bc'},
            state=WorkerVersionState.Error,
            docker_image=self.version_1.docker_image
        )
        with self.assertNumQueries(14):
            response = self.client.get(reverse('api:worker-versions', kwargs={'pk': str(self.worker_reco.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['count'], 2)
        # The revision created in this test has a recent timestamp so it is displayed first
        first_version = data['results'][0]
        revision = first_version.pop('revision')
        self.assertEqual(revision['id'], str(last_revision.id))
        self.assertEqual(revision['message'], 'A second version')
        self.assertDictEqual(first_version, {
            'id': str(last_version.id),
            'configuration': {'a': 'bc'},
            'docker_image': str(self.version_1.docker_image.id),
            'docker_image_iid': None,
            'docker_image_name': f'my_repo.fake/workers/worker/reco:{last_version.id}',
            'state': 'error',
            'gpu_usage': 'disabled',
            'model_usage': False,
            'worker': {
                'id': str(self.worker_reco.id),
                'name': self.worker_reco.name,
                'type': self.worker_reco.type.slug,
                'slug': self.worker_reco.slug,
            }
        })

    def test_workers_versions_list_requires_contributor(self):
        """
        User is not able to list worker versions with a guest access
        """
        self.repo.memberships.filter(user=self.user).update(level=Role.Guest.value)
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.get(reverse('api:workers-list'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'count': 0,
            'next': None,
            'number': 1,
            'previous': None,
            'results': []
        })

    def test_versions_list_public_worker(self):
        """
        Any authenticated user is able to list versions for public workers
        """
        user = User.objects.create_user('user42@test.fr', 'abcd')
        user.verified_email = True
        user.save()
        self.worker_reco.public = True
        self.worker_reco.save()
        self.client.force_login(user)
        with self.assertNumQueries(9):
            response = self.client.get(reverse('api:worker-versions', kwargs={'pk': str(self.worker_reco.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['count'], 1)
        self.assertEqual(data['results'][0]['id'], str(self.worker_reco.versions.first().id))

    def test_versions_list_filter_worker(self):
        self.client.force_login(self.user)
        worker_2 = Worker.objects.create(
            repository=self.repo,
            name='Worker 2',
            slug='worker_2',
            type=self.worker_type_dla
        )
        version_2 = WorkerVersion.objects.create(
            worker=worker_2,
            revision=self.rev,
            configuration={"test": "test2"},
            gpu_usage=WorkerVersionGPUUsage.Disabled
        )

        with self.assertNumQueries(14):
            response = self.client.get(reverse('api:worker-versions', kwargs={'pk': str(worker_2.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 1)
        version = data['results'][0]
        self.assertEqual(version['id'], str(version_2.id))

    def test_versions_list_simple_mode(self):
        """
        With the mode attribute set to simple, worker versions older than master/main
        or with no tag are excluded from the results
        """
        # An old branch revision
        non_tagged_rev = self.repo.revisions.create(hash='001', message='random', author='A1ice')
        non_tagged_rev.refs.create(name='a-second-branch', type=GitRefType.Branch, repository=self.repo)
        # An old tagged branch revision
        tagged_rev = self.repo.revisions.create(hash='002', message='v0.1', author='Bob')
        tagged_rev.refs.create(name='a-branch', type=GitRefType.Branch, repository=self.repo)
        tagged_rev.refs.create(name='0.1', type=GitRefType.Tag, repository=self.repo)
        # Master branch revision
        master_rev = self.repo.revisions.create(hash='003', message='uptodate', author='B0b')
        master_rev.refs.create(name='master', type=GitRefType.Branch, repository=self.repo)
        # Main branch revision
        main_rev = self.repo.revisions.create(hash='004', message='uptodate', author='B0b')
        main_rev.refs.create(name='main', type=GitRefType.Branch, repository=self.repo)
        # A recent revision with no reference
        recent_rev = self.repo.revisions.create(hash='005', message='uptodate', author='AL1C3')
        WorkerVersion.objects.bulk_create([
            WorkerVersion(
                worker=self.worker_reco,
                revision=rev,
                configuration={'test': '42'}
            ) for rev in (non_tagged_rev, tagged_rev, master_rev, main_rev, recent_rev)
        ])

        # Complete mode
        with self.assertNumQueries(13):
            response = self.client.get(
                reverse('api:worker-versions', kwargs={'pk': str(self.worker_reco.id)}),
                {'mode': 'complete'},
                HTTP_AUTHORIZATION=f'Ponos {self.task.token}',
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()['count'], 6)

        # Simple mode
        with self.assertNumQueries(10):
            response = self.client.get(
                reverse('api:worker-versions', kwargs={'pk': str(self.worker_reco.id)}),
                {'mode': 'simple'},
                HTTP_AUTHORIZATION=f'Ponos {self.task.token}',
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['count'], 3)
        self.assertCountEqual(
            [version['revision']['id'] for version in data['results']],
            [str(tagged_rev.id), str(master_rev.id), str(main_rev.id)]
        )

    def test_versions_post_requires_task(self):
        """
        Only Ponos tasks are able to create new versions
        """
        users = [None, self.user, self.superuser]
        for user in users:
            if user:
                self.client.force_login(user)
            response = self.client.post(
                reverse('api:worker-versions', kwargs={'pk': str(self.worker_reco.id)}),
                data={'revision_id': str(self.rev2.id), 'configuration': {"test": "test2"}}, format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_versions_post_non_existing_worker(self):
        response = self.client.post(
            reverse('api:worker-versions', kwargs={'pk': '12341234-1234-1234-1234-123412341234'}),
            data={'revision_id': str(self.rev2.id), 'configuration': {"test": "test2"}},
            format='json',
            HTTP_AUTHORIZATION=f'Ponos {self.task.token}',
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(response.json(), {'detail': 'Not found.'})

    def test_versions_post_available_requires_docker_image(self):
        response = self.client.post(
            reverse('api:worker-versions', kwargs={'pk': str(self.worker_reco.id)}),
            data={
                'revision_id': str(self.rev2.id),
                'configuration': {"test": "test2"},
                'state': 'available',
            },
            HTTP_AUTHORIZATION=f'Ponos {self.task.token}',
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'non_field_errors': [
                'A WorkerVersion must have a Docker image artifact and a Docker image ID set when it is available.'
            ]
        })

    def test_versions_post_return_existing_worker_version(self):
        # A worker version already exists for this worker and this revision
        response = self.client.post(
            reverse('api:worker-versions', kwargs={'pk': str(self.worker_reco.id)}),
            data={'revision_id': str(self.rev.id), 'configuration': {"test": "test1"}},
            format='json',
            HTTP_AUTHORIZATION=f'Ponos {self.task.token}',
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['id'], str(self.version_1.id))
        self.assertEqual(data['configuration'], {"test": 42})
        self.assertEqual(data['revision']['id'], str(self.rev.id))
        self.assertEqual(data['state'], 'available')

    def test_versions_post_empty(self):
        response = self.client.post(
            reverse('api:worker-versions', kwargs={'pk': str(self.worker_reco.id)}),
            HTTP_AUTHORIZATION=f'Ponos {self.task.token}',
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'revision_id': ['This field is required.'],
            'configuration': ['This field is required.']
        })

    def test_versions_post_unrelated_revision(self):
        """
        It is not possible to create a version between a worker and a revision of two different repositories
        """
        rev = Revision.objects.create(
            hash='1337',
            message='A worker import I am admin on',
            author='Yann',
            repo=self.creds.repos.get(url='http://gitlab/repo')
        )
        response = self.client.post(
            reverse('api:worker-versions', kwargs={'pk': str(self.worker_reco.id)}),
            data={'revision_id': str(rev.id), 'configuration': {"test": "test2"}},
            format='json',
            HTTP_AUTHORIZATION=f'Ponos {self.task.token}',
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'revision': ['The revision must be part of the same repository as the worker.']
        })

    def test_versions_post_create_worker_version(self):
        response = self.client.post(
            reverse('api:worker-versions', kwargs={'pk': str(self.worker_reco.id)}),
            data={'revision_id': str(self.rev2.id), 'configuration': {"test": "test2"}, 'model_usage': True},
            format='json',
            HTTP_AUTHORIZATION=f'Ponos {self.task.token}',
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        data = response.json()
        self.assertNotEqual(data['id'], str(self.version_1.id))
        self.assertEqual(data['configuration'], {"test": "test2"})
        self.assertEqual(data['revision']['id'], str(self.rev2.id))
        self.assertEqual(data['state'], 'created')
        self.assertEqual(data['gpu_usage'], 'disabled')
        self.assertEqual(data['model_usage'], True)

    def test_versions_post_create_worker_version_wrong_gpu_usage(self):
        response = self.client.post(
            reverse('api:worker-versions', kwargs={'pk': str(self.worker_reco.id)}),
            data={'revision_id': str(self.rev2.id), 'configuration': {"test": "test2"}, 'gpu_usage': 'not_supported'},
            format='json',
            HTTP_AUTHORIZATION=f'Ponos {self.task.token}',
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_no_user_configuration_ok(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "revision_id": str(self.rev2.id),
                "configuration": {"beep": "boop"},
                "gpu_usage": "disabled",
            },
            format="json",
            HTTP_AUTHORIZATION=f'Ponos {self.task.token}',
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_valid_user_configuration(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "revision_id": str(self.rev2.id),
                "configuration": {
                    "user_configuration": {
                        "demo_integer": {"title": "Demo Integer", "type": "int", "required": True, "default": 1},
                        "demo_boolean": {"title": "Demo Boolean", "type": "bool", "required": False, "default": True},
                    }
                },
                "gpu_usage": "disabled",
            },
            format="json",
            HTTP_AUTHORIZATION=f'Ponos {self.task.token}',
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.json()['configuration'], {
            "user_configuration": {
                "demo_integer": {
                    "title": "Demo Integer",
                    "type": "int",
                    "required": True,
                    "default": 1
                },
                "demo_boolean": {
                    "title": "Demo Boolean",
                    "type": "bool",
                    "required": False,
                    "default": True
                }
            }
        })

    def test_valid_user_configuration_dict(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "revision_id": str(self.rev2.id),
                "configuration": {
                    "user_configuration": {
                        "demo_dict": {"title": "Demo Dict", "type": "dict", "required": True, "default": {"a": "b", "c": "d"}},
                    }
                },
                "gpu_usage": "disabled",
            },
            format="json",
            HTTP_AUTHORIZATION=f'Ponos {self.task.token}',
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.json()['configuration'], {
            "user_configuration": {
                "demo_dict": {
                    "title": "Demo Dict",
                    "type": "dict",
                    "required": True,
                    "default": {"a": "b", "c": "d"}
                }
            }
        })

    def test_user_configuration_dict_strings_only(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "revision_id": str(self.rev2.id),
                "configuration": {
                    "user_configuration": {
                        "demo_dict": {"title": "Demo Dict", "type": "dict", "required": True, "default": {"a": ["12", "13"], "c": "d"}},
                    }
                },
                "gpu_usage": "disabled",
            },
            format="json",
            HTTP_AUTHORIZATION=f'Ponos {self.task.token}',
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_valid_user_configuration_enum(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "revision_id": str(self.rev2.id),
                "configuration": {
                    "user_configuration": {
                        "demo_choice": {"title": "Decisions", "type": "enum", "required": True, "default": 1, "choices": [1, 2, 3]}
                    }
                },
                "gpu_usage": "disabled",
            },
            format="json",
            HTTP_AUTHORIZATION=f'Ponos {self.task.token}',
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.json()['configuration'], {
            "user_configuration": {
                "demo_choice": {
                    "title": "Decisions",
                    "type": "enum",
                    "required": True,
                    "default": 1,
                    "choices": [1, 2, 3]
                }
            }
        })

    def test_valid_user_configuration_list(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "revision_id": str(self.rev2.id),
                "configuration": {
                    "user_configuration": {
                        "demo_list": {"title": "Demo List", "type": "list", "required": True, "subtype": "int", "default": [1, 2, 3, 4]},
                        "boolean_list": {"title": "It's a list of booleans", "type": "list", "required": False, "subtype": "bool", "default": [True, False, False]}
                    }
                },
                "gpu_usage": "disabled",
            },
            format="json",
            HTTP_AUTHORIZATION=f'Ponos {self.task.token}',
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.json()['configuration'], {
            "user_configuration": {
                "demo_list": {
                    "title": "Demo List",
                    "type": "list",
                    "subtype": "int",
                    "required": True,
                    "default": [1, 2, 3, 4]
                },
                "boolean_list": {
                    "title": "It's a list of booleans",
                    "type": "list",
                    "subtype": "bool",
                    "required": False,
                    "default": [True, False, False]
                }
            }
        })

    def test_invalid_user_configuration_list_requires_subtype(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "revision_id": str(self.rev2.id),
                "configuration": {
                    "user_configuration": {
                        "demo_list": {"title": "Demo List", "type": "list", "required": True, "default": [1, 2, 3, 4]},
                    }
                },
                "gpu_usage": "disabled",
            },
            format="json",
            HTTP_AUTHORIZATION=f'Ponos {self.task.token}',
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "configuration": {
                "user_configuration": [{
                    "demo_list": {
                        "subtype": ["The \"subtype\" field must be set for \"list\" type properties."]
                    }
                }]
            }
        })

    def test_invalid_user_configuration_list_wrong_default(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "revision_id": str(self.rev2.id),
                "configuration": {
                    "user_configuration": {
                        "demo_list": {"title": "Demo List", "type": "list", "required": True, "subtype": "int", "default": 12},
                    }
                },
                "gpu_usage": "disabled",
            },
            format="json",
            HTTP_AUTHORIZATION=f'Ponos {self.task.token}',
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "configuration": {
                "user_configuration": [{
                    "demo_list": {
                        "default": ["Default value is not of type list."]
                    }
                }]
            }
        })

    def test_invalid_user_configuration_list_wrong_subtype(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "revision_id": str(self.rev2.id),
                "configuration": {
                    "user_configuration": {
                        "demo_list": {"title": "Demo List", "type": "list", "required": True, "subtype": "dict", "default": [1, 2, 3, 4]},
                    }
                },
                "gpu_usage": "disabled",
            },
            format="json",
            HTTP_AUTHORIZATION=f'Ponos {self.task.token}',
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "configuration": {
                "user_configuration": [{
                    "demo_list": {
                        "subtype": ["Subtype can only be int, float, bool or string."]
                    }
                }]
            }
        })

    def test_invalid_user_configuration_list_wrong_default_subtype(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "revision_id": str(self.rev2.id),
                "configuration": {
                    "user_configuration": {
                        "demo_list": {"title": "Demo List", "type": "list", "required": True, "subtype": "int", "default": [1, 2, "three", 4]},
                    }
                },
                "gpu_usage": "disabled",
            },
            format="json",
            HTTP_AUTHORIZATION=f'Ponos {self.task.token}',
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "configuration": {
                "user_configuration": [{
                    "demo_list": {
                        "default": ["All items in the default value must be of type int."]
                    }
                }]
            }
        })

    def test_invalid_user_configuration_not_list_choices(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "revision_id": str(self.rev2.id),
                "configuration": {
                    "user_configuration": {
                        "demo_choice": {"title": "Decisions", "type": "enum", "required": True, "default": 1, "choices": "eeee"}
                    }
                },
                "gpu_usage": "disabled",
            },
            format="json",
            HTTP_AUTHORIZATION=f'Ponos {self.task.token}',
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "configuration": {
                "user_configuration": [{
                    "demo_choice": {
                        "choices": ["Expected a list of items but got type \"str\"."]
                    }
                }]
            }
        })

    def test_invalid_user_configuration_not_dict(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "revision_id": str(self.rev2.id),
                "configuration": {
                    "user_configuration": "non"
                },
                "gpu_usage": "disabled",
            },
            format="json",
            HTTP_AUTHORIZATION=f'Ponos {self.task.token}',
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"configuration": {"user_configuration": [["Expected a dictionary of items but got type \"str\"."]]}})

    def test_invalid_user_configuration_item_not_dict(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "revision_id": str(self.rev2.id),
                "configuration": {
                    "user_configuration": {
                        "secrets": ["aaaaaa"]
                    }
                },
                "gpu_usage": "disabled",
            },
            format="json",
            HTTP_AUTHORIZATION=f'Ponos {self.task.token}',
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"configuration": {
            "user_configuration": [{
                "secrets":
                {'__all__': ['User configuration field definitions should be of type dict, not list.']}
            }]
        }})

    def test_invalid_user_configuration_wrong_field_type(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "revision_id": str(self.rev2.id),
                "configuration": {
                    "user_configuration": {
                        "something": {
                            "title": "some thing",
                            "type": "uh oh",
                            "required": 2
                        }
                    }
                },
                "gpu_usage": "disabled",
            },
            format="json",
            HTTP_AUTHORIZATION=f'Ponos {self.task.token}',
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "configuration": {
                "user_configuration": [{
                    "something": {
                        "required": ["Must be a valid boolean."],
                        "type": ["Value is not of type UserConfigurationFieldType"]
                    }
                }]
            }
        })

    def test_invalid_user_configuration_wrong_default_type(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "revision_id": str(self.rev2.id),
                "configuration": {
                    "user_configuration": {
                        "one_float": {
                            "title": "a float",
                            "type": "float",
                            "default": "bonjour",
                            "required": True
                        }
                    }
                },
                "gpu_usage": "disabled",
            },
            format="json",
            HTTP_AUTHORIZATION=f'Ponos {self.task.token}',
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "configuration": {
                "user_configuration": [{
                    "one_float": {
                        "default": ["Default value is not of type float."]
                    }
                }]
            }
        })

    def test_invalid_user_configuration_choices_no_enum(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "revision_id": str(self.rev2.id),
                "configuration": {
                    "user_configuration": {
                        "something": {
                            "title": "some thing",
                            "type": "int",
                            "required": False,
                            "choices": [1, 2, 3]
                        }
                    }
                },
                "gpu_usage": "disabled",
            },
            format="json",
            HTTP_AUTHORIZATION=f'Ponos {self.task.token}',
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "configuration": {
                "user_configuration": [{
                    "something": {
                        "choices": ["The \"choices\" field can only be set for an \"enum\" type property."]
                    }
                }]
            }
        })

    def test_invalid_user_configuration_missing_key(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "revision_id": str(self.rev2.id),
                "configuration": {
                    "user_configuration": {
                        "demo_integer": {"type": "int", "required": True, "default": 1}
                    }
                },
                "gpu_usage": "disabled",
            },
            format="json",
            HTTP_AUTHORIZATION=f'Ponos {self.task.token}',
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(
            response.json(),
            {
                "configuration": {
                    "user_configuration": [{
                        "demo_integer": {
                            "title": ["This field is required."]
                        }
                    }]
                }
            }
        )

    def test_invalid_user_configuration_invalid_key(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "revision_id": str(self.rev2.id),
                "configuration": {
                    "user_configuration": {
                        "demo_integer": {
                            "title": "an integer",
                            "type": "int",
                            "required": True,
                            "default": 1,
                            "some_key": "oh no",
                        }
                    }
                },
                "gpu_usage": "disabled",
            },
            format="json",
            HTTP_AUTHORIZATION=f'Ponos {self.task.token}',
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "configuration": {
                "user_configuration": [
                    {
                        "demo_integer": {
                            "some_key": ["Configurable properties can only be defined using the following keys: title, type, required, default, subtype, choices."]
                        }
                    }
                ]
            }
        })

    def test_invalid_user_configuration_default_value(self):
        cases = [
            ({"type": "int", "default": False}, 'int'),
            ({"type": "int", "default": True}, 'int'),
            ({"type": "float", "default": False}, 'float'),
            ({"type": "float", "default": True}, 'float'),
            ({"type": "bool", "default": 0}, 'bool'),
            ({"type": "bool", "default": 1}, 'bool'),
            ({"type": "string", "default": 1}, 'string'),
            ({"type": "dict", "default": ["a", "b"]}, 'dict')
        ]
        for params, expected in cases:
            with self.subTest(**params):
                response = self.client.post(
                    reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
                    data={
                        "revision_id": str(self.rev2.id),
                        "configuration": {
                            "user_configuration": {
                                "param": {"title": 'param', **params}
                            }
                        },
                    },
                    format="json",
                    HTTP_AUTHORIZATION=f'Ponos {self.task.token}',
                )
                self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertEqual(response.json(), {"configuration": {"user_configuration": [{'param': {'default': [f'Default value is not of type {expected}.']}}]}})

    def test_retrieve_version_invalid_id(self):
        self.client.force_login(self.user)
        response = self.client.get(
            reverse('api:version-retrieve', kwargs={'pk': '12341234-1234-1234-1234-123412341234'})
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_retrieve_version_no_login(self):
        """
        Worker versions may be retrieved with no authentication in order
        to see the version on public resources
        """
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:version-retrieve', kwargs={'pk': str(self.version_1.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        revision = data.pop('revision')
        self.assertEqual(revision['id'], str(self.rev.id))
        self.assertDictEqual(data, {
            'id': str(self.version_1.id),
            'configuration': {"test": 42},
            'docker_image': str(self.version_1.docker_image.id),
            'docker_image_iid': None,
            'docker_image_name': f'my_repo.fake/workers/worker/reco:{self.version_1.id}',
            'state': 'available',
            'gpu_usage': 'disabled',
            'model_usage': False,
            'worker': {
                'id': str(self.worker_reco.id),
                'name': self.worker_reco.name,
                'type': self.worker_reco.type.slug,
                'slug': self.worker_reco.slug,
            }
        })

    def test_retrieve_version(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:version-retrieve', kwargs={'pk': str(self.version_1.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        revision = data.pop('revision')
        self.assertEqual(revision['id'], str(self.rev.id))
        self.assertDictEqual(data, {
            'id': str(self.version_1.id),
            'configuration': {"test": 42},
            'docker_image': str(self.version_1.docker_image.id),
            'docker_image_iid': None,
            'docker_image_name': f'my_repo.fake/workers/worker/reco:{self.version_1.id}',
            'state': 'available',
            'gpu_usage': 'disabled',
            'model_usage': False,
            'worker': {
                'id': str(self.worker_reco.id),
                'name': self.worker_reco.name,
                'type': self.worker_reco.type.slug,
                'slug': self.worker_reco.slug,
            }
        })

    def test_update_version_requires_login(self):
        response = self.client.patch(
            reverse('api:version-retrieve', kwargs={'pk': str(self.version_1.id)}),
            data={
                'configuration': {"test": "test2"},
                'docker_image': '12341234-1234-1234-1234-123412341234',
                'state': 'error',
                'gpu_usage': 'disabled',
                'model_usage': False,
            }, format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(response.json(), {'detail': 'Authentication credentials were not provided.'})

    def test_update_version_invalid_id(self):
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse('api:version-retrieve', kwargs={'pk': '12341234-1234-1234-1234-123412341234'}),
            data={
                'configuration': {"test": "test2"},
                'docker_image': '12341234-1234-1234-1234-123412341234',
                'state': 'error',
                'gpu_usage': 'disabled',
                'model_usage': False,
            }, format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_update_worker_version_non_task(self):
        """
        Only Ponos tasks are allowed to update a worker version
        """
        docker_image = self.version_1.docker_image
        self.version_1.state = WorkerVersionState.Created
        self.version_1.docker_image = None
        self.version_1.save()
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse('api:version-retrieve', kwargs={'pk': str(self.version_1.id)}),
            data={
                'configuration': {"test": "test2"},
                'docker_image': str(docker_image.id),
                'state': 'error',
                'gpu_usage': 'disabled',
                'model_usage': False,
            }, format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(response.json(), {'detail': 'Only Ponos tasks are allowed to update a worker version.'})

    def test_update_worker_version(self):
        """
        Update worker version artifact, configuration and state
        """
        docker_image = self.version_1.docker_image
        self.version_1.state = WorkerVersionState.Created
        self.version_1.docker_image = None
        self.version_1.save()

        response = self.client.patch(
            reverse('api:version-retrieve', kwargs={'pk': str(self.version_1.id)}),
            data={
                'configuration': {"test": "test2"},
                'docker_image': str(docker_image.id),
                'state': 'error',
            },
            format='json',
            HTTP_AUTHORIZATION=f'Ponos {self.task.token}',
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['id'], str(self.version_1.id))
        self.version_1.refresh_from_db()
        self.assertEqual(data['configuration'], {"test": "test2"})
        self.assertEqual(data['docker_image'], str(docker_image.id))
        self.assertIsNone(data['docker_image_iid'])
        self.assertEqual(data['state'], 'error')
        self.assertEqual(data['gpu_usage'], 'disabled')

    def test_cannot_update_worker_version_revision(self):
        self.version_1.state = WorkerVersionState.Created
        self.version_1.docker_image = None
        self.version_1.save()
        with self.assertNumQueries(8):
            response = self.client.patch(
                reverse('api:version-retrieve', kwargs={'pk': str(self.version_1.id)}),
                data={
                    'revision_id': str(self.rev2.id)
                },
                format='json',
                HTTP_AUTHORIZATION=f'Ponos {self.task.token}',
            )
            # revision_id just gets ignored
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['id'], str(self.version_1.id))
        self.version_1.refresh_from_db()
        self.assertEqual(data['revision']['id'], str(self.rev.id))

    def test_update_version_valid(self):
        self.version_1.state = WorkerVersionState.Created
        self.version_1.docker_image = None
        self.version_1.save()

        workflow = Workflow.objects.create(farm_id=get_default_farm_id(), recipe=RECIPE)
        workflow.start()
        task = workflow.tasks.get(slug='first')
        artifact = task.artifacts.create(
            path='path/to/file.json',
            size=100,
            content_type='application/json',
        )

        response = self.client.patch(
            reverse('api:version-retrieve', kwargs={'pk': str(self.version_1.id)}),
            data={
                'configuration': {"test": "test2"},
                'docker_image': str(artifact.id),
                'docker_image_iid': 'sha256:deadbeef',
                'state': 'available',
            },
            format='json',
            HTTP_AUTHORIZATION=f'Ponos {self.task.token}',
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['id'], str(self.version_1.id))
        self.assertEqual(data['configuration'], {"test": "test2"})
        self.assertEqual(data['docker_image'], str(artifact.id))
        self.assertEqual(data['docker_image_iid'], 'sha256:deadbeef')
        self.assertEqual(data['state'], 'available')
        self.assertEqual(data['gpu_usage'], 'disabled')

    def test_update_version_available_requires_docker_image(self):
        self.version_1.state = WorkerVersionState.Created
        self.version_1.docker_image = None
        self.version_1.save()

        response = self.client.patch(
            reverse('api:version-retrieve', kwargs={'pk': str(self.version_1.id)}),
            data={'state': 'available'},
            format='json',
            HTTP_AUTHORIZATION=f'Ponos {self.task.token}',
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'non_field_errors': [
                'A WorkerVersion must have a Docker image artifact and a Docker image ID set when it is available.'
            ]
        })

    def test_update_version_available_docker_image_not_null(self):
        workflow = Workflow.objects.create(farm_id=get_default_farm_id(), recipe=RECIPE)
        workflow.start()
        task = workflow.tasks.get(slug='first')
        self.version_1.docker_image = task.artifacts.create(
            path='path/to/file.json',
            size=100,
            content_type='application/json',
        )
        self.version_1.state = WorkerVersionState.Available
        self.version_1.save()

        response = self.client.patch(
            reverse('api:version-retrieve', kwargs={'pk': str(self.version_1.id)}),
            data={'docker_image': None},
            format='json',
            HTTP_AUTHORIZATION=f'Ponos {self.task.token}',
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'non_field_errors': [
                'A WorkerVersion must have a Docker image artifact and a Docker image ID set when it is available.'
            ]
        })

    def test_update_version_available_docker_image_iid_not_null(self):
        workflow = Workflow.objects.create(farm_id=get_default_farm_id(), recipe=RECIPE)
        workflow.start()
        task = workflow.tasks.get(slug='first')
        self.version_1.docker_image = task.artifacts.create(
            path='path/to/file.json',
            size=100,
            content_type='application/json',
        )
        self.version_1.docker_image_iid = 'something:latest'
        self.version_1.state = WorkerVersionState.Available
        self.version_1.save()

        response = self.client.patch(
            reverse('api:version-retrieve', kwargs={'pk': str(self.version_1.id)}),
            data={'docker_image_iid': None},
            format='json',
            HTTP_AUTHORIZATION=f'Ponos {self.task.token}',
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'non_field_errors': [
                'A WorkerVersion must have a Docker image artifact and a Docker image ID set when it is available.'
            ]
        })

    def test_property_docker_image_name(self):
        self.assertEqual(self.version_1.docker_image_name, f'my_repo.fake/workers/worker/reco:{self.version_1.id}')

    def test_property_docker_image_name_explicit(self):
        version_2 = WorkerVersion.objects.create(
            id="12341234-1234-1234-1234-123412341234",
            worker=self.worker_reco,
            revision=self.rev2,
            configuration={"test": "test1"},
            gpu_usage=WorkerVersionGPUUsage.Disabled
        )
        self.assertEqual(
            version_2.docker_image_name,
            'my_repo.fake/workers/worker/reco:12341234-1234-1234-1234-123412341234'
        )

    def test_property_slug(self):
        self.assertEqual(self.version_1.slug, f'reco_{str(self.version_1.id)[0:6]}')

    def test_property_slug_explicit(self):
        version_2 = WorkerVersion.objects.create(
            id="01234567-89ab-cdef-fedc-ba9876543210",
            worker=self.worker_reco,
            revision=self.rev2,
            configuration={"test": "test1"},
            gpu_usage=WorkerVersionGPUUsage.Disabled
        )
        self.assertEqual(version_2.slug, 'reco_012345')

    def test_property_docker_command_empty(self):
        self.assertEqual(self.version_1.docker_command, None)

    def test_property_docker_command_not_empty(self):
        self.version_1.configuration = {'docker': {'command': 'mysupercommand'}}
        self.assertEqual(self.version_1.docker_command, 'mysupercommand')
        self.version_1.configuration = {"test": "test1"}

    def test_property_docker_shm_size(self):
        self.version_1.configuration = {'docker': {'shm_size': '32b'}}
        self.assertEqual(self.version_1.docker_shm_size, '32b')

    def test_corpus_worker_version_no_login_public(self):
        self.assertTrue(self.corpus.public)
        # Use multiple versions to ensure we don't get duplicated queries
        self.corpus.worker_versions.set([self.version_1, self.version_2])

        with self.assertNumQueries(7):
            response = self.client.get(reverse('api:corpus-versions', kwargs={'pk': self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            'count': 2,
            'number': 1,
            'previous': None,
            'next': None,
            'results': [
                {
                    'id': str(self.version_2.id),
                    'configuration': {'test': 42},
                    'revision': {
                        'id': str(self.version_2.revision_id),
                        'hash': '1337',
                        'author': 'Test user',
                        'message': 'My w0rk3r',
                        'created': '2020-02-02T01:23:45.678000Z',
                        'commit_url': 'http://my_repo.fake/workers/worker/commit/1337',
                        'refs': []
                    },
                    'gpu_usage': 'disabled',
                    'model_usage': False,
                    'docker_image': str(self.version_2.docker_image_id),
                    'docker_image_iid': None,
                    'docker_image_name': f'my_repo.fake/workers/worker/dla:{self.version_2.id}',
                    'state': 'available',
                    'worker': {
                        'id': str(self.worker_dla.id),
                        'name': 'Document layout analyser',
                        'type': 'dla',
                        'slug': 'dla',
                    }
                },
                {
                    'id': str(self.version_1.id),
                    'configuration': {'test': 42},
                    'revision': {
                        'id': str(self.version_1.revision_id),
                        'hash': '1337',
                        'author': 'Test user',
                        'message': 'My w0rk3r',
                        'created': '2020-02-02T01:23:45.678000Z',
                        'commit_url': 'http://my_repo.fake/workers/worker/commit/1337',
                        'refs': []
                    },
                    'gpu_usage': 'disabled',
                    'model_usage': False,
                    'docker_image': str(self.version_1.docker_image_id),
                    'docker_image_iid': None,
                    'docker_image_name': f'my_repo.fake/workers/worker/reco:{self.version_1.id}',
                    'state': 'available',
                    'worker': {
                        'id': str(self.worker_reco.id),
                        'name': 'Recognizer',
                        'type': 'recognizer',
                        'slug': 'reco',
                    }
                },
            ],
        })

    def test_corpus_worker_version_no_login_private(self):
        self.corpus.public = False
        self.corpus.save()
        self.corpus.worker_versions.set([self.version_1, self.version_2])

        with self.assertNumQueries(1):
            response = self.client.get(reverse('api:corpus-versions', kwargs={'pk': self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertDictEqual(response.json(), {'detail': 'Not found.'})

    def test_corpus_worker_version_not_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        # Use multiple versions to ensure we don't get duplicated queries
        self.corpus.worker_versions.set([self.version_1, self.version_2])

        with self.assertNumQueries(11):
            response = self.client.get(reverse('api:corpus-versions', kwargs={'pk': self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            'count': 2,
            'number': 1,
            'previous': None,
            'next': None,
            'results': [
                {
                    'id': str(self.version_2.id),
                    'configuration': {'test': 42},
                    'revision': {
                        'id': str(self.version_2.revision_id),
                        'hash': '1337',
                        'author': 'Test user',
                        'message': 'My w0rk3r',
                        'created': '2020-02-02T01:23:45.678000Z',
                        'commit_url': 'http://my_repo.fake/workers/worker/commit/1337',
                        'refs': []
                    },
                    'gpu_usage': 'disabled',
                    'model_usage': False,
                    'docker_image': str(self.version_2.docker_image_id),
                    'docker_image_iid': None,
                    'docker_image_name': f'my_repo.fake/workers/worker/dla:{self.version_2.id}',
                    'state': 'available',
                    'worker': {
                        'id': str(self.worker_dla.id),
                        'name': 'Document layout analyser',
                        'type': 'dla',
                        'slug': 'dla',
                    }
                },
                {
                    'id': str(self.version_1.id),
                    'configuration': {'test': 42},
                    'revision': {
                        'id': str(self.version_1.revision_id),
                        'hash': '1337',
                        'author': 'Test user',
                        'message': 'My w0rk3r',
                        'created': '2020-02-02T01:23:45.678000Z',
                        'commit_url': 'http://my_repo.fake/workers/worker/commit/1337',
                        'refs': []
                    },
                    'gpu_usage': 'disabled',
                    'model_usage': False,
                    'docker_image': str(self.version_1.docker_image_id),
                    'docker_image_iid': None,
                    'docker_image_name': f'my_repo.fake/workers/worker/reco:{self.version_1.id}',
                    'state': 'available',
                    'worker': {
                        'id': str(self.worker_reco.id),
                        'name': 'Recognizer',
                        'type': 'recognizer',
                        'slug': 'reco',
                    }
                },
            ],
        })

    def test_corpus_worker_version_list(self):
        self.client.force_login(self.user)
        # Use multiple versions to ensure we don't get duplicated queries
        self.corpus.worker_versions.set([self.version_1, self.version_2])

        with self.assertNumQueries(11):
            response = self.client.get(reverse('api:corpus-versions', kwargs={'pk': self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            'count': 2,
            'number': 1,
            'previous': None,
            'next': None,
            'results': [
                {
                    'id': str(self.version_2.id),
                    'configuration': {'test': 42},
                    'revision': {
                        'id': str(self.version_2.revision_id),
                        'hash': '1337',
                        'author': 'Test user',
                        'message': 'My w0rk3r',
                        'created': '2020-02-02T01:23:45.678000Z',
                        'commit_url': 'http://my_repo.fake/workers/worker/commit/1337',
                        'refs': []
                    },
                    'gpu_usage': 'disabled',
                    'model_usage': False,
                    'docker_image': str(self.version_2.docker_image_id),
                    'docker_image_iid': None,
                    'docker_image_name': f'my_repo.fake/workers/worker/dla:{self.version_2.id}',
                    'state': 'available',
                    'worker': {
                        'id': str(self.worker_dla.id),
                        'name': 'Document layout analyser',
                        'type': 'dla',
                        'slug': 'dla',
                    }
                },
                {
                    'id': str(self.version_1.id),
                    'configuration': {'test': 42},
                    'revision': {
                        'id': str(self.version_1.revision_id),
                        'hash': '1337',
                        'author': 'Test user',
                        'message': 'My w0rk3r',
                        'created': '2020-02-02T01:23:45.678000Z',
                        'commit_url': 'http://my_repo.fake/workers/worker/commit/1337',
                        'refs': []
                    },
                    'gpu_usage': 'disabled',
                    'model_usage': False,
                    'docker_image': str(self.version_1.docker_image_id),
                    'docker_image_iid': None,
                    'docker_image_name': f'my_repo.fake/workers/worker/reco:{self.version_1.id}',
                    'state': 'available',
                    'worker': {
                        'id': str(self.worker_reco.id),
                        'name': 'Recognizer',
                        'type': 'recognizer',
                        'slug': 'reco',
                    }
                },
            ],
        })

    def test_corpus_worker_version_no_right(self):
        self.user.rights.all().delete()
        self.client.force_login(self.user)
        # Use multiple versions to ensure we don't get duplicated queries
        self.corpus.worker_versions.set([self.version_1, self.version_2])

        with self.assertNumQueries(11):
            response = self.client.get(reverse('api:corpus-versions', kwargs={'pk': self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            'count': 2,
            'number': 1,
            'previous': None,
            'next': None,
            'results': [
                {
                    'id': str(self.version_2.id),
                    'configuration': {'test': 42},
                    'revision': {
                        'id': str(self.version_2.revision_id),
                        'hash': '1337',
                        'author': 'Test user',
                        'message': 'My w0rk3r',
                        'created': '2020-02-02T01:23:45.678000Z',
                        'commit_url': 'http://my_repo.fake/workers/worker/commit/1337',
                        'refs': []
                    },
                    'gpu_usage': 'disabled',
                    'model_usage': False,
                    'docker_image': str(self.version_2.docker_image_id),
                    'docker_image_iid': None,
                    'docker_image_name': f'my_repo.fake/workers/worker/dla:{self.version_2.id}',
                    'state': 'available',
                    'worker': {
                        'id': str(self.worker_dla.id),
                        'name': 'Document layout analyser',
                        'type': 'dla',
                        'slug': 'dla',
                    }
                },
                {
                    'id': str(self.version_1.id),
                    'configuration': {'test': 42},
                    'revision': {
                        'id': str(self.version_1.revision_id),
                        'hash': '1337',
                        'author': 'Test user',
                        'message': 'My w0rk3r',
                        'created': '2020-02-02T01:23:45.678000Z',
                        'commit_url': 'http://my_repo.fake/workers/worker/commit/1337',
                        'refs': []
                    },
                    'gpu_usage': 'disabled',
                    'model_usage': False,
                    'docker_image': str(self.version_1.docker_image_id),
                    'docker_image_iid': None,
                    'docker_image_name': f'my_repo.fake/workers/worker/reco:{self.version_1.id}',
                    'state': 'available',
                    'worker': {
                        'id': str(self.worker_reco.id),
                        'name': 'Recognizer',
                        'type': 'recognizer',
                        'slug': 'reco',
                    }
                },
            ],
        })
