from collections import namedtuple
from unittest.mock import call, patch

import yaml
from django.test import override_settings
from rest_framework import status
from rest_framework.reverse import reverse

from arkindex.documents.models import Corpus, Element
from arkindex.ponos.models import State, Workflow
from arkindex.process.models import ActivityState, Process, ProcessMode, WorkerActivity, WorkerVersion
from arkindex.process.utils import get_default_farm_id
from arkindex.project.mixins import SelectionMixin
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role

RECIPE = '''
tasks:
  first:
    image: alpine
'''


class TestWorkflows(FixtureAPITestCase):
    """
    Test workflows from Arkindex existing elements
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.volume = Element.objects.get(name='Volume 1')
        cls.pages = Element.objects.get_descending(cls.volume.id).filter(type__slug='page', polygon__isnull=False)
        cls.private_corpus = Corpus.objects.create(name='private')
        cls.private_corpus.elements.create(
            type=cls.private_corpus.types.create(slug='volume', folder=True),
            name='Hidden',
        )

        # Workers ProcessMode
        cls.creds = cls.user.credentials.get()
        cls.repo = cls.creds.repos.get(url='http://my_repo.fake/workers/worker')
        cls.rev_1 = cls.repo.revisions.get()
        cls.rev_2 = cls.repo.revisions.create(
            hash='2',
            message='commit message',
            author='bob',
        )
        cls.version_1 = WorkerVersion.objects.get(worker__slug='reco')
        cls.worker_1 = cls.version_1.worker
        cls.version_2 = WorkerVersion.objects.get(worker__slug='dla')
        cls.worker_2 = cls.version_2.worker
        cls.version_3 = WorkerVersion.objects.get(worker__slug='worker-gpu')
        cls.worker_3 = cls.version_3.worker

    @override_settings(
        ARKINDEX_TASKS_IMAGE='tasks:latest',
        PONOS_RECIPE={'env': {'SOME_VAR': 'somevalue'}}
    )
    def test_page_workflow(self):
        self.client.force_login(self.user)

        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'corpus': str(self.corpus.id),
                'element_type': 'page',
                'load_children': True,
            },
            format='json',
        )

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        data = response.json()
        process = Process.objects.get(id=data['id'])
        self.assertIsNone(process.workflow)

        self.assertDictEqual(data, {
            'name': None,
            'id': str(process.id),
            'state': 'unscheduled',
            'corpus': str(self.volume.corpus.id),
            'files': [],
            'mode': 'workers',
            'revision': process.revision,
            'workflow': None,
            'folder_type': None,
            'element_type': 'page',
            'element': None,
            'template_id': None,
            'load_children': True,
            'use_cache': False,
            'use_gpu': False,
            'activity_state': ActivityState.Disabled.value,
            'element_name_contains': None,
            'model_id': None,
            'train_folder_id': None,
            'validation_folder_id': None,
            'test_folder_id': None,
        })

        # The process needs to be started to produce a workflow
        response = self.client.post(
            reverse('api:process-start', kwargs={'pk': str(process.id)}),
            # The process needs a worker run or thumbnails generation to start
            {'thumbnails': 'true'}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        process.refresh_from_db()
        self.assertIsNotNone(process.workflow)

        self.assertEqual(process.workflow.recipes['initialisation'].image, 'tasks:latest')
        self.assertEqual(
            process.workflow.recipes['initialisation'].command,
            'python -m arkindex_tasks.init_elements '
            f'{process.id} '
            '--chunks-number 1'
        )

    def test_volume_workflow(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'corpus': str(self.corpus.id),
                'element': str(self.volume.id),
                'load_children': True,
                'use_cache': True,
            },
            format='json'
        )

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        process = Process.objects.get(id=data['id'])
        self.assertIsNone(process.workflow)

        self.assertEqual(process.element, self.volume)
        self.assertEqual(data.pop('element')['id'], str(self.volume.id))
        self.assertDictEqual(data, {
            'name': None,
            'id': str(process.id),
            'state': 'unscheduled',
            'mode': 'workers',
            'corpus': str(self.corpus.id),
            'files': [],
            'revision': process.revision,
            'workflow': None,
            'folder_type': None,
            'element_type': None,
            'template_id': None,
            'load_children': True,
            'use_cache': True,
            'use_gpu': False,
            'activity_state': ActivityState.Disabled.value,
            'element_name_contains': None,
            'model_id': None,
            'train_folder_id': None,
            'validation_folder_id': None,
            'test_folder_id': None,
        })

    def test_single_page_workflow(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'corpus': str(self.corpus.id),
                'element': str(self.pages.first().id),
            },
            format='json'
        )

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        process = Process.objects.get(id=data['id'])
        self.assertEqual(process.element, self.pages.first())

    def test_create_element_workflow_requires_login(self):
        response = self.client.post(reverse('api:corpus-workflow'), {}, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_create_element_workflow_required_fields(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:corpus-workflow'), {}, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'corpus': ['This field is required.']}
        )

    def test_create_element_workflow_requires_corpus_admin(self):
        """
        Only an admin of the target corpus can create a workers process
        """
        self.client.force_login(self.user)
        self.corpus.memberships.filter(user=self.user).update(level=Role.Contributor.value)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'corpus': str(self.corpus.id),
                'name': 'Test',
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'corpus': ['You do not have an admin access to this corpus.']})

    def test_create_element_workflow_non_readable_corpus(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'corpus': str(self.private_corpus.id),
                'name': 'Test',
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'corpus': ['Corpus with this ID does not exist.']})

    def test_create_element_workflow_name_filter(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'corpus': str(self.corpus.id),
                'element_name_contains': self.pages.first().name[2:5],
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        process = Process.objects.get(id=data['id'])
        self.assertEqual(process.name_contains, self.pages.first().name[2:5])

    def test_create_element_workflow_fields_max_length(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'corpus': str(self.corpus.id),
                'name': 'A' * 200,
                'element_name_contains': 'B' * 300,
                'element_type': 'C' * 100,
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'name': ['Ensure this field has no more than 100 characters.'],
            'element_name_contains': ['Ensure this field has no more than 250 characters.'],
            'element_type': ['Ensure this field has no more than 50 characters.'],
        })

    def test_create_element_workflow_type_filter(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'corpus': str(self.corpus.id),
                'element_type': 'volume',
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        process = Process.objects.get(id=data['id'])
        self.assertEqual(process.element_type.slug, 'volume')

    def test_create_element_workflow_empty_queryset(self):
        """
        An empty QuerySet is ignored: the user can change the element filtering later,
        and running a process on no elements at all is harmless.
        """
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'corpus': str(self.corpus.id),
                'element_type': 'volume',
                'element_name_contains': 'ahahah',
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        process = Process.objects.get(id=data['id'])
        self.assertEqual(process.corpus, self.corpus)
        self.assertEqual(process.element_type.slug, 'volume')
        self.assertEqual(process.name_contains, 'ahahah')
        self.assertFalse(process.list_elements().exists())

    def test_create_element_workflow_empty_selection(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'corpus': str(self.corpus.id),
                'selection': True,
            },
            format='json'
        )
        mixin = SelectionMixin()
        mixin.request = namedtuple('Request', 'user')(self.user)
        self.assertEqual(mixin.get_selection(corpus_id=self.corpus.id).count(), 0)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        data = response.json()
        self.assertIn('__all__', data)
        self.assertListEqual(
            data['__all__'],
            ['No element match those filters.']
        )

    def test_create_element_workflow_wrong_type(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'corpus': str(self.corpus.id),
                'element_type': 'arachnid',
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'element_type': ['This type does not exist in corpus "Unit Tests"']
        })

    def test_create_element_workflow_mixed_element_filters(self):
        """
        Element parameter can be set with filtering parameters
        """
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'corpus': str(self.corpus.id),
                'element': str(self.pages.first().id),
                'type': 'page',
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_create_element_workflow_non_folder_no_zone(self):
        """
        An element that is not of a folder type and has no image or polygon
        should still be usable to create a process
        """
        self.client.force_login(self.user)
        element = self.corpus.elements.create(type=self.pages.first().type, name='Kill me please')
        with self.assertNumQueries(12):
            response = self.client.post(
                reverse('api:corpus-workflow'),
                {
                    'corpus': str(self.corpus.id),
                    'element': str(element.id),
                },
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        process = Process.objects.get(element=element, mode=ProcessMode.Workers)
        self.assertDictEqual(response.json(), {
            'name': None,
            'id': str(process.id),
            'state': 'unscheduled',
            'corpus': str(self.volume.corpus.id),
            'files': [],
            'mode': 'workers',
            'revision': process.revision,
            'workflow': None,
            'folder_type': None,
            'element_type': None,
            'element': {
                'id': str(element.id),
                'name': 'Kill me please',
                'type': 'page',
                'zone': None,
                'corpus': {
                    'id': str(self.corpus.id),
                    'name': 'Unit Tests',
                    'public': True,
                },
                'rotation_angle': 0,
                'mirrored': False,
                'thumbnail_url': None,
                'thumbnail_put_url': None,
            },
            'template_id': None,
            'load_children': False,
            'use_cache': False,
            'use_gpu': False,
            'activity_state': ActivityState.Disabled.value,
            'element_name_contains': None,
            'model_id': None,
            'train_folder_id': None,
            'validation_folder_id': None,
            'test_folder_id': None,
        })

    def test_create_element_workflow_mixed_selection_filters(self):
        """
        Element and selection fields cannot be set together
        """
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'corpus': str(self.corpus.id),
                'element': str(self.volume.id),
                'selection': True,
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {
                'element': ['This field cannot be set with the selection field.'],
                'selection': ['This field cannot be set with the element field.']
            }
        )

    def test_create_element_workflow_selection_filter(self):
        self.client.force_login(self.user)
        page = self.pages.first()
        self.client.post(
            reverse('api:elements-selection'),
            data={'ids': [page.id]},
            format='json'
        )
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'corpus': str(self.corpus.id),
                'selection': True,
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        process = Process.objects.get(id=data['id'])
        self.assertIsNone(process.workflow)
        self.assertEqual(process.elements.get(), page)

        # The process needs to be started to produce a workflow
        response = self.client.post(
            reverse('api:process-start', kwargs={'pk': str(process.id)}),
            # The process needs a worker run or thumbnails generation to start
            {'thumbnails': 'true'}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        process.refresh_from_db()
        self.assertIsNotNone(process.workflow)

        self.assertEqual(
            process.workflow.recipes['initialisation'].command,
            'python -m arkindex_tasks.init_elements '
            f'{process.id} '
            '--chunks-number 1'
        )

        response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': process.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 1)
        self.assertEqual(data['results'][0]['id'], str(page.id))

    @override_settings(ARKINDEX_FEATURES={'selection': False})
    def test_create_element_workflow_selection_no_selection(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'corpus': str(self.corpus.id),
                'selection': True,
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'selection': ['Selection is not available on this instance.']})

    def test_create_element_workflow_process_name(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'corpus': str(self.corpus.id),
                'name': 'blah',
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        process = Process.objects.get(id=data['id'])
        self.assertEqual(process.name, 'blah')

    def test_thumbnails_generation_only(self):
        """
        Generating thumbnails without any worker must generate a process task
        and a thumbnails task, when the process is started
        """
        self.client.force_login(self.user)
        corpus_type = self.corpus.types.first()
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'corpus': str(self.corpus.id),
                'element_type': corpus_type.slug,
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        process = Process.objects.get(id=data['id'])
        self.assertIsNone(process.workflow)

        response = self.client.post(
            reverse('api:process-start', kwargs={'pk': str(process.id)}),
            {
                "chunks": 3,
                'thumbnails': True,
            }
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        process.refresh_from_db()
        self.assertIsNotNone(process.workflow)

        self.assertEqual(
            process.workflow.recipes['initialisation'].command,
            'python -m arkindex_tasks.init_elements '
            f'{process.id} '
            '--chunks-number 3'
        )
        for i in range(1, 4):
            self.assertEqual(
                process.workflow.recipes[f'thumbnails_{i}'].command,
                f'python3 -m arkindex_tasks.generate_thumbnails /data/initialisation/elements_chunk_{i}.json'
            )

    @patch('arkindex.project.triggers.process_tasks.initialize_activity.delay')
    @override_settings(ARKINDEX_TASKS_IMAGE='registry.teklia.com/tasks')
    def test_workers_multiple_worker_runs(self, activities_delay_mock):
        process_2 = self.corpus.processes.create(
            activity_state=ActivityState.Pending,
            creator=self.user,
            mode=ProcessMode.Workers,
            load_children=True,
        )
        run_1 = process_2.worker_runs.create(
            version=self.version_1,
            parents=[],
        )
        process_2.worker_runs.create(
            version=self.version_2,
            parents=[run_1.id],
        )

        workflow_tmp = Workflow.objects.create(farm_id=get_default_farm_id(), recipe=RECIPE)
        workflow_tmp.start()

        self.assertIsNone(process_2.workflow)
        self.assertFalse(self.corpus.worker_versions.exists())

        self.client.force_login(self.user)
        with self.assertNumQueries(28):
            response = self.client.post(
                reverse('api:process-start', kwargs={'pk': str(process_2.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()['id'], str(process_2.id))

        process_2.refresh_from_db()
        self.assertEqual(process_2.state, State.Unscheduled)
        self.assertIsNotNone(process_2.workflow)
        run_1 = process_2.worker_runs.get(version=self.version_1)
        run_2 = process_2.worker_runs.get(version=self.version_2)

        self.assertDictEqual(yaml.safe_load(process_2.workflow.recipe)['tasks'], {
            'initialisation':
                {
                    'command': 'python -m arkindex_tasks.init_elements '
                               f'{process_2.id} --chunks-number 1',
                    'image': 'registry.teklia.com/tasks'
                },
            f'reco_{str(self.version_1.id)[0:6]}':
                {
                    'command': None,
                    'image': f'my_repo.fake/workers/worker/reco:{self.version_1.id}',
                    'artifact': str(self.version_1.docker_image.id),
                    'parents': ['initialisation'],
                    'shm_size': None,
                    'env': {
                        'TASK_ELEMENTS': '/data/initialisation/elements.json',
                        'WORKER_VERSION_ID': str(self.version_1.id),
                        'ARKINDEX_WORKER_RUN_ID': str(run_1.id),
                    }
                },
            f'dla_{str(self.version_2.id)[0:6]}':
                {
                    'command': None,
                    'image': f'my_repo.fake/workers/worker/dla:{self.version_2.id}',
                    'artifact': str(self.version_1.docker_image.id),
                    'parents': [f'reco_{str(self.version_1.id)[0:6]}'],
                    'shm_size': None,
                    'env': {
                        'TASK_ELEMENTS': '/data/initialisation/elements.json',
                        'WORKER_VERSION_ID': str(self.version_2.id),
                        'ARKINDEX_WORKER_RUN_ID': str(run_2.id),
                    }
                }
        })

        # Check worker activities has been created for concerned elements on both runs
        self.assertEqual(activities_delay_mock.call_count, 1)
        self.assertEqual(activities_delay_mock.call_args, call(process_2))
        self.assertEqual(process_2.activity_state, ActivityState.Pending)

        # Manually trigger activity creations
        from arkindex.process.tasks import initialize_activity
        with self.assertNumQueries(8):
            initialize_activity(process_2)
        process_2.refresh_from_db()
        self.assertEqual(process_2.activity_state, ActivityState.Ready)
        elements_ids = self.corpus.elements.values_list('id', flat=True)
        self.assertCountEqual(
            elements_ids,
            WorkerActivity.objects.filter(worker_version=self.version_1).values_list('element_id', flat=True)
        )
        self.assertCountEqual(
            elements_ids,
            WorkerActivity.objects.filter(worker_version=self.version_2).values_list('element_id', flat=True)
        )

        # Check that the corpus worker version cache has been updated
        self.assertCountEqual(self.corpus.worker_versions.all(), [self.version_1, self.version_2])

    @override_settings(ARKINDEX_TASKS_IMAGE='registry.teklia.com/tasks')
    def test_create_process_use_cache_option(self):
        """
        A process with the `use_cache` parameter creates an initialization task with the --use-cache flag
        """
        process_2 = self.corpus.processes.create(creator=self.user, mode=ProcessMode.Workers)
        process_2.worker_runs.create(
            version=self.version_1,
            parents=[],
        )
        process_2.use_cache = True
        process_2.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(25):
            response = self.client.post(
                reverse('api:process-start', kwargs={'pk': str(process_2.id)})
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()['use_cache'], True)
        process_2.refresh_from_db()
        self.assertEqual(process_2.state, State.Unscheduled)
        self.assertIsNotNone(process_2.workflow)
        self.assertDictEqual(yaml.safe_load(process_2.workflow.recipe)['tasks'], {
            'initialisation':
                {
                    'command': 'python -m arkindex_tasks.init_elements '
                               f'{process_2.id} --chunks-number 1 '
                               '--use-cache',
                    'image': 'registry.teklia.com/tasks'
                },
            f'reco_{str(self.version_1.id)[0:6]}':
                {
                    'command': None,
                    'image': f'my_repo.fake/workers/worker/reco:{self.version_1.id}',
                    'artifact': str(self.version_1.docker_image.id),
                    'parents': ['initialisation'],
                    'shm_size': None,
                    'env': {
                        'TASK_ELEMENTS': '/data/initialisation/elements.json',
                        'WORKER_VERSION_ID': str(self.version_1.id),
                        'ARKINDEX_WORKER_RUN_ID': str(process_2.worker_runs.get().id),
                    }
                },
        })

    @override_settings(ARKINDEX_TASKS_IMAGE='registry.teklia.com/tasks')
    def test_create_process_use_gpu_option(self):
        """
        A process with the `use_gpu` parameter enables the `requires_gpu` attribute of tasks than need one
        """
        process_2 = self.corpus.processes.create(creator=self.user, mode=ProcessMode.Workers)
        process_2.worker_runs.create(
            version=self.version_3,
            parents=[],
        )
        process_2.use_gpu = True
        process_2.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(25):
            response = self.client.post(reverse('api:process-start', kwargs={'pk': str(process_2.id)}))

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()['use_gpu'], True)
        process_2.refresh_from_db()
        self.assertEqual(process_2.state, State.Unscheduled)
        self.assertIsNotNone(process_2.workflow)
        self.assertDictEqual(yaml.safe_load(process_2.workflow.recipe)['tasks'], {
            'initialisation':
                {
                    'command': 'python -m arkindex_tasks.init_elements '
                               f'{process_2.id} --chunks-number 1',
                    'image': 'registry.teklia.com/tasks'
                },
            f'worker-gpu_{str(self.version_3.id)[0:6]}':
                {
                    'command': None,
                    'image': f'my_repo.fake/workers/worker/worker-gpu:{self.version_3.id}',
                    'artifact': str(self.version_3.docker_image.id),
                    'parents': ['initialisation'],
                    'shm_size': None,
                    'env': {
                        'TASK_ELEMENTS': '/data/initialisation/elements.json',
                        'WORKER_VERSION_ID': str(self.version_3.id),
                        'ARKINDEX_WORKER_RUN_ID': str(process_2.worker_runs.get().id),
                    },
                    'requires_gpu': True,
                },
        })

    def test_retry_keeps_requires_gpu(self):
        """
        When a process is retried, the newly created tasks keep the same requires_gpu values
        """
        process = self.corpus.processes.create(creator=self.user, mode=ProcessMode.Workers)
        process.worker_runs.create(
            version=self.version_3,
            parents=[],
        )
        process.use_gpu = True
        process.save()
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:process-start', kwargs={'pk': str(process.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        process.refresh_from_db()
        self.assertEqual(process.state, State.Unscheduled)

        init_task, worker_task = process.workflow.tasks.filter(run=0)
        self.assertEqual(init_task.requires_gpu, False)
        self.assertEqual(worker_task.requires_gpu, True)

        process.workflow.tasks.update(state=State.Stopped)

        response = self.client.post(reverse('api:process-retry', kwargs={'pk': str(process.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        init_task, worker_task = process.workflow.tasks.filter(run=1)
        self.assertEqual(init_task.requires_gpu, False)
        self.assertEqual(worker_task.requires_gpu, True)

    def test_use_gpu_unsupported(self):
        process = self.corpus.processes.create(creator=self.user, mode=ProcessMode.Workers)
        process.worker_runs.create(
            version=self.version_2,
            parents=[],
        )
        process.use_gpu = True
        process.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse('api:process-start', kwargs={'pk': str(process.id)}),
                {'use_gpu': 'true'}
            )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(
            response.json(),
            {'use_gpu': 'The process is configured to use GPU, but does not include any workers that support GPU usage.'}
        )

    def test_use_gpu_some_unsupported(self):
        process = self.corpus.processes.create(creator=self.user, mode=ProcessMode.Workers)
        process.worker_runs.create(
            version=self.version_2,
            parents=[],
        )
        process.worker_runs.create(
            version=self.version_3,
            parents=[],
        )
        process.use_gpu = True
        process.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(29):
            response = self.client.post(
                reverse('api:process-start', kwargs={'pk': str(process.id)}),
                {'use_gpu': 'true'}
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        process.refresh_from_db()
        init_task, worker_2_task, worker_3_task = process.workflow.tasks.filter(run=0)
        self.assertEqual(init_task.requires_gpu, False)
        self.assertEqual(worker_2_task.requires_gpu, False)
        self.assertEqual(worker_3_task.requires_gpu, True)
