import uuid
from datetime import datetime, timezone

from django.conf import settings
from django.contrib.auth.models import AbstractBaseUser
from django.contrib.contenttypes.fields import GenericForeignKey, GenericRelation
from django.contrib.contenttypes.models import ContentType
from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from django_rq.queues import get_queue
from enumfields import Enum, EnumField
from rq.utils import as_text

from arkindex.users.managers import UserManager
from arkindex.users.providers import get_provider


class Role(Enum):
    Admin = 100
    Contributor = 50
    Guest = 10


class Right(models.Model):
    """
    Defines an access right for users and groups to a generic target.
    """
    id = models.UUIDField(default=uuid.uuid4, primary_key=True, editable=False)

    # Owner of the right (XOR)
    user = models.ForeignKey('users.User', null=True, on_delete=models.CASCADE, related_name='rights')
    group = models.ForeignKey('users.Group', null=True, on_delete=models.CASCADE, related_name='rights')

    # Target of the right
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    content_id = models.UUIDField(editable=False)
    content_object = GenericForeignKey('content_type', 'content_id')

    level = models.PositiveIntegerField(
        help_text='Maximum privilege level.',
        validators=[
            MinValueValidator(1),
            MaxValueValidator(100)
        ]
    )

    class Meta:
        constraints = [
            # User XOR Group is the owner of this access right
            models.CheckConstraint(
                name='user_xor_group',
                check=(
                    models.Q(group_id__isnull=False, user_id__isnull=True)
                    | models.Q(group_id__isnull=True, user_id__isnull=False)
                ),
            ),
            # Ensure that the rights owner only has one Right instance for the same target.
            # Since either user or group can be NULL, and a NULL in the fields of a unique constraint causes the
            # constraint to still allow duplicates, this constraint is split in two partial constraints.
            models.UniqueConstraint(
                fields=['user', 'content_id', 'content_type'],
                name='right_user_unique_target',
                condition=models.Q(user__isnull=False),
            ),
            models.UniqueConstraint(
                fields=['group', 'content_id', 'content_type'],
                name='right_group_unique_target',
                condition=models.Q(group__isnull=False),
            ),
        ]


class User(AbstractBaseUser):
    email = models.EmailField(
        verbose_name='email address',
        max_length=255,
        db_collation='case_insensitive',
    )
    display_name = models.CharField(max_length=120)
    transkribus_email = models.EmailField(
        max_length=255,
        null=True,
        blank=True,
    )
    is_active = models.BooleanField(default=True)
    is_admin = models.BooleanField(default=False)
    verified_email = models.BooleanField(default=False)

    selected_elements = models.ManyToManyField(
        'documents.Element',
        through='documents.Selection',
        related_name='selection_users',
    )

    objects = UserManager()

    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)

    USERNAME_FIELD = 'email'
    REQUIRED_FIELDS = ['display_name', ]

    class Meta:
        constraints = [
            models.UniqueConstraint(fields=['email'], name='email_unique'),
        ]

    def __str__(self):
        return self.email

    def has_perm(self, perm, obj=None):
        "Does the user have a specific permission?"
        # Simplest possible answer: Yes, always
        return True

    def has_module_perms(self, app_label):
        "Does the user have permissions to view the app `app_label`?"
        # Simplest possible answer: Yes, always
        return True

    def has_scopes(self, *scopes):
        if not scopes:
            return True
        assert all(isinstance(scope, Scope) for scope in scopes), 'Scopes should be from the Scope enum'
        return self.user_scopes.filter(scope__in=scopes).count() == len(set(scopes))

    def has_scope(self, scope):
        assert isinstance(scope, Scope), 'Scope should be from the Scope enum'
        # TODO: Could be optimized in a way similar to arkindex.ponos.models.Workflow.get_state()
        # See https://gitlab.com/teklia/arkindex/ponos/-/blob/e5989f3e1dd1f6d7d93ba940b7ed2c5471ddd34f/ponos/models.py#L370
        return self.user_scopes.filter(scope=scope).exists()

    @property
    def is_staff(self):
        "Is the user a member of staff?"
        # Simplest possible answer: All admins are staff
        return self.is_admin

    def get_rq_jobs(self):
        """
        List RQ jobs linked to this user's ID in all queues, cleaning up any deleted jobs from the user registry
        """
        queue = get_queue('default')
        registry = queue.user_registry(self.id)
        job_ids = registry.get_job_ids()
        to_delete = set()

        with queue.connection.pipeline() as pipeline:
            for job_id in job_ids:
                job_id = as_text(job_id)
                pipeline.hgetall(queue.job_class.key_for(job_id))

            job_payloads = pipeline.execute()

            for job_id, job_payload in zip(job_ids, job_payloads):
                if job_payload:
                    job = queue.job_class(job_id, connection=queue.connection, serializer=queue.serializer)
                    job.restore(job_payload)
                    yield job
                else:
                    # When there is no payload returned by the HGETALL for this key, this means the key did not exist.
                    # We will remove the key from the registry so we no longer have to access it next time.
                    to_delete.add(job_id)
                    # ZREM can be called with multiple keys at once, but with enough keys we'll get a broken pipe error,
                    # so to make sure we can really clean up a huge amount of job IDs, we delete in chunks.
                    if len(to_delete) >= settings.REDIS_ZREM_CHUNK_SIZE:
                        # Do not use the pipeline to delete here as we want to delete immediately
                        # Deleting using the pipeline would send too many keys at once and also overwhelm Redis.
                        queue.connection.zrem(registry.key, *to_delete)
                        to_delete = set()

            if to_delete:
                queue.connection.zrem(registry.key, *to_delete)


class Group(models.Model):
    id = models.UUIDField(default=uuid.uuid4, primary_key=True, editable=False)
    name = models.CharField(max_length=64)
    public = models.BooleanField(default=False)
    memberships = GenericRelation('users.Right', 'content_id', related_query_name='group_target')
    use_in_new_project = models.BooleanField(default=False)

    def add_member(self, user, level):
        return Right.objects.create(
            content_object=self,
            user=user,
            level=level,
        )


class OAuthStatus(Enum):
    Created = 'created'
    Done = 'done'
    Error = 'error'


class OAuthCredentials(models.Model):
    id = models.UUIDField(default=uuid.uuid4, primary_key=True, editable=False)
    user = models.ForeignKey('users.User', on_delete=models.CASCADE, related_name='credentials')
    provider_url = models.URLField()
    status = EnumField(OAuthStatus, default=OAuthStatus.Created)
    token = models.CharField(max_length=64, blank=True, null=True)
    refresh_token = models.CharField(max_length=64, blank=True, null=True)
    expiry = models.DateTimeField(blank=True, null=True)
    account_name = models.CharField(max_length=100, blank=True, null=True)

    def __str__(self):
        token_str = f"- {self.token[:6]}" if self.token else ''
        return f"{self.provider_name} - {self.user.email} {token_str}".strip()

    # Make this constant into a database field to be able to handle OAuth with services other than GitLab
    provider_name = 'gitlab'

    @property
    def provider_class(self):
        return get_provider(self.provider_name)

    @property
    def provider(self):
        return self.provider_class(credentials=self)

    @property
    def git_provider_class(self):
        from arkindex.process.providers import from_oauth
        return from_oauth(self.provider_name)

    @property
    def git_provider(self):
        return self.git_provider_class(credentials=self)

    @property
    def expired(self):
        return self.expiry is None or self.expiry < datetime.now(timezone.utc)

    class Meta:
        verbose_name = 'OAuth credentials'
        verbose_name_plural = 'OAuth credentials'


class Scope(Enum):

    UploadS3Image = 'upload_s3_image'
    """
    Allows uploading images to an S3 bucket using the CreateImage endpoint and `s3_put_url`.
    This could allow someone to upload very large files or make DoS attacks on the local IIIF image server.
    """

    CreateIIIFImage = 'create_iiif_image'
    """
    Allows adding images to Arkindex from external IIIF servers using the CreateIIIFImage endpoint.
    This could allow someone to make Arkindex perform DoS attacks on other IIIF servers due to the image checks.
    """

    S3Ingest = 's3_ingest'
    """
    Allows access to S3 ingestion features.
    """


class UserScope(models.Model):
    user = models.ForeignKey('users.User', related_name='user_scopes', on_delete=models.CASCADE)
    scope = EnumField(Scope, max_length=50)

    class Meta:
        unique_together = (
            'user', 'scope'
        )
